/*

 Automatically include the right filter code + common routines

 filter.cpp
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
 */

#include "filter.h"

////////////////////////////////////////////////////////////////////////
// Include FFT specific code:

#if defined FFT_NR
#include "filter_nr.cpp"
#elif defined FFT_FFTW
#include "filter_fftw.cpp"
#else
#error Please define here which source code to use.
#endif

////////////////////////////////////////////////////////////////////////
// Common routines to both codes:

float filterFunction(float px, float py, float pz) {
	
	// px, py, pz are assumed to be in the range [0,1] here
	if (px>0.5) px=1-px; if (py>0.5) py=1-py; if (pz>0.5) pz=1-pz;
	// fix me: should use lattice momentum here...?
  float temp = px*px+py*py+pz*pz; 
  temp = sqrt(4*temp/3);
	
	switch (FILTER_FUNC) {
		case 0:
			return 1.0;
		case 1:
  		return exp(-temp/FILTER_WIDTH);
  	case 2:
  		if (temp<=FILTER_WIDTH)
  			return 1.0;
  		else
  			return 0.0;
  	case 3:
  		return 1/(temp+FILTER_WIDTH);
  	default:
  		return 1.0;
  }
  
}

void filterFields() {

	cout << "==> Filtering Fields" << endl;

  SUNmatrix ***temp,norm;
  
  // allocate memory for temporary storage variable
	temp = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) temp[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) temp[sx][sy] = new SUNmatrix[NUM+2];
	
	switch (INITIAL_CONDITION_TYPE)
	{
		case 0:
			// w fields
			for (int j=0;j<NW;j++) { 
				
				if (MSG_FILTER) cout << "====> Filtering W #" << j+1 << endl;
				
				for (int sx=1;sx<=NUMT;sx++) 
				for (int sy=1;sy<=NUMT;sy++)
		    for (int sz=1;sz<=NUM;sz++) { 
		      temp[sx][sy][sz] = w[sx][sy][sz][j];
		    }
		    
		    if (MSG_FILTER) {
			    cout << "FILTER DEBUG: before passing to filter" << endl;
			    for (int sx=0;sx<NUMT;sx++) {
			    	for (int sy=0;sy<NUMT;sy++) {
			    		cout << temp[sx+1][sy+1][1].a << " ";
			    	}
			    	cout << endl;
			    }
			  }
		           
				filterField(temp);
		
		    if (MSG_FILTER) {
			    cout << "FILTER DEBUG: after passing to filter" << endl;
			    for (int sx=0;sx<NUMT;sx++) {
			    	for (int sy=0;sy<NUMT;sy++) {
			    		cout << temp[sx+1][sy+1][1].a << " ";
			    	}
			    	cout << endl;
			    }
			  }
		
		    // copy back to w structure
				for (int sx=1;sx<=NUMT;sx++) 
		    for (int sy=1;sy<=NUMT;sy++)
		    for (int sz=1; sz<=NUM;sz++) { 
		      w[sx][sy][sz][j] = temp[sx][sy][sz];
		    }
		
		  }		
			
			// gaurantee that filtered configuration obeys gauss law
			// by subtracting average over directions; set ow to match
			for (int sx=1;sx<=NUMT;sx++) {
		    for (int sy=1;sy<=NUMT;sy++) {
		      for (int sz=1; sz<=NUM;sz++) {
		      	norm.reset(); 
		    		for (int j=0;j<NW;j++) norm += w[sx][sy][sz][j];
		    		for (int j=0;j<NW;j++)
				    {
				      w[sx][sy][sz][j]-=norm/NW;
				      ow[sx][sy][sz][j]=w[sx][sy][sz][j];
				    }
		    	}
		    }
		  }
		  break;
		case 1:
			// b fields - z direction			
			for (int sx=1;sx<=NUMT;sx++) 
			  for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1;sz<=NUM;sz++) 
	          temp[sx][sy][sz] = toAlgebraElement(u[sx][sy][sz])/2;
			filterField(temp);
	    // convert back to links
			for (int sx=1;sx<=NUMT;sx++) 
	      for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1; sz<=NUM;sz++) {
	          #if defined SU2 || defined SU2double
	            u[sx][sy][sz].a = u[sx][sy][sz].a;
							u[sx][sy][sz].b = temp[sx][sy][sz].a;
							u[sx][sy][sz].c = temp[sx][sy][sz].b;
							u[sx][sy][sz].d = temp[sx][sy][sz].c;
						#elif defined SU3 || defined SU3double || defined SU3tvmet
							u[sx][sy][sz] = SUNLieMatrix(tr(u[sx][sy][sz])/2,temp[sx][sy][sz].element(0),temp[sx][sy][sz].element(1),
							                               temp[sx][sy][sz].element(2),temp[sx][sy][sz].element(3),temp[sx][sy][sz].element(4),
							                               temp[sx][sy][sz].element(5),temp[sx][sy][sz].element(6),temp[sx][sy][sz].element(7)); 
						#else
						#error Please provide your SU(N) code here.
						#endif
					  u[sx][sy][sz].normalize();
	    }
	    // x
	    for (int sx=1;sx<=NUMT;sx++) 
			  for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1;sz<=NUM;sz++) 
	          temp[sx][sy][sz] = toAlgebraElement(ux[sx][sy][sz])/2;
			filterField(temp);
	    // convert back to links
			for (int sx=1;sx<=NUMT;sx++) 
	      for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1; sz<=NUM;sz++) {
	          #if defined SU2 || defined SU2double
	            ux[sx][sy][sz].a = ux[sx][sy][sz].a;
							ux[sx][sy][sz].b = temp[sx][sy][sz].a;
							ux[sx][sy][sz].c = temp[sx][sy][sz].b;
							ux[sx][sy][sz].d = temp[sx][sy][sz].c;
						#elif defined SU3 || defined SU3double || defined SU3tvmet
							ux[sx][sy][sz] = SUNLieMatrix(tr(ux[sx][sy][sz])/2,temp[sx][sy][sz].element(0),temp[sx][sy][sz].element(1),
							                               temp[sx][sy][sz].element(2),temp[sx][sy][sz].element(3),temp[sx][sy][sz].element(4),
							                               temp[sx][sy][sz].element(5),temp[sx][sy][sz].element(6),temp[sx][sy][sz].element(7));  
						#else
						#error Please provide your SU(N) code here.
						#endif
					  ux[sx][sy][sz].normalize();
	    }
	    // y
			for (int sx=1;sx<=NUMT;sx++) 
			  for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1;sz<=NUM;sz++) 
	          temp[sx][sy][sz] = toAlgebraElement(uy[sx][sy][sz])/2;
			filterField(temp);
	    // convert back to links
			for (int sx=1;sx<=NUMT;sx++) 
	      for (int sy=1;sy<=NUMT;sy++)
	        for (int sz=1; sz<=NUM;sz++) {
	          #if defined SU2 || defined SU2double
	            uy[sx][sy][sz].a = uy[sx][sy][sz].a;
							uy[sx][sy][sz].b = temp[sx][sy][sz].a;
							uy[sx][sy][sz].c = temp[sx][sy][sz].b;
							uy[sx][sy][sz].d = temp[sx][sy][sz].c;
						#elif defined SU3 || defined SU3double || defined SU3tvmet
							uy[sx][sy][sz] = SUNLieMatrix(tr(uy[sx][sy][sz])/2,temp[sx][sy][sz].element(0),temp[sx][sy][sz].element(1),
							                               temp[sx][sy][sz].element(2),temp[sx][sy][sz].element(3),temp[sx][sy][sz].element(4),
							                               temp[sx][sy][sz].element(5),temp[sx][sy][sz].element(6),temp[sx][sy][sz].element(7)); 
						#else
						#error Please provide your SU(N) code here.
						#endif
					  uy[sx][sy][sz].normalize();
	    }	    
		  break;
		default:
			break;
	}		
	return;
}

void filterField(SUNmatrix*** field) {

	double *FT3darray;

  SUNmatrix ***temp;
  
  // allocate memory for temporary storage variable
	temp = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) temp[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) temp[sx][sy] = new SUNmatrix[NUM+2];
		
	FT3darray = new double[NUMT*NUMT*NUM*2];

  int SUNmax = field[1][1][1].maxelement();
	for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
	{	
	  for (int sx=0;sx<NUMT;sx++) 
		  for (int sy=0;sy<NUMT;sy++)
	      for (int sz=0; sz<NUM;sz++) { 
		      // get real part
		      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=field[sx+1][sy+1][sz+1].element(SUNelement);
		      // set imaginary part to zero
		      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0;
	  }
	 
	  filterFieldElement(FT3darray);
	  
	  // put the real value of the filtered distribution back into temp
	  for (int sx=0;sx<NUMT;sx++) 
		  for (int sy=0;sy<NUMT;sy++)
	      for (int sz=0; sz<NUM;sz++) {
	  		temp[sx+1][sy+1][sz+1].addelement(SUNelement, FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]);
	  }
	}
	
	// set field equal to temp
	for (int sx=0;sx<NUMT;sx++) 
	 for (int sy=0;sy<NUMT;sy++)
	    for (int sz=0; sz<NUM;sz++) { 
	  		field[sx+1][sy+1][sz+1] = temp[sx+1][sy+1][sz+1];
	}
  
	return;
}

bool powerOfTwo(int n)
{
	int i = n;
	while ( 
		(i > 0) // positive value
		&& 
		((i & 1) == 0) // divisible by two
		)
	{
		i = i >> 1; // divide by two
	}

	if (i == 1) return true; // pure power of two
	else return false; // other primes exist.
};
