/*
 
 crossplatform.h
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
*/

/* Implements platform independent functions:

   1) "nice" sleep,
   i.e. resources are freed for other processes to run.
   This function depends on the operating system. 
   
   2) mkdir
*/

#ifndef __crossplatform_h__
#define __crossplatform_h__

///////////////////////////////////////////////////////////////////
//  Sleep

#if defined WIN32
	/////////////////////////////////////////////////////
	// Sleep unter Windows:
	// #include <windows.h>
	// void Sleep(int ms); sleep in milliseconds
	//
	// Windows.h can be downloaded with the Platform SDK:
	// http://msdn.microsoft.com/platformsdk

	#include <windows.h>
	inline void mysleep(int ms)
	{
		Sleep(ms);
	};

#elif defined LINUX || __unix__
	/////////////////////////////////////////////////////
	// Sleep unter Linux:
	//	#include <unistd.h>
	//	int usleep(useconds_t useconds); sleep in microseconds.

	#include <unistd.h>
	inline void mysleep(int ms)
	{
		usleep(1000 * ms);
	};

#else
	// Please define a sleep() function for your system!
	// Default behavior is to not sleep at all:
	inline void mysleep(int ms)
	{
		return;	
	};
#endif 

///////////////////////////////////////////////////////////////////
//  mkdir

/*
    Code based on:
    http://www.galileo-press.de/openbook/c_von_a_bis_z/c_019_002.htm
*/

#include <stdio.h>
#include <stdlib.h>
#ifdef __unix__
	#include <sys/types.h>
	#include <sys/stat.h>
	inline int mymkdir(const char *path, int mode)
	{
		return mkdir(path, mode);
	};
#elif __WIN32__ // || _MS_DOS_
    #include <dir.h>
	inline int mymkdir(const char *path, int mode)
	{
		return _mkdir(path); // No mode for MS-DOS based systems.
	};
#elif WIN32
    #include <direct.h>  /* Visual C++ */
	inline int mymkdir(const char *path, int mode)
	{
		return _mkdir(path); // No mode for MS-DOS based systems.
	};
#else
	//Please define a mkdir() function for your system!
	// Default behavior is to not create anything:
	inline int mymkdir(const char *path, int mode)
	{
	};
#endif

#endif // __crossplatform_h__
