/*
 
 mpicore.h
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
*/

/* Core functionality of the MPI implementation is sourced out */

#ifndef __mpicore_h__
#define __mpicore_h__

#include <assert.h>

// External variables:
extern int myrank;
extern const bool MSG_GARBAGE;

/////////////////////////////////////////////////////////////////////
// FIFOBuffer
// (simple First-In-First-Out Buffer)

template <class T>
class FIFOBuffer
{
private:
	int maxlen;   // Maximum length of buffer
	T *pbuf;    // Buffer space
	int first;    // location of first element
	int last;     // location of last element + 1
	              // (if first==last, no element exists)

public:
	FIFOBuffer(int newmaxlen);
	~FIFOBuffer();

	int push(T newval);
	T pull();
	int empty(); // true if empty

	void dumpstate(); // for developing purposes
};


template <class T>
FIFOBuffer<T>::FIFOBuffer(int newmaxlen)
{
	maxlen = newmaxlen;
	pbuf = new T[newmaxlen];
	// ? test if allocation successful
	first = 0;
	last = 0;
};

template <class T>
FIFOBuffer<T>::~FIFOBuffer()
{
	if (pbuf != NULL) delete pbuf;
};

template <class T>
int FIFOBuffer<T>::push(T newval)
{
	int newlast;
	newlast = last + 1;
	if (newlast >= maxlen)
	{
		assert(newlast == maxlen);

		newlast = 0;
	};

	if (newlast == first)
	{
		// Buffer full
		cout << myrank << ": WARNING: FIFOBuffer::push(): Buffer full" << endl;
		return 1; // Error
	}
	else
	{
		pbuf[last] = newval;
		last = newlast;
		return 0; // OK
	};
};

template <class T>
T FIFOBuffer<T>::pull()
{
	T *preturn;
	if (first == last)
	{
		// No element in buffer
		cout << myrank << ": WARNING: FIFOBuffer::pull(): No element in buffer" << endl;
		return *( (T*) NULL) ;  // just to avoid compiler warning...
	}
	else
	{
		preturn = pbuf + first; // obtain pointer to address
		first++;
		if (first >= maxlen)
		{
			assert(first == maxlen);

			first = 0;
		};
		return *preturn;
	};
};

template <class T>
int FIFOBuffer<T>::empty()
{
	return (first == last);
};

/*
template <class T>
void FIFOBuffer<T>::dumpstate()
{
	cout << myrank << ": FIFO state Generic: ";
	int i = first;
	while (i != last)
	{
		cout << *((int *)(pbuf + i)) 
			<< (char)(*(((int *)(pbuf + i))+ 1) + 'a' )
			<< (char)(*(((int *)(pbuf + i))+ 2) + 'a' ) << ", ";
		i++;
		if (i>=maxlen) i = 0;
	};
	cout << endl;
	return;
};
*/

/////////////////////////////////////////////////////////////////////
// Garbage related functions

// List of allocated slices that are currently not in use:
template <class T>
class GarbageCan
{
private:
	T **garbagelist; 
	int garbagemax; // length of list
	int garbagenum; // actual number of garbage

public:
	GarbageCan(int newgarbagemax);
	~GarbageCan();

	T *GetNewItem();
	void DisposeItem(T *yuck);
};

template <class T>
GarbageCan<T>::GarbageCan(int newgarbagemax)
{
	garbagemax = newgarbagemax;
	garbagelist = new T*[garbagemax];
	garbagenum = 0;
};

template <class T>
GarbageCan<T>::~GarbageCan()
{
	// Clean-up:
	for (int i = 0; i<garbagenum; i++)
	{
		delete garbagelist[i];
	};
	delete garbagelist;
}

// Obtain allocated Simulation Slice - either fresh or from garbage can.
template <class T>
T *GarbageCan<T>::GetNewItem()
{
	if (garbagenum > 0)
	{
		if (MSG_GARBAGE) cout << myrank << ": New Item from Garbage" << endl;

		// Oh, we found some garbage. Let us sell this as
		// fresh simulation space
		garbagenum--;
		return garbagelist[garbagenum];
		// NOTE: We give back the last item that has been added
		//       to the garbage list. It has the added advantage
		//       that it might still be in the memory cache.
	}
	else
	{
		if (MSG_GARBAGE) cout << myrank << ": Creating new Item" << endl;

		// Garbage can empty. Let us make fresh simulation space.
		T *pslice = new T();
		return pslice;
	}
};

// Get rid of them
// NOTE: Currently there is no method implemented to collect garbage
//       from time to time. Memory can only increase.
template <class T>
void GarbageCan<T>::DisposeItem(T *yuck)
{
	// nothing to throw away?
	if (yuck == NULL) return;

	// Garbage can full?
	if (garbagenum >= garbagemax)
	{
		cout << myrank << ": WARNING: YUCK! DisposeItem(): Garbage can full!" << endl;
		delete yuck; // bye bye..
	};

	// Don't wanna touch waste. Away with it:
	garbagelist[garbagenum]=yuck;
	garbagenum++;
	return;
};
/////////////////////////////////////////////////////////////////////

#endif // __mpicore_h__
