#ifndef __su2mat_h__
#define __su2mat_h__
using namespace std;

#include <iostream>
#include <assert.h>
#include <math.h>

/* 
 * SU2 Algebra Matrix Class
 *
 * All elements in the SU(2) Algebra can be expressed as
 * 
 *   U = a t_1 + b t_2 + c t_3
 *
 * Where t_i are the SU(2) generators which are given by 
 * 1/2 of the Pauli matrices used above, i.e. t_i = 1/2 s_i.
 *
*/

template <class T>
class SU2matrix {

public:    

    T a,b,c;
    
    SU2matrix<T>(T a, T b, T c)
    {
	this->a = a; 
	this->b = b; 
	this->c = c;
    }

    SU2matrix<T>()
    { 
	this->a = 0; 
	this->b = 0; 
	this->c = 0;
    }

	void reset()
	{
	this->a = 0;
	this->b = 0;
	this->c = 0;
	}

    SU2matrix<T> operator=(const SU2matrix &m)
    {
        this->a = m.a;
        this->b = m.b;
        this->c = m.c;
        return *this;
    }

    template <class TS>
    SU2matrix<T> operator*(const TS &d)
    {
        SU2matrix<T> temp;
        temp.a = d*this->a;
        temp.b = d*this->b;
        temp.c = d*this->c;
        return temp;
    }

//    SU2matrix<T> operator*=(const T &d)
//    {
//        this->a *= d;
//        this->b *= d;
//        this->c *= d;
//        return *this;
//    }

    template <class TS>
	SU2matrix<T> operator*=(const TS &d)
    {
        this->a *= d;
        this->b *= d;
        this->c *= d;
        return *this;
    }

    template <class TS>
    SU2matrix<T> operator/(const TS &d)
    {
        SU2matrix<T> temp;
        temp.a = this->a/d;
        temp.b = this->b/d;
        temp.c = this->c/d;
        return temp;
    }

    template <class TS>
    SU2matrix<T> operator/=(const TS &d)
    {
        this->a /= d;
        this->b /= d;
        this->c /= d;
        return *this;
    }

    //template <class TS>
    //SU2matrix<T> operator*(const TS &i)
    //{
    //    SU2matrix<T> temp;
    //    temp.a = i*this->a;
    //    temp.b = i*this->b;
    //    temp.c = i*this->c;
    //    return temp;
    //}

	//template <class TS>
 //   SU2matrix<T> operator*=(const TS &i)
 //   {
 //       this->a *= i;
 //       this->b *= i;
 //       this->c *= i;
 //       return *this;
 //   }

	//template <class TS>
 //   SU2matrix<T> operator/(const TS &i)
 //   {
 //       SU2matrix<T> temp;
 //       temp.a = this->a/i;
 //       temp.b = this->b/i;
 //       temp.c = this->c/i;
 //       return temp;
 //   }

	//template <class TS>
 //   SU2matrix<T> operator/=(const TS &i)
 //   {
 //       this->a /= i;
 //       this->b /= i;
 //       this->c /= i;
 //       return *this;
 //   }

    SU2matrix<T> operator+(const SU2matrix<T> &m)
    {
        SU2matrix<T> temp;
        temp.a = this->a + m.a; 
        temp.b = this->b + m.b; 
        temp.c = this->c + m.c;
        return temp;
    }

    SU2matrix<T> operator+=(const SU2matrix<T> &m)
    {
        this->a += m.a;
        this->b += m.b;
        this->c += m.c;
        return *this;
    }

    SU2matrix<T> operator-(const SU2matrix<T> &m)
    {
        SU2matrix<T> temp;
        temp.a = this->a - m.a; 
        temp.b = this->b - m.b; 
        temp.c = this->c - m.c;
        return temp;
    }

    SU2matrix<T> operator-=(const SU2matrix<T> &m)
    {
        this->a -= m.a;
        this->b -= m.b;
        this->c -= m.c;
        return *this;
    }

    bool operator==(const SU2matrix<T> &m)
    {
     	return (this->a == m.a) && (this->b == m.b) && (this->c == m.c);
    }

	// for easy access to members
	double element(int i)
	{
		switch(i)
		{
		case 0:
			return a;
			break;
		case 1:
			return b;
			break;
		case 2:
			return c;
			break;
		default:
			assert(false); // This should never happen.
			return 0.;
		};
	};

	// for easy access to members
    template <class TS>
	void addelement(int i, TS value)
	{
		switch(i)
		{
		case 0:
			a += value;
			break;
		case 1:
			b += value;
			break;
		case 2:
			c += value;
			break;
		default:
			assert(false); // This should never happen.
		};
	};

	// Number of members to be accessed by index.
	static int maxelement()
	{
		return 3;
	}

	static int N()
	{
		return 2; // SU(N)
	}

    /* used for testing 
    gaugeMat toMatrix() {

    	gaugeMat t1,t2,t3,temp;
	dcomp i(0,1);

	t1 = 0,0.5,0.5,0;
	t2 = 0,-0.5*i,0.5*i,0;
	t3 = 0.5,0,0,-0.5;

	temp  = t1*this->a;
	temp += t2*this->b;
	temp += t3*this->c;
	
	return temp;
    }*/

    void print()
    {
        cout << "(" << a << "," << b << "," << c << ")\n";
    }
    
private:
    
};


/* 
 * SU2 Lie Group Matrix Class
 *
 * All elements in the Lie Group of SU(2) can be expressed as
 * 
 *   U = a I + i b s_1 + i c s_2 + i d s_3
 *
 * Where I is the identity matrix and s_i are the Pauli matrices
 *
*/

template <class T, class TL>
class SU2LieMatrix {

public:    

    TL a,b,c,d;
    
    SU2LieMatrix<T,TL>(TL a, TL b, TL c, TL d)
    {
        this->a = a; 
        this->b = b; 
        this->c = c;
        this->d = d;
    }

    SU2LieMatrix<T,TL>()
    { 
        this->a = 0; 
        this->b = 0; 
        this->c = 0;
        this->d = 0;
    }

	/*
	// convert Algebra group to traceless Lie Group element
	SU2LieMatrix algebra_to_liegroup(SU2matrix m)
	{
		SU2LieMatrix lm;
		// m.a contains the identity element and thus the trace.
		// .b, .c, .d are traceless
		// Factor 2 because of different normalization
		lm.a = 0.;
		lm.b = (double)(0.5 * m.a);
		lm.c = (double)(0.5 * m.b);
		lm.d = (double)(0.5 * m.c);
		return lm;
	}
	*/
	// convert Algebra group to traceless Lie Group element
	void algebra_to_liegroup(SU2matrix<T> m)
	{
		// m.a contains the identity element and thus the trace.
		// .b, .c, .d are traceless
		// Factor 2 because of different normalization
		a = 0.;
		b = 0.5 * m.a;
		c = 0.5 * m.b;
		d = 0.5 * m.c;
	}

	void normalize()
    {
        double temp = sqrt(this->a*this->a + this->b*this->b + this->c*this->c + this->d*this->d);
        this->a /= temp;
        this->b /= temp;
        this->c /= temp;
        this->d /= temp;
    }

    double getNorm()
    {
        return sqrt(this->a*this->a + this->b*this->b + this->c*this->c + this->d*this->d);
    }

	void reset()
	{
	this->a = 0;
	this->b = 0;
	this->c = 0;
	this->d = 0;
	}

    SU2LieMatrix<T,TL> operator+(const SU2LieMatrix<T,TL> &m)
    {
        SU2LieMatrix<T,TL> temp;
        temp.a = this->a + m.a; 
        temp.b = this->b + m.b; 
        temp.c = this->c + m.c;
        temp.d = this->d + m.d;
        return temp;
    }

	SU2LieMatrix<T,TL> operator+=(const SU2LieMatrix<T,TL> &m)
	{
		this->a += m.a; 
		this->b += m.b; 
		this->c += m.c;
		this->d += m.d;
		return *this;
	}

    SU2LieMatrix<T,TL> operator-(const SU2LieMatrix<T,TL> &m)
    {
        SU2LieMatrix<T,TL> temp;
        temp.a = this->a - m.a; 
        temp.b = this->b - m.b; 
        temp.c = this->c - m.c;
        temp.d = this->d - m.d;
        return temp;
    }

    template <class TS>
    SU2LieMatrix<T,TL> operator/(const TS &i)
    {
        SU2LieMatrix<T,TL> temp;
        temp.a = this->a/i;
        temp.b = this->b/i;
        temp.c = this->c/i;
        temp.d = this->d/i;
        return temp;
    }

    SU2LieMatrix<T,TL> operator*(const SU2LieMatrix<T,TL> &m)
    {
        SU2LieMatrix<T,TL> temp;
         temp.a = this->a*m.a - this->b*m.b - this->c*m.c - this->d*m.d;
         temp.b = this->a*m.b + this->b*m.a - this->c*m.d + this->d*m.c;
         temp.c = this->a*m.c + this->b*m.d + this->c*m.a - this->d*m.b;
         temp.d = this->a*m.d - this->b*m.c + this->c*m.b + this->d*m.a;
         return temp;
    }

     bool operator==(const SU2LieMatrix<T,TL> &m)
     {
	     return (this->a == m.a) && (this->b == m.b) && (this->c == m.c)
		     && (this->d == m.d); 
     }

	static int N()
	{
		return 2; // SU(N)
	}

    /* used for testing 
    gaugeMat toMatrix() {

    	gaugeMat id,s1,s2,s3,temp;
	dcomp i(0,1);

	id = 1,0,0,1; 
	s1 = 0,i,i,0;
	s2 = 0,1,-1,0;
	s3 = i,0,0,-i;

	temp  = id*this->a;
	temp += s1*this->b;
	temp += s2*this->c;
	temp += s3*this->d;

	return temp;
    }*/

    void print()
    {
        cout.precision(20);
        cout << "(" << this->a << "," << this->b << "," << this->c << "," << this->d;
        cout << ") : norm-1=" << getNorm()-1.0 << endl;
    }
    
private:
   

};

// Further declarations:
template <class T, class TS>
SU2matrix<T> operator*(const TS d,const SU2matrix<T> m)
{
    SU2matrix<T> temp;
    temp.a = d*m.a;
    temp.b = d*m.b;
    temp.c = d*m.c;
    return temp;
}

//template <class T, class TS>
//SU2matrix<T> operator*(const TS i,const SU2matrix<T> m)
//{
//    SU2matrix<T> temp;
//    temp.a = i*m.a;
//    temp.b = i*m.b;
//    temp.c = i*m.c;
//    return temp;
//}

//computes Hermitian conjugate of Lie Matrix
template <class T, class TL>
SU2LieMatrix<T,TL> Hermitian(SU2LieMatrix<T,TL> m)
{
  SU2LieMatrix<T,TL> temp;
  temp.a=m.a;
  temp.b=-m.b;
  temp.c=-m.c;
  temp.d=-m.d;
  return temp;
}

//
// transport left computes Udagger.M.U 
//
template <class T, class TL>
SU2matrix<T> transportLeft(SU2matrix<T> m,SU2LieMatrix<T,TL> u) {
	SU2matrix<T> temp;
	temp.a = (2*u.a*u.a-1)*m.a + 2*u.b*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.b*u.d-m.c*u.c);
	temp.b = (2*u.a*u.a-1)*m.b + 2*u.c*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.c*u.b-m.a*u.d);
	temp.c = (2*u.a*u.a-1)*m.c + 2*u.d*(u.b*m.a + u.c*m.b + u.d*m.c) - 2*u.a*(m.a*u.c-m.b*u.b);
	return temp;
}

//
// transport right computes U.M.Udagger
//
template <class T, class TL>
SU2matrix<T> transportRight(SU2matrix<T> m,SU2LieMatrix<T,TL> u) {
	SU2matrix<T> temp;
	temp.a = (2*u.a*u.a-1)*m.a + 2*u.b*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.b*u.d-m.c*u.c);
	temp.b = (2*u.a*u.a-1)*m.b + 2*u.c*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.c*u.b-m.a*u.d);
	temp.c = (2*u.a*u.a-1)*m.c + 2*u.d*(u.b*m.a + u.c*m.b + u.d*m.c) + 2*u.a*(m.a*u.c-m.b*u.b);
	return temp;
}

template <class T>
double tr2(SU2matrix<T> m) {
    	return (m.a*m.a + m.b*m.b + m.c*m.c)/2;
}

template <class T>
double tr(SU2matrix<T> m,SU2matrix<T> n) {
    	return (m.a*n.a + m.b*n.b + m.c*n.c)/2;
}

template <class T, class TL>
double tr(SU2LieMatrix<T,TL> m)
{
	return 2 * m.a;
}

// computes i times commuator of two SU2 matrices
template <class T>
SU2matrix<T> icommute(SU2matrix<T> u,SU2matrix<T> v) {
	SU2matrix<T> temp;
	temp.a = u.c*v.b - u.b*v.c;
	temp.b = u.a*v.c - u.c*v.a;
	temp.c = u.b*v.a - u.a*v.b;
	return temp;
}

// computes i[u,i[u,v]]
template <class T>
SU2matrix<T> cuuv(SU2matrix<T> u,SU2matrix<T> v) {
	SU2matrix<T> temp;
	temp.a  = u.a*(u.b*v.b+u.c*v.c);
	temp.a -= v.a*(u.b*u.b+u.c*u.c);
	temp.b  = u.b*(u.a*v.a+u.c*v.c);
	temp.b -= v.b*(u.a*u.a+u.c*u.c);
	temp.c  = u.c*(u.a*v.a+u.b*v.b);
	temp.c -= v.c*(u.a*u.a+u.b*u.b);
	return temp;
}

// computes a Lie element from exp(i SU2matrix) via analytic formula
template <class T>
SU2LieMatrix<T, double> mexpi(SU2matrix<T> m) {
	SU2LieMatrix<T, double> lm;
	double x = sqrt(m.a*m.a + m.b*m.b + m.c*m.c);	
	if (x==0) { lm.a=1; lm.b=0; lm.c=0; lm.d=0; return lm; }
	lm.a = cos(x/2);
	double s = sin(x/2); 
	//float s = sqrt(1-lm.a*lm.a); 
	lm.b = m.a*s/x; lm.c = m.b*s/x; lm.d = m.c*s/x;
	return lm;
}

// convert Lie Group element to Algebra group element by subtracting the trace
template <class T, class TL>
SU2matrix<T> toAlgebraElement(SU2LieMatrix<T,TL> m)
{
	SU2matrix<T> lm;
	// m.a contains the identity element and thus the trace.
	// .b, .c, .d are traceless
	// Factor 2 because of different normalization
	lm.a = 2 * m.b;
	lm.b = 2 * m.c;
	lm.c = 2 * m.d;
	return lm;
}

// aliases:
template <class T, class TL>
inline SU2matrix<T> Adjoint(SU2LieMatrix<T,TL> m)
{
	return toAlgebraElement(m);
};

#endif // __su2mat_h__
