#ifndef __su3mat_tvmet_h__
#define __su3mat_tvmet_h__
#include <math.h>
#include <tvmet/tvmet.h>
#include <tvmet/Matrix.h>
#include <tvmet/NumericTraits.h>

#define TXMAX	30	// max number of iteration for computation of matrix exponential
#define TXACC	1.0e-07 // target accuracy goal for computation of matrix exponential

using namespace std;
using namespace tvmet;

typedef complex<double> dcomp;
typedef Matrix<dcomp,3,3> tvmetMatrix;

static dcomp I(0.0,1.0);

/* 
 * SU3 matrix class represented by tvmet matrices
 *
*/

class SU3matrix {
private:
	// somehow this mechanism does not work, so it is mandatory to 
	// call initGellMannMatrices() before any SU3 matrix is instantiated.
	static bool we_should_call_initGellMannMatrices; // = true (in su3mat-tvmet.cpp)

public:
	// These static members occupy memory space only once.
	static tvmetMatrix id,s1,s2,s3,s4,s5,s6,s7,s8; 

    tvmetMatrix mat;

	static void initGellMannMatrices()
	{
		// Initialization should happen once by the first contructor called.
        id = 1,0,0,0,1,0,0,0,1;
        s1 = 0,1,0,1,0,0,0,0,0;
        s2 = 0,-I,0,I,0,0,0,0,0;
        s3 = 1,0,0,0,-1,0,0,0,0;
        s4 = 0,0,1,0,0,0,1,0,0;
        s5 = 0,0,-I,0,0,0,I,0,0;
        s6 = 0,0,0,0,0,1,0,1,0;
        s7 = 0,0,0,0,0,-I,0,I,0;
        s8 = 1. / sqrt(3.0),0,0,0, 1. / sqrt(3.0),0,0,0,-2. / sqrt(3.0);
		we_should_call_initGellMannMatrices = false;
	}

    SU3matrix()
    { 
	// commented out for efficiency 
	// need to make sure that the mat is initialized to zero when necessary by hand
	//this->mat = 0;  
		if (we_should_call_initGellMannMatrices) initGellMannMatrices();
    }

    explicit SU3matrix(const tvmetMatrix &m)
    { 
		if (we_should_call_initGellMannMatrices) initGellMannMatrices();
		this->mat = m;
    }

    // initializes a Lie group element -- no normalization is assumed here
    explicit SU3matrix(double a0, double a1, double a2, double a3, double a4, double a5, double a6, double a7, double a8) 
    {
		if (we_should_call_initGellMannMatrices) initGellMannMatrices();
        this->mat = id*a0 + s1*a1 + s2*a2 + s3*a3 + s4*a4 + s5*a5 + s6*a6 + s7*a7 + s8*a8; 
    }

    // initializes an Algebra group element
    explicit SU3matrix(double a1, double a2, double a3, double a4, double a5, double a6, double a7, double a8) 
    {
		if (we_should_call_initGellMannMatrices) initGellMannMatrices();
	// note different normalization of basis elements here
        this->mat = (s1*a1 + s2*a2 + s3*a3 + s4*a4 + s5*a5 + s6*a6 + s7*a7 + s8*a8)*0.5; 
    }

	void reset()
	{
		this->mat = 0.;
	}

    SU3matrix operator*(const SU3matrix &m)
    {	
	SU3matrix temp;
        temp.mat = this->mat*m.mat;
	return temp;
    }

    SU3matrix operator*=(const SU3matrix &m)
    {	
        this->mat = this->mat*m.mat;
        return *this;
    }

    SU3matrix operator*(const double &d)
    {
        SU3matrix temp;
        temp.mat = this->mat;
	temp.mat *= d;
        return temp;
    }

    SU3matrix operator*=(const double &d)
    {
        this->mat *= d;
        return *this;
    }

    SU3matrix operator*(const int &i)
    {
        SU3matrix temp;
        temp.mat = this->mat;
	temp.mat *= i;
        return temp;
    }

    SU3matrix operator*=(const int &i)
    {
        this->mat *= i;
        return *this;
    }

    SU3matrix operator/(const double &d)
    {
        SU3matrix temp;
        temp.mat = this->mat;
	temp.mat /= d;
        return temp;
    }

    SU3matrix operator/=(const double &d)
    {
        this->mat /= d;
        return *this;
    }

    SU3matrix operator/(const int &i)
    {
        SU3matrix temp;
        temp.mat = this->mat;
	temp.mat /= i;
        return temp;
    }

    SU3matrix operator/=(const int &i)
    {
        this->mat /= i;
        return *this;
    }

    SU3matrix operator+(const SU3matrix &m)
    {
        SU3matrix temp;
        temp.mat  = this->mat;
	temp.mat += m.mat;
        return temp;
    }

    SU3matrix operator+=(const SU3matrix &m)
    {
        this->mat += m.mat;
        return *this;
    }

    SU3matrix operator-(const SU3matrix &m)
    {
        SU3matrix temp;
        temp.mat  = this->mat;
	temp.mat -= m.mat;
        return temp;
    }

    SU3matrix operator-=(const SU3matrix &m)
    {
        this->mat -= m.mat;
        return *this;
    }

    void print()
    {
/*
		cout.precision(20);
		for(int i=0;i<3;i++)
		{
			for(int j=0;j<3;j++)
			{
				cout << real(mat(i,j)) << " + I " << imag(mat(i,j));
				if ((i<2)||(j<2)) cout << ",";
			};
		};
		cout << endl;
*/
		cout.precision(8);
		cout << "(" 
			<< mat(1-1,1-1) << "," 
			<< mat(1-1,2-1) << "," 
			<< mat(1-1,3-1) << "\n " 
			<< mat(2-1,1-1) << "," 
			<< mat(2-1,2-1) << "," 
			<< mat(2-1,3-1) << "\n " 
			<< mat(3-1,1-1) << "," 
			<< mat(3-1,2-1) << "," 
			<< mat(3-1,3-1) << ")" << endl;
	}

private:

};

// Additional function declarations:

SU3matrix toAlgebraElement(SU3matrix m);

SU3matrix operator*(const double d,const SU3matrix &m);
SU3matrix operator*(const int i,const SU3matrix &m);

// Hermitian Conjugate
SU3matrix dagger(SU3matrix &m); 
// Same function (for compatibility with SU2)
SU3matrix Hermitian(SU3matrix &m);

//
// transport left computes dagger(U).M.U
//
inline SU3matrix transportLeft(SU3matrix &m,SU3matrix &u) 
{
	// [ ] ??? This function should return an Algebra Element
	//         according to SU2 implementation. Is that correct ???
//	return dagger(u)*m*u; 
	return toAlgebraElement(dagger(u)*m*u); 
}

//
// transport right computes U.M.dagger(U)
//
inline SU3matrix transportRight(SU3matrix &m,SU3matrix &u) 
{ 
	// [ ] ??? This function should return an Algebra Element
	//         according to SU2 implementation. Is that correct ???
//	return u*m*dagger(u);
	return toAlgebraElement(u*m*dagger(u));
}

double tr(SU3matrix &m,SU3matrix &n);

double tr2(SU3matrix &m);

// computes i times commuator of two SU3 matrices
SU3matrix icommute(SU3matrix &u,SU3matrix &v);

// computes i[u,i[u,v]]
SU3matrix cuuv(SU3matrix &u,SU3matrix &v);

// computes matrix exponential using taylor expansion
// exp(i m) = 1 + i m + (i m)^2/2 + ...
SU3matrix mexpi(SU3matrix &m);

SU3matrix mexpiOld(SU3matrix &m);

// aliases:
inline SU3matrix Adjoint(SU3matrix m)
{
	return toAlgebraElement(m);
};

#endif // __su3mat_tvmet_h__
