// SU(3) matrices
// Andreas Ipp, July 10, 2006

#ifndef __su3mat_h__
#define __su3mat_h__

#include <assert.h>
#include <math.h>
#include <complex>

const int TXMAX = 30;    // max number of iteration for computation of matrix exponential
const double TXACC = 1.0e-10; //1.0e-07; // target accuracy goal for computation of matrix exponential

using namespace std;

const complex<double> I(0.0,1.0);
const double sqrt3 = 1.73205080756887729352744634151; // sqrt(3) to 30 digits

// note different normalization of basis elements here
// compared to the (original) SU3tvmet
const double LF = 0.5; // Normalization factor between Matrix and LieMatrix elements.
const double LFinv = 2.0; // Normalization factor back (should be 1/LF).

// The two class declarations defined below:
template <class T> class SU3matrix;
template <class T, class TL> class SU3LieMatrix;


/* 
 * SU3 Algebra Matrix Class
 *
 * All elements in the SU(3) Algebra can be expressed as
 *
 *   U = a_i * l_i
 *
 * where l_i are the Gell-Mann matrices
 *
        id = 1,0,0,0,1,0,0,0,1;
        s1 = 0,1,0,1,0,0,0,0,0;
        s2 = 0,-I,0,I,0,0,0,0,0;
        s3 = 1,0,0,0,-1,0,0,0,0;
        s4 = 0,0,1,0,0,0,1,0,0;
        s5 = 0,0,-I,0,0,0,I,0,0;
        s6 = 0,0,0,0,0,1,0,1,0;
        s7 = 0,0,0,0,0,-I,0,I,0;
        s8 = 1. / sqrt(3.0),0,0,0, 1. / sqrt(3.0),0,0,0,-2. / sqrt(3.0);
 *
*/

template <class T>
class SU3matrix {

public:    

    T a1,a2,a3,a4,a5,a6,a7,a8;
    
    SU3matrix<T>(T A1, T A2, T A3, T A4, T A5, T A6, T A7, T A8)
    {
		a1 = A1; 
        a2 = A2; 
        a3 = A3; 
        a4 = A4; 
        a5 = A5; 
        a6 = A6; 
        a7 = A7; 
        a8 = A8; 
    }

    SU3matrix<T>()
    {
		reset();
	}

    void set(T A1, T A2, T A3, T A4, T A5, T A6, T A7, T A8)
    {
        a1 = A1; 
        a2 = A2; 
        a3 = A3; 
        a4 = A4; 
        a5 = A5; 
        a6 = A6; 
        a7 = A7; 
        a8 = A8; 
    }

	void reset()
	{
		a1 = 0;
		a2 = 0;
		a3 = 0;
		a4 = 0;
		a5 = 0;
		a6 = 0;
		a7 = 0;
		a8 = 0;
	}

    SU3matrix<T> operator=(const SU3matrix<T> &m)
    {
        this->a1 = m.a1;
        this->a2 = m.a2;
        this->a3 = m.a3;
        this->a4 = m.a4;
        this->a5 = m.a5;
        this->a6 = m.a6;
        this->a7 = m.a7;
        this->a8 = m.a8;
        return *this;
    }

	SU3matrix<T> operator+(const SU3matrix<T> &m)
    {
		SU3matrix<T> temp;
		temp.a1 = this->a1 + m.a1; 
		temp.a2 = this->a2 + m.a2; 
		temp.a3 = this->a3 + m.a3; 
		temp.a4 = this->a4 + m.a4; 
		temp.a5 = this->a5 + m.a5; 
		temp.a6 = this->a6 + m.a6; 
		temp.a7 = this->a7 + m.a7; 
		temp.a8 = this->a8 + m.a8; 
		return temp;
    }

	SU3matrix<T> operator+=(const SU3matrix<T> &m)
    {
		a1 += m.a1; 
		a2 += m.a2; 
		a3 += m.a3; 
		a4 += m.a4; 
		a5 += m.a5; 
		a6 += m.a6; 
		a7 += m.a7; 
		a8 += m.a8; 
		return *this;
    }

    SU3matrix<T> operator-(const SU3matrix<T> &m)
    {
		SU3matrix<T> temp;
		temp.a1 = this->a1 - m.a1; 
		temp.a2 = this->a2 - m.a2; 
		temp.a3 = this->a3 - m.a3; 
		temp.a4 = this->a4 - m.a4; 
		temp.a5 = this->a5 - m.a5; 
		temp.a6 = this->a6 - m.a6; 
		temp.a7 = this->a7 - m.a7; 
		temp.a8 = this->a8 - m.a8; 
		return temp;
    }

	SU3matrix<T> operator-=(const SU3matrix<T> &m)
    {
		a1 -= m.a1; 
		a2 -= m.a2; 
		a3 -= m.a3; 
		a4 -= m.a4; 
		a5 -= m.a5; 
		a6 -= m.a6; 
		a7 -= m.a7; 
		a8 -= m.a8; 
		return *this;
    }

	template <class TS>
    SU3matrix<T> operator*(const TS &i)
	{
		SU3matrix<T> temp;
		temp.a1 = this->a1*i; 
		temp.a2 = this->a2*i; 
		temp.a3 = this->a3*i; 
		temp.a4 = this->a4*i; 
		temp.a5 = this->a5*i; 
		temp.a6 = this->a6*i; 
		temp.a7 = this->a7*i; 
		temp.a8 = this->a8*i; 
		return temp;
    }

	template <class TS>
    SU3matrix<T> operator*=(const TS &i)
	{
		a1 *= i; 
		a2 *= i; 
		a3 *= i; 
		a4 *= i; 
		a5 *= i; 
		a6 *= i; 
		a7 *= i; 
		a8 *= i; 
		return *this;
    }

	template <class TS>
	SU3matrix<T> operator/(const TS &i)
	{
		SU3matrix<T> temp;
		temp.a1 = this->a1/i; 
		temp.a2 = this->a2/i; 
		temp.a3 = this->a3/i; 
		temp.a4 = this->a4/i; 
		temp.a5 = this->a5/i; 
		temp.a6 = this->a6/i; 
		temp.a7 = this->a7/i; 
		temp.a8 = this->a8/i; 
		return temp;
    }

	template <class TS>
    SU3matrix<T> operator/=(const TS &i)
	{
		a1 /= i; 
		a2 /= i; 
		a3 /= i; 
		a4 /= i; 
		a5 /= i; 
		a6 /= i; 
		a7 /= i; 
		a8 /= i; 
		return *this;
    }

	// for easy access to members
	T element(int i)
	{
		assert((i >= 0) && (i < 8)); // valid bounds

		// it is assumed that elements are arranged contiguously, e.g.:
		assert( &a1 + 1 == &a2);

		return *(&a1 + i);  // returns a1 if i=0; a2 if i=1; a3 if i=2; etc.
	};

	// for easy access to members
	void addelement(int i, T value)
	{
		assert((i >= 0) && (i < 8)); // valid bounds

		// it is assumed that elements are arranged contiguously, e.g.:
		assert( &a1 + 7 == &a8);  // we check some other element

		*(&a1 + i) += value;
	}

	// Number of members to be accessed by index.
	static int maxelement()
	{
		return 8;
	}

	static int N()
	{
		return 3; // SU(N)
	}

    void print()
    {
		cout.precision(20);
		cout << "(" 
			<< a1 << "," 
			<< a2 << "," 
			<< a3 << "," 
			<< a4 << "," 
			<< a5 << "," 
			<< a6 << "," 
			<< a7 << "," 
			<< a8 << ")" << endl;
    }
    
};


/* 
 * SU3 Lie Group Matrix Class
 *
 * All elements in the Lie Group of SU(3) can be expressed as
 * complex 3x3 matrices
 *
 *   U = a_0 * id + a_i * l_i
 *
 * l_i are the Gell-Mann matrices
 *
        id = 1,0,0,0,1,0,0,0,1;
        s1 = 0,1,0,1,0,0,0,0,0;
        s2 = 0,-I,0,I,0,0,0,0,0;
        s3 = 1,0,0,0,-1,0,0,0,0;
        s4 = 0,0,1,0,0,0,1,0,0;
        s5 = 0,0,-I,0,0,0,I,0,0;
        s6 = 0,0,0,0,0,1,0,1,0;
        s7 = 0,0,0,0,0,-I,0,I,0;
        s8 = 1. / sqrt(3.0),0,0,0, 1. / sqrt(3.0),0,0,0,-2. / sqrt(3.0);
*/

/* Use the following MATHEMATICA program to generate the functions below:

  MATHEMATICA:

id = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}};
s1 = {{0, 1, 0}, {1, 0, 0}, {0, 0, 0}};
s2 = {{0, -I, 0}, {I, 0, 0}, {0, 0, 0}};
s3 = {{1, 0, 0}, {0, -1, 0}, {0, 0, 0}};
s4 = {{0, 0, 1}, {0, 0, 0}, {1, 0, 0}};
s5 = {{0, 0, -I}, {0, 0, 0}, {I, 0, 0}};
s6 = {{0, 0, 0}, {0, 0, 1}, {0, 1, 0}};
s7 = {{0, 0, 0}, {0, 0, -I}, {0, I, 0}};
s8 = {{1/Sqrt[3], 0, 0}, {0, 1/Sqrt[3], 0}, {0, 0, -2/Sqrt[3]}};

out = {{m11, m12, m13}, {m21, m22, m23}, {m31, m32, m33}};

AssignMatrix[mat1_, mat2_] := StringJoin @@ ((ToString[CForm[#1]] <> " = " <> \
ToString[CForm[#2]] <> ";\n") & @@@ Transpose[{Flatten[mat1], Flatten[mat2]}]);

*/
template <class T, class TL> // Template Lie class (must be complex)
class SU3LieMatrix {

public:    

    TL m11, m12, m13,
	   m21, m22, m23,
	   m31, m32, m33;
    
	/*
		MATHEMATICA:
		AssignMatrix[out, A0*id + A1*s1 + A2*s2 + A3*s3 + A4*s4 + A5*s5 + A6*s6 + A7*s7 + A8*s8]
	*/
	/*
	SU3LieMatrix_ORIG<T,TL>(T A0, T A1, T A2, T A3, T A4, T A5, T A6, T A7, T A8)
    {
		// LF (normalization factor) is not needed when initializing
		// all 9 elements.
		m11 = ((TL) A0 + (TL) A3 + (TL) A8/sqrt3);
		m12 = ((TL) A1 - I * (TL) A2);
		m13 = ((TL) A4 - I * (TL) A5);
		m21 = ((TL) A1 + I * (TL) A2);
		m22 = ((TL) A0 - (TL) A3 + (TL) A8/sqrt3);
		m23 = ((TL) A6 - I * (TL) A7);
		m31 = ((TL) A4 + I * (TL) A5);
		m32 = ((TL) A6 + I * (TL) A7);
		m33 = ((TL) A0 - (2.*((TL) A8))/sqrt3);
	}
	*/

	// All elements except for A0 are multiplied by I
	// (see implementation of SU(2))
	SU3LieMatrix<T,TL>(T A0, T A1, T A2, T A3, T A4, T A5, T A6, T A7, T A8)
    {
		// LF (normalization factor) is not needed when initializing
		// all 9 elements.
		m11 = ((TL) A0) + I * ((TL) A3 + (TL) A8/sqrt3);
		m12 = I * ((TL) A1 - I * (TL) A2);
		m13 = I * ((TL) A4 - I * (TL) A5);
		m21 = I * ((TL) A1 + I * (TL) A2);
		m22 = ((TL) A0) + I * (- (TL) A3 + (TL) A8/sqrt3);
		m23 = I * ((TL) A6 - I * (TL) A7);
		m31 = I * ((TL) A4 + I * (TL) A5);
		m32 = I * ((TL) A6 + I * (TL) A7);
		m33 = ((TL) A0) + I * (- (2.*((TL) A8))/sqrt3);
	}

	/*
		MATHEMATICA:
		AssignMatrix[out,ma1*s1+ma2*s2+ma3*s3+ma4*s4+ma5*s5+ma6*s6+ma7*s7+ma8*s8]
	*/
/*
	// implicit conversion is dangerous! => use algebra_to_liegroup() below
    SU3LieMatrix<T,TL>(const SU3matrix<T> &m)
    {
		// LF is a normalization factor
		m11 = LF * ((TL) m.a3 + (TL) m.a8/sqrt3);
		m12 = LF * ((TL) m.a1 - I * (TL) m.a2);
		m13 = LF * ((TL) m.a4 - I * (TL) m.a5);
		m21 = LF * ((TL) m.a1 + I * (TL) m.a2);
		m22 = LF * (-((TL) m.a3) + (TL) m.a8/sqrt3);
		m23 = LF * ((TL) m.a6 - I * (TL) m.a7);
		m31 = LF * ((TL) m.a4 + I * (TL) m.a5);
		m32 = LF * ((TL) m.a6 + I * (TL) m.a7);
		m33 = LF * ((-2.0*((TL) m.a8))/sqrt3);
    }
*/

	/*
		MATHEMATICA:
		AssignMatrix[out,ma1*s1+ma2*s2+ma3*s3+ma4*s4+ma5*s5+ma6*s6+ma7*s7+ma8*s8]
	*/
	// convert Algebra element to (traceless) Lie Group element
	//template <class T, class TL>
	void algebra_to_liegroup_orig(const SU3matrix<T> &m)
	{
		// LF is a normalization factor
		m11 = LF * ((TL) m.a3 + (TL) m.a8/sqrt3);
		m12 = LF * ((TL) m.a1 - I * (TL) m.a2);
		m13 = LF * ((TL) m.a4 - I * (TL) m.a5);
		m21 = LF * ((TL) m.a1 + I * (TL) m.a2);
		m22 = LF * (-((TL) m.a3) + (TL) m.a8/sqrt3);
		m23 = LF * ((TL) m.a6 - I * (TL) m.a7);
		m31 = LF * ((TL) m.a4 + I * (TL) m.a5);
		m32 = LF * ((TL) m.a6 + I * (TL) m.a7);
		m33 = LF * ((-2.0*((TL) m.a8))/sqrt3);
	}

	// Algebra and LieGroup element differ further by a factor i
	// (see su(2) implementation)
	void algebra_to_liegroup(const SU3matrix<T> &m)
	{
		// LF is a normalization factor
		m11 = I * LF * ((TL) m.a3 + (TL) m.a8/sqrt3);
		m12 = I * LF * ((TL) m.a1 - I * (TL) m.a2);
		m13 = I * LF * ((TL) m.a4 - I * (TL) m.a5);
		m21 = I * LF * ((TL) m.a1 + I * (TL) m.a2);
		m22 = I * LF * (-((TL) m.a3) + (TL) m.a8/sqrt3);
		m23 = I * LF * ((TL) m.a6 - I * (TL) m.a7);
		m31 = I * LF * ((TL) m.a4 + I * (TL) m.a5);
		m32 = I * LF * ((TL) m.a6 + I * (TL) m.a7);
		m33 = I * LF * ((-2.0*((TL) m.a8))/sqrt3);
	}

/*
	SU3LieMatrix<T,TL> operator=(const SU3matrix<T> &m)
    {
		m11 = m.a3 + m.a8/sqrt3;
		m12 = m.a1 - I * m.a2;
		m13 = m.a4 - I * m.a5;
		m21 = m.a1 + I * m.a2;
		m22 = -m.a3 + m.a8/sqrt3;
		m23 = m.a6 - I * m.a7;
		m31 = m.a4 + I * m.a5;
		m32 = m.a6 + I * m.a7;
		m33 = (-2*m.a8)/sqrt3;
        return *this;
    }
*/

    
    SU3LieMatrix<T,TL> operator=(const SU3LieMatrix<T,TL> &m)
    {
        this->m11 = m.m11;
        this->m12 = m.m12;
        this->m13 = m.m13;
        this->m21 = m.m21;
        this->m22 = m.m22;
        this->m23 = m.m23;
        this->m31 = m.m31;
        this->m32 = m.m32;
        this->m33 = m.m33;
        return *this;
    }

    SU3LieMatrix()
    {
		reset();
	}

	/*
		MATHEMATICA:
		AssignMatrix[out,0*id]
	*/
	void reset()
	{
		m11 = 0;
		m12 = 0;
		m13 = 0;
		m21 = 0;
		m22 = 0;
		m23 = 0;
		m31 = 0;
		m32 = 0;
		m33 = 0;
	}

	/*
		MATHEMATICA:
		thisin = {{thism11, thism12, thism13}, {thism21, thism22, thism23}, {thism31, thism32, thism33}};
		Tr[Conjugate[ Transpose[thisin] ].thisin]
	*/
	void normalize()
	{
		double norm = sqrt(
			real(m11)*real(m11) + imag(m11)*imag(m11)
			+ real(m12)*real(m12) + imag(m12)*imag(m12)
			+ real(m13)*real(m13) + imag(m13)*imag(m13)
			+ real(m21)*real(m21) + imag(m21)*imag(m21)
			+ real(m22)*real(m22) + imag(m22)*imag(m22)
			+ real(m23)*real(m23) + imag(m23)*imag(m23)
			+ real(m31)*real(m31) + imag(m31)*imag(m31)
			+ real(m32)*real(m32) + imag(m32)*imag(m32)
			+ real(m33)*real(m33) + imag(m33)*imag(m33) );
		if (norm > 0)
		{
			m11 /= norm;
			m12 /= norm;
			m13 /= norm;
			m21 /= norm;
			m22 /= norm;
			m23 /= norm;
			m31 /= norm;
			m32 /= norm;
			m33 /= norm;
		};
	};

	// Factor 0.5 in the norm comes from 
	// Gell-Mann matrices, whose tr^2 = 2.
	// Here we try to agree with the SU(2) implementation.
	// MAKES ONLY SENSE for trace-less objects...
	//   because tr^2(id) = 3, tr^2(Gell-Mann-matrix) = 2..
	//   they don't mix well..
	double getNorm()
	{
		return sqrt( 0.5 * 
			(real(m11)*real(m11) + imag(m11)*imag(m11)
			+ real(m12)*real(m12) + imag(m12)*imag(m12)
			+ real(m13)*real(m13) + imag(m13)*imag(m13)
			+ real(m21)*real(m21) + imag(m21)*imag(m21)
			+ real(m22)*real(m22) + imag(m22)*imag(m22)
			+ real(m23)*real(m23) + imag(m23)*imag(m23)
			+ real(m31)*real(m31) + imag(m31)*imag(m31)
			+ real(m32)*real(m32) + imag(m32)*imag(m32)
			+ real(m33)*real(m33) + imag(m33)*imag(m33)) );
	};

    SU3LieMatrix<T,TL> operator+(const SU3LieMatrix<T,TL> &m)
    {
		SU3LieMatrix<T,TL> temp;
		temp.m11 = this->m11 + m.m11;
		temp.m12 = this->m12 + m.m12;
		temp.m13 = this->m13 + m.m13;
		temp.m21 = this->m21 + m.m21;
		temp.m22 = this->m22 + m.m22;
		temp.m23 = this->m23 + m.m23;
		temp.m31 = this->m31 + m.m31;
		temp.m32 = this->m32 + m.m32;
		temp.m33 = this->m33 + m.m33;
		return temp;
    }

	SU3LieMatrix<T,TL> operator+=(const SU3LieMatrix<T,TL> &m)
    {
		this->m11 += m.m11;
		this->m12 += m.m12;
		this->m13 += m.m13;
		this->m21 += m.m21;
		this->m22 += m.m22;
		this->m23 += m.m23;
		this->m31 += m.m31;
		this->m32 += m.m32;
		this->m33 += m.m33;
		return *this;
    }

	SU3LieMatrix<T,TL> operator-(const SU3LieMatrix<T,TL> &m)
    {
		SU3LieMatrix<T,TL> temp;
		temp.m11 = this->m11 - m.m11;
		temp.m12 = this->m12 - m.m12;
		temp.m13 = this->m13 - m.m13;
		temp.m21 = this->m21 - m.m21;
		temp.m22 = this->m22 - m.m22;
		temp.m23 = this->m23 - m.m23;
		temp.m31 = this->m31 - m.m31;
		temp.m32 = this->m32 - m.m32;
		temp.m33 = this->m33 - m.m33;
		return temp;
    }
/*
 * // we define this below as operator(m, i)

	template <class TS>
	SU3LieMatrix<T,TL> operator*(const TS &i)
    {
		SU3LieMatrix<T,TL> temp;
		temp.m11 = this->m11 * i;
		temp.m12 = this->m12 * i;
		temp.m13 = this->m13 * i;
		temp.m21 = this->m21 * i;
		temp.m22 = this->m22 * i;
		temp.m23 = this->m23 * i;
		temp.m31 = this->m31 * i;
		temp.m32 = this->m32 * i;
		temp.m33 = this->m33 * i;
		return temp;
    }
*/
	template <class TS>
	SU3LieMatrix<T,TL> operator/(const TS &i)
    {
		SU3LieMatrix<T,TL> temp;
		temp.m11 = this->m11 / (TL) i;
		temp.m12 = this->m12 / (TL) i;
		temp.m13 = this->m13 / (TL) i;
		temp.m21 = this->m21 / (TL) i;
		temp.m22 = this->m22 / (TL) i;
		temp.m23 = this->m23 / (TL) i;
		temp.m31 = this->m31 / (TL) i;
		temp.m32 = this->m32 / (TL) i;
		temp.m33 = this->m33 / (TL) i;
		return temp;
    }

	/*
		MATHEMATICA:
		thisin = {{thism11, thism12, thism13}, {thism21, thism22, thism23}, {thism31, thism32, thism33}};
		min = {{Zmm11, Zmm12, Zmm13}, {Zmm21, Zmm22, Zmm23}, {Zmm31, Zmm32, Zmm33}};
		AssignMatrix[out, thisin . min]
	*/
/*
    /// We define this below as operator*(m,n)
    SU3LieMatrix<T,TL> operator*(const SU3LieMatrix<T,TL> &m)
    {
		SU3LieMatrix<T,TL> temp;
		temp.m11 = this->m11*m.m11 + this->m12*m.m21 + this->m13*m.m31;
		temp.m12 = this->m11*m.m12 + this->m12*m.m22 + this->m13*m.m32;
		temp.m13 = this->m11*m.m13 + this->m12*m.m23 + this->m13*m.m33;
		temp.m21 = this->m21*m.m11 + this->m22*m.m21 + this->m23*m.m31;
		temp.m22 = this->m21*m.m12 + this->m22*m.m22 + this->m23*m.m32;
		temp.m23 = this->m21*m.m13 + this->m22*m.m23 + this->m23*m.m33;
		temp.m31 = this->m31*m.m11 + this->m32*m.m21 + this->m33*m.m31;
		temp.m32 = this->m31*m.m12 + this->m32*m.m22 + this->m33*m.m32;
		temp.m33 = this->m31*m.m13 + this->m32*m.m23 + this->m33*m.m33;
		return temp;
    }
*/

	static int N()
	{
		return 3; // SU(N)
	}
	
    void print()
    {
		cout.precision(8);
		cout << "(" 
			<< m11 << "," 
			<< m12 << "," 
			<< m13 << "\n " 
			<< m21 << "," 
			<< m22 << "," 
			<< m23 << "\n " 
			<< m31 << "," 
		 	<< m32 << "," 
		        << m33 << ")";
		 cout << " : norm-1=" << getNorm()-1.0 << endl;

    }
    
};


// Additional function declarations:

template <class TS, class T, class TL>
SU3LieMatrix<T,TL> operator*(const SU3LieMatrix<T,TL> &m, const TS &i)
    {
	SU3LieMatrix<T,TL> temp;
	temp.m11 = m.m11 * i;
	temp.m12 = m.m12 * i;
	temp.m13 = m.m13 * i;
	temp.m21 = m.m21 * i;
	temp.m22 = m.m22 * i;
	temp.m23 = m.m23 * i;
	temp.m31 = m.m31 * i;
	temp.m32 = m.m32 * i;
	temp.m33 = m.m33 * i;
	return temp;
    }


template <class TS, class T> // scalar class
SU3matrix<T> operator*(const TS d,const SU3matrix<T> &m)
{
	SU3matrix<T> temp;
	temp.a1 = (T) d * m.a1;
	temp.a2 = (T) d * m.a2;
	temp.a3 = (T) d * m.a3;
	temp.a4 = (T) d * m.a4;
	temp.a5 = (T) d * m.a5;
	temp.a6 = (T) d * m.a6;
	temp.a7 = (T) d * m.a7;
	temp.a8 = (T) d * m.a8;
	return temp;
}

template <class TS, class T, class TL>
SU3LieMatrix<T,TL> operator*(const TS d, const SU3LieMatrix<T,TL> &m)
{
	SU3LieMatrix<T,TL> temp;
	temp.m11 = (TL) d * m.m11;
	temp.m12 = (TL) d * m.m12;
	temp.m13 = (TL) d * m.m13;
	temp.m21 = (TL) d * m.m21;
	temp.m22 = (TL) d * m.m22;
	temp.m23 = (TL) d * m.m23;
	temp.m31 = (TL) d * m.m31;
	temp.m32 = (TL) d * m.m32;
	temp.m33 = (TL) d * m.m33;
	return temp;
}

// Specialization of this class:
template <class T, class TL>
SU3LieMatrix<T,TL> operator*(const SU3LieMatrix<T,TL> &m, const SU3LieMatrix<T,TL> &n)
    {
	SU3LieMatrix<T,TL> temp;
	temp.m11 = m.m11*n.m11 + m.m12*n.m21 + m.m13*n.m31;
	temp.m12 = m.m11*n.m12 + m.m12*n.m22 + m.m13*n.m32;
	temp.m13 = m.m11*n.m13 + m.m12*n.m23 + m.m13*n.m33;
	temp.m21 = m.m21*n.m11 + m.m22*n.m21 + m.m23*n.m31;
	temp.m22 = m.m21*n.m12 + m.m22*n.m22 + m.m23*n.m32;
	temp.m23 = m.m21*n.m13 + m.m22*n.m23 + m.m23*n.m33;
	temp.m31 = m.m31*n.m11 + m.m32*n.m21 + m.m33*n.m31;
	temp.m32 = m.m31*n.m12 + m.m32*n.m22 + m.m33*n.m32;
	temp.m33 = m.m31*n.m13 + m.m32*n.m23 + m.m33*n.m33;
	return temp;
    }

// Hermitian Conjugate (conjugate of transpose)
template <class T, class TL>
SU3LieMatrix<T,TL> dagger(const SU3LieMatrix<T,TL> &m) {
	SU3LieMatrix<T,TL> temp;
	temp.m11 = conj(m.m11);
	temp.m12 = conj(m.m21);
	temp.m13 = conj(m.m31);
	temp.m21 = conj(m.m12);
	temp.m22 = conj(m.m22);
	temp.m23 = conj(m.m32);
	temp.m31 = conj(m.m13);
	temp.m32 = conj(m.m23);
	temp.m33 = conj(m.m33);
	return temp;
}

// Same function (for compatibility with SU2)
template <class T, class TL>
SU3LieMatrix<T,TL> Hermitian(const SU3LieMatrix<T,TL> &m) {
	return dagger(m);
}

//
// transport left computes dagger(U).M.U
//
template <class T, class TL>
SU3matrix<T> transportLeft(const SU3matrix<T> &M, const SU3LieMatrix<T,TL> &u)
{ 
	SU3LieMatrix<T,TL> m;
	m.algebra_to_liegroup(M);
	return toAlgebraElement(dagger(u)*m*u); 
}

//
// transport right computes U.M.dagger(U)
//
template <class T, class TL>
SU3matrix<T> transportRight(const SU3matrix<T> &M, const SU3LieMatrix<T,TL> &u) 
{
	SU3LieMatrix<T,TL> m;
	m.algebra_to_liegroup(M);
	return toAlgebraElement(u*m*dagger(u));
}

/*
	MATHEMATICA:
	mmin = {{mm11, mm12, mm13}, {mm21, mm22, mm23}, {mm31, mm32, mm33}};
	nmin = {{nm11, nm12, nm13}, {nm21, nm22, nm23}, {nm31, nm32, nm33}};
	CForm[Tr[mmin . nmin]]
*/	
// Calculate real(trace(m*n))
template <class T>
double tr_ORIG(const SU3matrix<T> &M, const SU3matrix<T> &N) {
	SU3LieMatrix<T, complex<double> > m;
	m.algebra_to_liegroup(M);
	SU3LieMatrix<T, complex<double> > n;
	n.algebra_to_liegroup(N);
	return real(m.m11*n.m11 + m.m21*n.m12 + m.m31*n.m13 
		+ m.m12*n.m21 + m.m22*n.m22 + m.m32*n.m23 
		+ m.m13*n.m31 + m.m23*n.m32 + m.m33*n.m33);
}

// tr(M,N) = tr(m.dagger(n))
template <class T>
double tr(const SU3matrix<T> &M, const SU3matrix<T> &N) {
	SU3LieMatrix<T, complex<double> > m;
	m.algebra_to_liegroup(M);
	SU3LieMatrix<T, complex<double> > n;
	n.algebra_to_liegroup(N);
	return    real(m.m11*conj(n.m11)
			+ m.m12*conj(n.m12)
			+ m.m13*conj(n.m13)
			+ m.m21*conj(n.m21)
			+ m.m22*conj(n.m22)
			+ m.m23*conj(n.m23)
			+ m.m31*conj(n.m31)
			+ m.m32*conj(n.m32)
			+ m.m33*conj(n.m33));
}

// real(tr(M))
template <class T, class TL>
double tr(const SU3LieMatrix<T,TL> &m) {
	return    real(m.m11) + real(m.m22) + real(m.m33);
}

template <class T>
double tr2_ORIG(const SU3matrix<T> &M) {
	SU3LieMatrix<T, complex<double> > m;
	m.algebra_to_liegroup(M);
	return real(m.m11*m.m11 + m.m21*m.m12 + m.m31*m.m13 
		+ m.m12*m.m21 + m.m22*m.m22 + m.m32*m.m23 
		+ m.m13*m.m31 + m.m23*m.m32 + m.m33*m.m33);
}

// [ ] Is tr2 given by tr(m * dagger(m)) ?
// This would then be getNorm()^2
template <class T>
double tr2(const SU3matrix<T> &M) {
	SU3LieMatrix<T, complex<double> > m;
	m.algebra_to_liegroup(M);
	return    real(m.m11)*real(m.m11) + imag(m.m11)*imag(m.m11)
			+ real(m.m12)*real(m.m12) + imag(m.m12)*imag(m.m12)
			+ real(m.m13)*real(m.m13) + imag(m.m13)*imag(m.m13)
			+ real(m.m21)*real(m.m21) + imag(m.m21)*imag(m.m21)
			+ real(m.m22)*real(m.m22) + imag(m.m22)*imag(m.m22)
			+ real(m.m23)*real(m.m23) + imag(m.m23)*imag(m.m23)
			+ real(m.m31)*real(m.m31) + imag(m.m31)*imag(m.m31)
			+ real(m.m32)*real(m.m32) + imag(m.m32)*imag(m.m32)
			+ real(m.m33)*real(m.m33) + imag(m.m33)*imag(m.m33);
}


// returns the real trace
template <class T, class TL>
double trace(const SU3LieMatrix<T,TL> &m) {
	return real(m.m11) + real(m.m22) + real(m.m33);
}

// computes i times commutator of two SU3 matrices
// I*(u*v - v*u)
template <class T>
SU3matrix<T> icommuteSLOW(const SU3matrix<T> &U, const SU3matrix<T> &V)
{
	SU3LieMatrix<T, complex<double> > u(U);
	SU3LieMatrix<T, complex<double> > v(V);
	return toAlgebraElement((u*v - v*u)*I);
}

// same as above, only explicit calculation
// Should give same results up to numerical issues.
/*
	MATHEMATICA:
	icomm = I*((ma1*s1 + ma2*s2 + ma3*s3 + ma4*s4 + ma5*s5 + ma6*s6 + ma7*s7 + 
                ma8*s8).(na1*s1 + na2*s2 + na3*s3 + na4*s4 + na5*s5 + 
                na6*s6 + na7*s7 + na8*s8) 
				- (na1*s1 + na2*s2 + na3*s3 + na4*s4 + na5*s5 + na6*s6 + na7*s7 + 
                na8*s8).(ma1*s1 + ma2*s2 + ma3*s3 + ma4*s4 + ma5*s5 + 
                ma6*s6 + ma7*s7 + ma8*s8));
	AssignMatrix[outGroup,Simplify[Tr[icomm. #1]]&/@{s1,s2,s3,s4,s5,s6,s7,s8}]
*/
// computes i times commutator of two SU3 matrices
// I*(u*v - v*u)
template <class T>
SU3matrix<T> icommute(const SU3matrix<T> &m, const SU3matrix<T> &n)
{
	// LF is a normalization factor (== LFinv*LF*LF)
	SU3matrix<T> temp;
	temp.a1 = LF * (2*m.a3*n.a2 - 2*m.a2*n.a3 + m.a7*n.a4 - m.a6*n.a5 + m.a5*n.a6 - m.a4*n.a7);
	temp.a2 = LF * (-2*m.a3*n.a1 + 2*m.a1*n.a3 + m.a6*n.a4 + m.a7*n.a5 - m.a4*n.a6 - m.a5*n.a7);
	temp.a3 = LF * (2*m.a2*n.a1 - 2*m.a1*n.a2 + m.a5*n.a4 - m.a4*n.a5 - m.a7*n.a6 + m.a6*n.a7);
	temp.a4 = LF * (-(m.a7*n.a1 + m.a6*n.a2 + m.a5*n.a3 - m.a3*n.a5 - sqrt3*m.a8*n.a5 - m.a2*n.a6 - 
		m.a1*n.a7 + sqrt3*m.a5*n.a8));
	temp.a5 = LF * (m.a6*n.a1 - m.a7*n.a2 + m.a4*n.a3 - m.a3*n.a4 - sqrt3*m.a8*n.a4 - m.a1*n.a6 + 
		m.a2*n.a7 + sqrt3*m.a4*n.a8);
	temp.a6 = LF * (-(m.a5*n.a1) + m.a4*n.a2 + m.a7*n.a3 - m.a2*n.a4 + m.a1*n.a5 - m.a3*n.a7 + 
		sqrt3*m.a8*n.a7 - sqrt3*m.a7*n.a8);
	temp.a7 = LF * (m.a4*n.a1 + m.a5*n.a2 - m.a6*n.a3 - m.a1*n.a4 - m.a2*n.a5 + m.a3*n.a6 - 
		sqrt3*m.a8*n.a6 + sqrt3*m.a6*n.a8);
	temp.a8 = LF * (sqrt3*(m.a5*n.a4 - m.a4*n.a5 + m.a7*n.a6 - m.a6*n.a7));
	return temp;
}


// computes i[u,i[u,v]]
template <class T>
SU3matrix<T> cuuv(const SU3matrix<T> &u, const SU3matrix<T> &v)
{
	return icommute(u,icommute(u,v));
}

// computes matrix exponential using taylor expansion
// exp(i m) = 1 + i m + (i m)^2/2 + ...
template <class T>
SU3LieMatrix<T, complex<double> > mexpi(const SU3matrix<T> &m) {
	SU3LieMatrix<T, complex<double> > lm(1,0,0,0,0,0,0,0,0);
	SU3LieMatrix<T, complex<double> > x;
	x.algebra_to_liegroup(m); 
	// Factor I is explicitely included in algebra_to_liegroup()
	// x = x * I; 
	SU3LieMatrix<T, complex<double> > t(1,0,0,0,0,0,0,0,0);
	int i;

	for (i=1;i<=TXMAX;i++)
	{
		t = t*x/((double)i);
		lm += t;
		if (fabs(trace(t*t))<TXACC) break;
	}
	if (i==TXMAX) cout << "WARNING:  Reached Max Terms in mexpi" << endl;
	return lm;
}

/*
	MATHEMATICA:
	outGroup={a1,a2,a3,a4,a5,a6,a7,a8};
	AssignMatrix[outGroup,Tr[out. #1]&/@{s1,s2,s3,s4,s5,s6,s7,s8}]
*/
// convert Lie Group element to Algebra group element by subtracting the trace
// Instead of subtracting the trace, we project out all other elements.
template <class T, class TL>
SU3matrix<T> toAlgebraElement_ORIG(const SU3LieMatrix<T,TL> &lm)
{
	SU3matrix<T> m;
	// LFinv is the inverse normalization factor between LieMatrix and matrix.
	m.a1 = (T) LFinv * 0.5*(real(lm.m12) + real(lm.m21));
	m.a2 = (T) - LFinv * 0.5*(imag(lm.m12) - imag(lm.m21));  // real(I*..) = -imag(..)
	m.a3 = (T) LFinv * 0.5*(real(lm.m11) - real(lm.m22));
	m.a4 = (T) LFinv * 0.5*(real(lm.m13) + real(lm.m31));
	m.a5 = (T) - LFinv * 0.5*(imag(lm.m13) - imag(lm.m31));  // real(I*..) = -imag(..)
	m.a6 = (T) LFinv * 0.5*(real(lm.m23) + real(lm.m32));
	m.a7 = (T) - LFinv * 0.5*(imag(lm.m23) - imag(lm.m32));  // real(I*..) = -imag(..)
	m.a8 = (T) LFinv * (0.5*(real(lm.m11) + real(lm.m22)) - real(lm.m33)) / sqrt3;
	// Don't need factor 2 for different normalization
	// as the whole matrix is stored instead of the basis,
	// but need factor 2 for tr(..)= 2.
	return m;
}

//// ???? Has this ADHOC prescription any theoretical justification????
// SO, here we subtract not only the trace, but we first divide by i.
// Different from the ORIG routine, we obtain:
// real(-i X) -> imag(X)
// imag(-i X) -> -real(X) 
template <class T, class TL>
SU3matrix<T> toAlgebraElement(const SU3LieMatrix<T,TL> &lm)
{
	SU3matrix<T> m;
	// LFinv is the inverse normalization factor between LieMatrix and matrix.
	m.a1 = (T) LFinv * 0.5*(imag(lm.m12) + imag(lm.m21));
	m.a2 = (T) LFinv * 0.5*(real(lm.m12) - real(lm.m21));  // real(I*..) = -imag(..)
	m.a3 = (T) LFinv * 0.5*(imag(lm.m11) - imag(lm.m22));
	m.a4 = (T) LFinv * 0.5*(imag(lm.m13) + imag(lm.m31));
	m.a5 = (T) LFinv * 0.5*(real(lm.m13) - real(lm.m31));  // real(I*..) = -imag(..)
	m.a6 = (T) LFinv * 0.5*(imag(lm.m23) + imag(lm.m32));
	m.a7 = (T) LFinv * 0.5*(real(lm.m23) - real(lm.m32));  // real(I*..) = -imag(..)
	m.a8 = (T) LFinv * (0.5*(imag(lm.m11) + imag(lm.m22)) - imag(lm.m33)) / sqrt3;
	// Don't need factor 2 for different normalization
	// as the whole matrix is stored instead of the basis,
	// but need factor 2 for tr(..)= 2.
	return m;
}
// aliases:
template <class T, class TL>
inline SU3matrix<T> Adjoint(SU3LieMatrix<T,TL> m)
{
	return toAlgebraElement(m);
};


#endif // __su3mat_h__
