#include <iostream>
#include <stdlib.h>
#include <stdio.h>

// For Windows, we need to set the following to access M_PI
#if defined WIN32
#define _USE_MATH_DEFINES
#endif 

#include <assert.h>
#include <math.h>

#include <fftw3.h>

#include "outputroutines.h"
#include "sunmat.h"
#include "solve.h"

// Windows does not define the _round_ function, so we have to do it by hand:
#if defined WIN32
inline double round(double r)
{
	return floor(r + 0.5);
};
#endif

/*
// Short demo how to use FFTW
void testtransform()
{
	cout << "Start testtransform" << endl;
	int N = 100;
	fftw_complex *in, *out;
	fftw_plan p;
	//...
	in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
	out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
	p = fftw_plan_dft_1d(N, in, out, FFTW_FORWARD, FFTW_ESTIMATE);
     fftw_plan fftw_plan_dft_3d(int nx, int ny, int nz,
                                fftw_complex *in, fftw_complex *out,
                                int sign, unsigned flags);
	//...
	fftw_execute(p); // repeat as needed
	//...
	fftw_destroy_plan(p);
	fftw_free(in); fftw_free(out);
	cout << "End testtransform" << endl;
};
*/

// Static structure to keep the FFT matrix and the according plan in memory
fftw_complex *my_fftw_matrix;
int my_fftw_size;
double my_fftw_norm;
fftw_plan my_fftw_plan;
fftw_plan my_fftw_plan_inverse;
double *cos_NUMT; // Contains a table of 1-cos(2 Pi x / NUMT)
double *cos_NUM;

void fft_init()
{
	if (MSG_FFT) cout << "Master: FFTW init()" << endl;

	// Allocate the memory
	my_fftw_size = NUMT * NUMT * NUM;
	my_fftw_norm = (double) my_fftw_size;

	my_fftw_matrix = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * my_fftw_size);

	// Make a plan (i.e. determine the fastest way to 
	// calulate the FFT on this machine.)

	const unsigned my_fftw_flags = FFTW_MEASURE;
		// Flag can be FFTW_MEASURE or FFTW_ESTIMATE or other.

	my_fftw_plan = fftw_plan_dft_3d(NUMT, NUMT, NUM, 
		my_fftw_matrix, my_fftw_matrix, FFTW_FORWARD, my_fftw_flags);

	fftw_print_plan(my_fftw_plan);

	my_fftw_plan_inverse = fftw_plan_dft_3d(NUMT, NUMT, NUM, 
		my_fftw_matrix, my_fftw_matrix, FFTW_BACKWARD, my_fftw_flags);

	fftw_print_plan(my_fftw_plan_inverse);

	// Fill the cosine tables:
	cos_NUMT = new double[NUMT];
	cos_NUM = new double[NUM];

	for (int i=0; i<NUMT; i++)
	{
		cos_NUMT[i] = 1. - cos(2 * M_PI * i / (double) NUMT);
	};
	for (int i=0; i<NUM; i++)
	{
		cos_NUM[i] = 1. - cos(2 * M_PI * i / (double) NUM);
	};

	// Set load/save buffer for angle dumps:
	// Share memory with FFTW buffer
	loadsavebuffer = (SUNmatrix_BASIC_TYPE *)(void *) my_fftw_matrix;

	// Should be obvious that we fit in there (could cancel my_fftw_size, but to be explicit...)
	assert(sizeof(SUNmatrix_BASIC_TYPE) * my_fftw_size <= sizeof(fftw_complex) * my_fftw_size);

	if (MSG_FFT) cout << "Master: FFTW init() OK" << endl;
}

inline void do_fft()
{
	//cout << "Master: FFTW do_fft()" << endl;

	fftw_execute(my_fftw_plan_inverse);
}

inline void do_fft_inverse()
{
	//cout << "Master: FFTW do_fft_inverse()" << endl;

	fftw_execute(my_fftw_plan);
	for (int i=0; i<my_fftw_size; i++)
	{
		my_fftw_matrix[i][0] /= my_fftw_norm; // normalize real
		my_fftw_matrix[i][1] /= my_fftw_norm; // and imaginary part.
	};
}

void fft_end()
{
	cout << "Master: FFTW end()" << endl;
	delete[] cos_NUMT;
	delete[] cos_NUM;
	fftw_destroy_plan(my_fftw_plan);
	fftw_destroy_plan(my_fftw_plan_inverse);
	fftw_free(my_fftw_matrix);

	// loadsavebuffer is freed together with my_fftw_matrix.
}

void u_divergence_to_fft(int index, double *conv)
{
	//cout << "Master: Calculate u divergence" << endl;

	//generate divergences
	int n = 0;
	double re; // real part
	for (int sx=1;sx<=NUMT;sx++) 
	 for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	{
		re = Adjoint(u[sx][sy][sz]).element(index);
		re -= Adjoint(u[sx][sy][sz-1]).element(index);
		re += Adjoint(ux[sx][sy][sz]).element(index);
		re -= Adjoint(ux[sx-1][sy][sz]).element(index);
		re += Adjoint(uy[sx][sy][sz]).element(index);
		re -= Adjoint(uy[sx][sy-1][sz]).element(index);

		my_fftw_matrix[n][0] = re; // real part
		my_fftw_matrix[n][1] = 0.; // imaginary part
		n++;

		*conv += re * re;
	}
	assert( n == NUMT * NUMT * NUM );  // we assigned values to all fields.
};

// Boundary conditions only for P (pix, piy, piz) and U (u, ux, uy)
void enforceBCsOnlyforPU()
{

	//fill in z=0,NUM+1 values for main x,y
	for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++) { 
		pix[sx][sy][0] = pix[sx][sy][NUM]; 	pix[sx][sy][NUM+1] = pix[sx][sy][1];
		piy[sx][sy][0] = piy[sx][sy][NUM]; 	piy[sx][sy][NUM+1] = piy[sx][sy][1];
		ez[sx][sy][0] = ez[sx][sy][NUM]; 	ez[sx][sy][NUM+1] = ez[sx][sy][1];
		u[sx][sy][0] = u[sx][sy][NUM]; 		u[sx][sy][NUM+1] = u[sx][sy][1];
		ux[sx][sy][0] = ux[sx][sy][NUM]; 	ux[sx][sy][NUM+1] = ux[sx][sy][1];
		uy[sx][sy][0] = uy[sx][sy][NUM]; 	uy[sx][sy][NUM+1] = uy[sx][sy][1];
	}

	//now fill in y=0,NUMT+1 values for main x and all z
	for (int sx=1;sx<=NUMT;sx++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
		pix[sx][0][sz] = pix[sx][NUMT][sz]; 	pix[sx][NUMT+1][sz] = pix[sx][1][sz];
		piy[sx][0][sz] = piy[sx][NUMT][sz]; 	piy[sx][NUMT+1][sz] = piy[sx][1][sz];
		ez[sx][0][sz] = ez[sx][NUMT][sz]; 	ez[sx][NUMT+1][sz] = ez[sx][1][sz];
		u[sx][0][sz] = u[sx][NUMT][sz]; 	u[sx][NUMT+1][sz] = u[sx][1][sz];
		ux[sx][0][sz] = ux[sx][NUMT][sz]; 	ux[sx][NUMT+1][sz] = ux[sx][1][sz];
		uy[sx][0][sz] = uy[sx][NUMT][sz]; 	uy[sx][NUMT+1][sz] = uy[sx][1][sz];
	}

	//finally fill in x=0,NUMT+1 values for general y,z

	for (int sy=0;sy<=NUMT+1;sy++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
		pix[0][sy][sz] = pix[NUMT][sy][sz]; 	pix[NUMT+1][sy][sz] = pix[1][sy][sz];
		piy[0][sy][sz] = piy[NUMT][sy][sz]; 	piy[NUMT+1][sy][sz] = piy[1][sy][sz];
		ez[0][sy][sz] = ez[NUMT][sy][sz]; 	ez[NUMT+1][sy][sz] = ez[1][sy][sz];
		u[0][sy][sz] = u[NUMT][sy][sz]; 		u[NUMT+1][sy][sz] = u[1][sy][sz];
		ux[0][sy][sz] = ux[NUMT][sy][sz]; 	ux[NUMT+1][sy][sz] = ux[1][sy][sz];
		uy[0][sy][sz] = uy[NUMT][sy][sz]; 	uy[NUMT+1][sy][sz] = uy[1][sy][sz];
	}

}


//fixes Coulomb gauge in 3d
void gaugefix3d(int maxiter)
{
//  	cout << "Master: gaugefix3d()" << endl;

	int nx=NUMT;
	int ny=NUMT;
	int nz=NUM;
	int n=0;

	SUNLieMatrix ***gux;
	gux = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) gux[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) gux[sx][sy] = new SUNLieMatrix[NUM+2];

	double conv=1;
	int gfiter;
	for (gfiter=0;gfiter<maxiter;gfiter++)
	{ 
		if (conv/NUMT/NUMT/NUM<1.e-15) break;

		conv=0;

		// Reset gux
		for (int sx=1;sx<=NUMT;sx++)
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			gux[sx][sy][sz].reset();
		}

		// Big loop for all components of u, ux, uy
		int SUNmax = pix[1][1][1].maxelement();
		for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
		{
			u_divergence_to_fft(SUNelement, &conv); //generate divergences

			do_fft(); //solve Poisson's equation: (L1a)
	      
			my_fftw_matrix[0][0] = 0.;  // remove constant part
			my_fftw_matrix[0][1] = 0.;

			n = 0;
			for (int j=0;j<nx;j++) //solve for X
			for (int k=0;k<ny;k++)
			for (int m=0;m<nz;m++)
			{
				if (n != 0)
				{
					//double fact = 0.5 / (cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
					double fact = -0.5 / (cos_NUMT[j] + cos_NUMT[k] + cos_NUM[m]);  // faster
					my_fftw_matrix[n][0] *= fact;
					my_fftw_matrix[n][1] *= fact;
				}
				n++;
			}
			assert( n == NUMT * NUMT * NUM );
			
			do_fft_inverse(); //transform back

	//	  cout << "FFTW:::" << gfiter << " >>> "; 
	//		for (int i =0; i<10; i++)
	//			cout << 0.5*my_fftw_matrix[i][0] << " + i " << 0.5*my_fftw_matrix[i][1] <<", ";

	//		cout << endl;

			//assume GT has been done already and gux is empty
			n = 0;
			SUNmatrix temp;
			SUNLieMatrix temp2;
			for (int sx=0;sx<NUMT;sx++) 
			for (int sy=0;sy<NUMT;sy++)
			for (int sz=0;sz<NUM;sz++)
			{
				temp.reset();
				temp.addelement(SUNelement, my_fftw_matrix[n][0]);
				temp2.algebra_to_liegroup(temp);
				gux[sx+1][sy+1][sz+1] += temp2;
				n++;
			}
			assert( n == NUMT * NUMT * NUM );
		};

		//// 
      
		//printf("just for debugging: %e\n",conv/NUMT/NUMT/NUM);

		//prepare gauge transform

		SUNmatrix temp;
		for (int sx=1;sx<=NUMT;sx++)
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			temp = toAlgebraElement(gux[sx][sy][sz]);
			//gux[sx][sy][sz]=mexpi(temp*(-1.0)); // !! Compiler error: ambiguous line, as (double)1.0 could be float or int
			gux[sx][sy][sz]=mexpi(-1 * temp);  // V1
			gux[sx][sy][sz].normalize();
		}

		//BCs
		for (int sx=1;sx<=NUMT;sx++) 
		for (int sy=1;sy<=NUMT;sy++) { 
			gux[sx][sy][0] = gux[sx][sy][NUM];  gux[sx][sy][NUM+1] = gux[sx][sy][1];
		}
		for (int sx=1;sx<=NUMT;sx++) 
		for (int sz=0;sz<=NUM+1;sz++) { 
			gux[sx][0][sz] = gux[sx][NUMT][sz]; gux[sx][NUMT+1][sz] = gux[sx][1][sz];
		}
		for (int sy=0;sy<=NUMT+1;sy++) 
		for (int sz=0;sz<=NUM+1;sz++) { 
			gux[0][sy][sz] = gux[NUMT][sy][sz]; gux[NUMT+1][sy][sz] = gux[1][sy][sz];
		}

		//do gauge Transform

		for (int sx=1;sx<=NUMT;sx++) 
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			ux[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*ux[sx][sy][sz]*gux[sx+1][sy][sz];
			uy[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*uy[sx][sy][sz]*gux[sx][sy+1][sz];
			u[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*u[sx][sy][sz]*gux[sx][sy][sz+1];
			pix[sx][sy][sz]=transportLeft(pix[sx][sy][sz],gux[sx][sy][sz]);
			piy[sx][sy][sz]=transportLeft(piy[sx][sy][sz],gux[sx][sy][sz]);
			ez[sx][sy][sz]=transportLeft(ez[sx][sy][sz],gux[sx][sy][sz]);

			ux[sx][sy][sz].normalize();  //reunitarize
			uy[sx][sy][sz].normalize();
			u[sx][sy][sz].normalize();
		}

		enforceBCsOnlyforPU();
      
    } // end of main loop

	//just for debugging:
	//printf("debug: %e\n",conv/NUMT/NUMT/NUM);

	// For Debugging:
	//cout << "Converged at "<< gfiter << "iterations, conv = ";
	//cout <<conv/NUMT/NUMT/NUM << endl;

	if (conv/NUMT/NUMT/NUM>1.e-15)
	{
		cout << "GF warning: did not converge at "<< maxiter << "iterations, conv = ";
		cout <<conv/NUMT/NUMT/NUM << endl;
	}

	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) delete [] gux[sx][sy];
	for (int sx=0;sx<NUMT+2;sx++) delete [] gux[sx];
	delete [] gux; 

}

///////////////////////////////////////////////////
// Helper functions:

  //load momentum
void load_momentum(SUNmatrix ***field, int SUNelement)
{
	int n = 0;
	for (int sx=1;sx<=NUMT;sx++)
	for (int sy=1;sy<=NUMT;sy++)
	for (int sz=1;sz<=NUM;sz++)
	{
		my_fftw_matrix[n][0] = field[sx][sy][sz].element(SUNelement);
		my_fftw_matrix[n][1] = 0.;
		n++;
	}
	assert( n == NUMT * NUMT * NUM );
};

void count_bin_content(int maxptot, int bins, int *countptot)
{
	int n=0;
	for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	for (int sz=0;sz<NUM;sz++)
	{
		if (n!=0)
		{
			//double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
			double disprel=sqrt(2.*(cos_NUMT[sx] + cos_NUMT[sy] + cos_NUM[sz]));
			int ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
			int costheta= (int) round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));
			assert( ptot < maxptot );
			assert( costheta < bins );

			countptot[ptot*bins+costheta]++;
		};
		n++;
	};
	assert( n == NUMT * NUMT * NUM );
};

void assign_to_bins(int maxptot, int bins, double *dndp, double *dndpz)
{
	int n = 0;
	for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	for (int sz=0;sz<NUM;sz++)
	{
		if (n!=0)
		{
			//double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
			double disprel=sqrt(2.*(cos_NUMT[sx] + cos_NUMT[sy] + cos_NUM[sz]));
			int ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
			int costheta= (int) round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));
			assert( ptot < maxptot );
			assert( costheta < bins );

			//real part
			dndp[ptot*bins+costheta]+=my_fftw_matrix[n][0]*my_fftw_matrix[n][0]/disprel;
			//imaginary part
			dndp[ptot*bins+costheta]+=my_fftw_matrix[n][1]*my_fftw_matrix[n][1]/disprel;

			//dndpz only
			//disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));
			disprel=sqrt(2.*(cos_NUM[sz]));

			if (sz != 0)
			{
				if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
				{
					// Take mean over all x and y
					dndpz[sz]+=my_fftw_matrix[n][0]*my_fftw_matrix[n][0]/disprel;
					//imaginary part
					dndpz[sz]+=my_fftw_matrix[n][1]*my_fftw_matrix[n][1]/disprel;
				}
				else
				{
					// Original code only probes at x==y==0
					dndpz[sz]+=my_fftw_matrix[sz][0]*my_fftw_matrix[sz][0]/disprel;
					//imaginary part
					dndpz[sz]+=my_fftw_matrix[sz][1]*my_fftw_matrix[sz][1]/disprel;
				};

			};
		};
		n++;
	}
	assert( n == NUMT * NUMT * NUM );
};

void measure_angle_field(SUNmatrix ***field, int maxptot, int bins, double *dndp, double *dndpz)
{
	int SUNmax = field[1][1][1].maxelement();
	for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
	{
		load_momentum(field, SUNelement);
		do_fft();
		assign_to_bins(maxptot, bins, dndp, dndpz);
	};
};


void measureangleforbins(double time, int usebins)
{
	fstream out,outpz;
	char fname[255];

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	int maxptot = (int)sqrt((double)(NUMT*NUMT*2+NUM*NUM)/9.)+1;
	//number of costheta bins
	//int bins = 10+1;
	int bins = usebins + 1;  // take value from params.txt
	int totbins=maxptot*bins;
	int *countptot = new int[totbins]; //need to count number of degenerate pt's
	double *dndp = new double[totbins]; 
	double *dndpz = new double[NUM+1]; 

	for (int sx=0;sx<totbins;sx++)
	{
		countptot[sx]=0;
		dndp[sx]=0.0;
	}
	for (int sx=0;sx<=NUM;sx++)
	{
		dndpz[sx]=0.0;
	}

	count_bin_content(maxptot, bins, countptot);

	measure_angle_field(pix, maxptot, bins, dndp, dndpz);
	measure_angle_field(piy, maxptot, bins, dndp, dndpz);
	measure_angle_field(ez, maxptot, bins, dndp, dndpz);
	

	// write to file

	sprintf(fname,"%s/snapshot/f_%dbins_%.2f.dat",pdirname,usebins,time);
	out.open(fname, ios::out);

	sprintf(fname,"%s/snapshot/fdpz_%.2f.dat",pdirname,time);
	outpz.open(fname, ios::out);

	for (int go=0;go<maxptot;go++)
	{
		out << go << "\t";
		for (int inner=0;inner<bins;inner++)
		{
			if (countptot[go*bins+inner]!=0)
				out << dndp[go*bins+inner]/countptot[go*bins+inner] << "\t";
			else 
				out << 0.0 << "\t";
		} 
		out << endl;
	}

	for (int go=1;go<=NUM;go++)
	{
		outpz << go << "\t";
		outpz << dndpz[go]<< endl;
	}

	out.close();
	outpz.close();

	if (time == 0.)
	{
		// The first time only, we also put out the information 
		// about the distribution of events in the bins:

		sprintf(fname,"%s/snapshot/f_%dbins_count.dat",pdirname,usebins);
		out.open(fname, ios::out);

		int checksum = 0;
		for (int go=0;go<maxptot;go++)
		{
			out << go << "\t";
			for (int inner=0;inner<bins;inner++)
			{
				out << countptot[go*bins+inner] << "\t";
				checksum += countptot[go*bins+inner];
			} 
			out << endl;
		}
		assert( checksum == NUMT * NUMT * NUM - 1); // We should record ALL events.
		out.close();
	};

	//clean up
	delete [] dndp;
	delete [] dndpz;
	delete [] countptot;
}

void measureangle(double time)
{
  gaugefix3d(150);

  angledumpsave(time); // Save gauge fixed fields for optional later analysis

  measureangleforbins(time, ANGLEBINS1);  // take value from params.txt
  measureangleforbins(time, ANGLEBINS2);  // take value from params.txt
  measureangleforbins(time, ANGLEBINS3);  // take value from params.txt
}
