/*
   Output measurement of angle

   Original code by Paul Romatschke

   uses FFT routine from Numerical Recipies
*/

#include <iostream>
#include <stdlib.h>
#include <stdio.h>

// For Windows, we need to set the following to access M_PI
#if defined WIN32
#define _USE_MATH_DEFINES
#endif 

#include <math.h>

#include "outputroutines.h"
#include "sunmat.h"
#include "solve.h"

// Windows does not define the _round_ function, so we have to do it by hand:
#if defined WIN32
inline double round(double r)
{
	return floor(r + 0.5);
};
#endif

static bool can_do_FFT_NR;

bool power_of_two(int n)
{
	int i = n;
	while ( 
		(i > 0) // positive value
		&& 
		((i & 1) == 0) // divisible by two
		)
	{
		i = i >> 1; // divide by two
	}

	if (i == 1) return true; // pure power of two
	else return false; // other primes exist.
};

void fft_init()
{
	loadsavebuffer = NULL;

#if defined SU2 || defined SU2double
	int fft_error = 0;
	//if ( NUMT != NUM ) fft_error = 1;
	if ( ! power_of_two(NUMT) ) fft_error = 2;
	
	if (fft_error > 0)
	{
		cout << endl;
		cout << "   ****************************************************" << endl;
		cout << "   *                                                  *" << endl;
		cout << "   *  WARNING: FFT from Numerical Recipies does not   *" << endl;
		//if (fft_error == 1)
		//cout << "   *  work with NUMT != NUM.                          *" << endl;
		if (fft_error == 2)
		cout << "   *  work with NUMT not a strict power of two.       *" << endl;		
		cout << "   *                                                  *" << endl;
		cout << "   *  Can not calculate measureangle().               *" << endl;
		cout << "   *  Either change NUMT and/or NUM or                *" << endl;
		cout << "   *  try 'make solve.. FFT=FFTW'                     *" << endl;
		cout << "   *                                                  *" << endl;
		cout << "   ****************************************************" << endl;
		cout << endl;

		can_do_FFT_NR = false;
	}
	else
	{
		can_do_FFT_NR = true;

		// Some additional initializations:
		
		// Set load/save buffer for angle dumps:
		loadsavebuffer = new SUNmatrix_BASIC_TYPE[NUMT * NUMT * NUM];
		assert(loadsavebuffer != NULL);
	};
#else
	// FFT_NR currently only works with SU(2)
	cout << endl;
	cout << "   ****************************************************" << endl;
	cout << "   *                                                  *" << endl;
	cout << "   *  WARNING: FFT from Numerical Recipies does not   *" << endl;
	cout << "   *  work with ";
	cout.width(30); cout << left << SUNmatrix_string << "        *" << endl;
	cout << "   *                                                  *" << endl;
	cout << "   *  Can not calculate measureangle().               *" << endl;
	cout << "   *  Try 'make solve.. FFT=FFTW'                     *" << endl;
	cout << "   *                                                  *" << endl;
	cout << "   ****************************************************" << endl;
	cout << endl;
	can_do_FFT_NR = false;
#endif

	return;
};

void fft_end()
{
	if (loadsavebuffer != NULL) delete[] loadsavebuffer;

	return;
};

// Paul's original code only works with SU2
#if defined SU2 || defined SU2double

// Boundary conditions only for P (pix, piy, piz) and U (u, ux, uy)
void enforceBCsOnlyforPU()
{

//fill in z=0,NUM+1 values for main x,y
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 
	pix[sx][sy][0] = pix[sx][sy][NUM]; 	pix[sx][sy][NUM+1] = pix[sx][sy][1];
	piy[sx][sy][0] = piy[sx][sy][NUM]; 	piy[sx][sy][NUM+1] = piy[sx][sy][1];
	ez[sx][sy][0] = ez[sx][sy][NUM]; 	ez[sx][sy][NUM+1] = ez[sx][sy][1];
	u[sx][sy][0] = u[sx][sy][NUM]; 		u[sx][sy][NUM+1] = u[sx][sy][1];
	ux[sx][sy][0] = ux[sx][sy][NUM]; 	ux[sx][sy][NUM+1] = ux[sx][sy][1];
	uy[sx][sy][0] = uy[sx][sy][NUM]; 	uy[sx][sy][NUM+1] = uy[sx][sy][1];
    }
  
  //now fill in y=0,NUMT+1 values for main x and all z
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      	pix[sx][0][sz] = pix[sx][NUMT][sz]; 	pix[sx][NUMT+1][sz] = pix[sx][1][sz];
	piy[sx][0][sz] = piy[sx][NUMT][sz]; 	piy[sx][NUMT+1][sz] = piy[sx][1][sz];
	ez[sx][0][sz] = ez[sx][NUMT][sz]; 	ez[sx][NUMT+1][sz] = ez[sx][1][sz];
	u[sx][0][sz] = u[sx][NUMT][sz]; 	u[sx][NUMT+1][sz] = u[sx][1][sz];
	ux[sx][0][sz] = ux[sx][NUMT][sz]; 	ux[sx][NUMT+1][sz] = ux[sx][1][sz];
	uy[sx][0][sz] = uy[sx][NUMT][sz]; 	uy[sx][NUMT+1][sz] = uy[sx][1][sz];
    }
  
  //finally fill in x=0,NUMT+1 values for general y,z

  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      pix[0][sy][sz] = pix[NUMT][sy][sz]; 	pix[NUMT+1][sy][sz] = pix[1][sy][sz];
      piy[0][sy][sz] = piy[NUMT][sy][sz]; 	piy[NUMT+1][sy][sz] = piy[1][sy][sz];
      ez[0][sy][sz] = ez[NUMT][sy][sz]; 	ez[NUMT+1][sy][sz] = ez[1][sy][sz];
      u[0][sy][sz] = u[NUMT][sy][sz]; 		u[NUMT+1][sy][sz] = u[1][sy][sz];
      ux[0][sy][sz] = ux[NUMT][sy][sz]; 	ux[NUMT+1][sy][sz] = ux[1][sy][sz];
      uy[0][sy][sz] = uy[NUMT][sy][sz]; 	uy[NUMT+1][sy][sz] = uy[1][sy][sz];
    }

}


//fixes Coulomb gauge in 3d
void gaugefix3d(int maxiter)
{
  
  //FT
  extern void FT3d(double* pArr, int nx, int ny, int nz, int sign);

  int nx=NUMT;
  int ny=NUMT;
  int nz=NUM;
  int l=0;
  double* FT3darray = new double[NUMT*NUMT*NUM*2];
  
  SUNLieMatrix ***gux;
  gux = new SUNLieMatrix**[NUMT+2]; 
  for (int sx=0;sx<NUMT+2;sx++) gux[sx] = new SUNLieMatrix*[NUMT+2];
  for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) gux[sx][sy] = new SUNLieMatrix[NUM+2];
  
  
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++)
      {
	gux[sx][sy][sz].a=1;
	gux[sx][sy][sz].b=0;
	gux[sx][sy][sz].c=0;
	gux[sx][sy][sz].d=0;
      }


  //BCs
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 
	gux[sx][sy][0] = gux[sx][sy][NUM]; 	gux[sx][sy][NUM+1] = gux[sx][sy][1];
    }
  for (int sx=1;sx<=NUMT;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
	gux[sx][0][sz] = gux[sx][NUMT][sz]; 	gux[sx][NUMT+1][sz] = gux[sx][1][sz];
    }
  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      gux[0][sy][sz] = gux[NUMT][sy][sz]; 	gux[NUMT+1][sy][sz] = gux[1][sy][sz];
    }

  


  double conv=1;
  int gfiter;
  for (gfiter=0;gfiter<maxiter;gfiter++)
    { 
      if (conv/NUMT/NUMT/NUM<1.e-15)
	break;

      conv=0;

      //generate divergences
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(u[sx][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(ux[sx][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).a;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).a;
	    

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	    
	  }
      
      //solve Poisson's equation: (L1a)

      FT3d(FT3darray,NUMT,NUMT,NUM,1);
      
      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back

	  if (MSG_FFT)
	  {
		  cout << "NR:::" << gfiter << " >>> "; 
			for (int i =0; i<10; i++)
				cout << FT3darray[2*i] << " + i " << FT3darray[2*i+1] <<", ";
			cout << endl;
	  };

       //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].b=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];

	  if (MSG_FFT)
	  {
	  		cout << "gux " << gux[1][1][1].b << endl;
	  };

      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(ux[sx][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(u[sx][sy][sz]).b;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).b;

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	    
	  }

      FT3d(FT3darray,NUMT,NUMT,NUM,1);

      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back


      //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].c=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];



      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]=Adjoint(ux[sx][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(ux[sx-1][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(uy[sx][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(uy[sx][sy-1][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]+=Adjoint(u[sx][sy][sz]).c;
	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]-=Adjoint(u[sx][sy][sz-1]).c;

	    FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))+1]=0.0;

	    conv+=FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))]*FT3darray[2*((sx-1)*NUMT*NUM+(sy-1)*NUM+(sz-1))];
	  }

      //solve Poisson's equation: (L1c)

      FT3d(FT3darray,NUMT,NUMT,NUM,1);

      
      FT3darray[0]=0; // remove constant part
      FT3darray[1]=0;
      //conv=0;
      for (int j=0;j<nx;j++) //solve for X
	for (int k=0;k<ny;k++)
	  for (int m=0;m<nz;m++)
	  {
	    l=m+k*nz+j*ny*nz;
	    l=(l<<1);
	    if (l!=0)
	      {
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
		l++;
		FT3darray[l]=FT3darray[l]/2/(cos(2*M_PI*j/nx)+cos(2*M_PI*k/ny)+cos(2*M_PI*m/nz)-3);
	      }
	    }
      
      FT3d(FT3darray,NUMT,NUMT,NUM,-1); //transform back

      //assume GT has been done already and gux is empty
      for (int sx=0;sx<NUMT;sx++) 
	for (int sy=0;sy<NUMT;sy++)
	  for (int sz=0;sz<NUM;sz++)
	    gux[sx+1][sy+1][sz+1].d=FT3darray[2*((sx)*NUMT*NUM+(sy)*NUM+(sz))];
      
      //printf("just for debugging: %e\n",conv/NUMT/NUMT/NUM);

      //prepare gauge transform

      SUNmatrix temp;
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	  {
	    temp.a=gux[sx][sy][sz].b;
	    temp.b=gux[sx][sy][sz].c;
	    temp.c=gux[sx][sy][sz].d;
	    //gux[sx][sy][sz]=mexpi(temp*(-1.0)); // !! Compiler error: ambiguous line, as (double)1.0 could be float or int
	    gux[sx][sy][sz]=mexpi(-1 * temp);  // V1
	    gux[sx][sy][sz].normalize();
	  }

      

      //BCs
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++) { 
	  gux[sx][sy][0] = gux[sx][sy][NUM]; 	gux[sx][sy][NUM+1] = gux[sx][sy][1];
	}
      for (int sx=1;sx<=NUMT;sx++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
	  gux[sx][0][sz] = gux[sx][NUMT][sz]; 	gux[sx][NUMT+1][sz] = gux[sx][1][sz];
	}
      for (int sy=0;sy<=NUMT+1;sy++) 
	for (int sz=0;sz<=NUM+1;sz++) { 
	  gux[0][sy][sz] = gux[NUMT][sy][sz]; 	gux[NUMT+1][sy][sz] = gux[1][sy][sz];
	}

      //do gauge Transform

      for (int sx=1;sx<=NUMT;sx++) 
	for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
	    {
	      ux[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*ux[sx][sy][sz]*gux[sx+1][sy][sz];
	      uy[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*uy[sx][sy][sz]*gux[sx][sy+1][sz];
	      u[sx][sy][sz]=Hermitian(gux[sx][sy][sz])*u[sx][sy][sz]*gux[sx][sy][sz+1];
	      pix[sx][sy][sz]=transportLeft(pix[sx][sy][sz],gux[sx][sy][sz]);
	      piy[sx][sy][sz]=transportLeft(piy[sx][sy][sz],gux[sx][sy][sz]);
	      ez[sx][sy][sz]=transportLeft(ez[sx][sy][sz],gux[sx][sy][sz]);
	      
	      ux[sx][sy][sz].normalize();  //reunitarize
	      uy[sx][sy][sz].normalize();
	      u[sx][sy][sz].normalize();
	    }
      
      enforceBCsOnlyforPU();
      
    } // end of main loop

  //just for debugging:
  //printf("debug: %e\n",conv/NUMT/NUMT/NUM);

  // For Debugging:
	if (MSG_FFT)
	{
		cout << "Converged at "<< gfiter << "iterations, conv = ";
		cout <<conv/NUMT/NUMT/NUM << endl;
	};

  if (conv/NUMT/NUMT/NUM>1.e-15)
    {
      cout << "GF warning: did not converge at "<< maxiter << "iterations, conv = ";
      cout <<conv/NUMT/NUMT/NUM << endl;
    }
  delete [] FT3darray;

  for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) delete [] gux[sx][sy];
  for (int sx=0;sx<NUMT+2;sx++) delete [] gux[sx];
  delete [] gux; 

}


void measureangleforbins(double time, int usebins)
{
  extern void FT3d(double* pArr, int nx, int ny,int nz, int sign);

  if (usebins < 0) return; // Nothing to do

  fstream out,outpz,outdistrib;
  char fname[255];

  char *pdirname = "data";
  if (NUMx == 1) pdirname = "data_mpi";
  
  double* FT3darray = new double[NUMT*NUMT*NUM*2];

  int ptot=0;
  int costheta=0;

  int maxptot = (int)sqrt((double)(NUMT*NUMT*2+NUM*NUM)/9.)+1;
  //number of costheta bins
  //int bins = 10+1;
  int bins = usebins + 1;
  int totbins=maxptot*bins;
  int *countptot = new int[totbins]; //need to count number of degenerate pt's
  double *dndp = new double[totbins]; 
  double *dndpz = new double[NUM+1]; 

  for (int sx=0;sx<totbins;sx++)
    {
      countptot[sx]=0;
      dndp[sx]=0.0;
    }
  for (int sx=0;sx<=NUM;sx++)
    {
      dndpz[sx]=0.0;
    }


  //load momentum
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=pix[sx+1][sy+1][sz+1].a;
	FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
      }
  //do FT
  FT3d(FT3darray,NUMT,NUMT,NUM,1);

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
	    costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));
	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
            if (costheta>bins-1)
              cout << "FAILURE!!!!" << endl;

	    //only this first time
	    countptot[ptot*bins+costheta]++;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
	    
	    //dndpz only
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));
		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	}
  //remaining contributions
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=pix[sx+1][sy+1][sz+1].b;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	  }

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=pix[sx+1][sy+1][sz+1].c;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    if (sz!=0)
	      {
		disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));
		
		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=piy[sx+1][sy+1][sz+1].a;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=piy[sx+1][sy+1][sz+1].b;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;


	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=piy[sx+1][sy+1][sz+1].c;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }


  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=ez[sx+1][sy+1][sz+1].a;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
	    costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }

for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=ez[sx+1][sy+1][sz+1].b;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;

	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }

  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	{
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=ez[sx+1][sy+1][sz+1].c;
	  FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0.0;
	}
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  for (int sx=0;sx<NUMT;sx++) 
    for (int sy=0;sy<NUMT;sy++)
      for (int sz=0;sz<NUM;sz++)
	if (sx+sy+sz!=0)
	  {
	    double disprel=sqrt(2.*(3.-cos(2.*M_PI*sx/NUMT)-cos(2.*M_PI*sy/NUMT)-cos(2*M_PI*sz/NUM)));
	    ptot= (int) floor(disprel*(maxptot-1)/sqrt(12.0));
            costheta= (int)
round((bins-1)*(2*sin(M_PI*sz/NUM)/disprel));

	    if (ptot>maxptot-1)
	      cout << "FAILURE!!!!" << endl;
	    //real part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
	    //imaginary part
	    dndp[ptot*bins+costheta]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
	    

	    //dndpz only
	    if (sz!=0)
	      {
	    disprel=sqrt(2.*(1-cos(2*M_PI*sz/NUM)));

		if (ANGLE_MEASUREMENT_DNDPZ_MEAN)
		{
			// Take mean over all x and y
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]/disprel;
		}
		else
		{
			// Original code only probes at x==y==0
			dndpz[sz]+=FT3darray[2*(sz)]*FT3darray[2*(sz)]/disprel;
			//imaginary part
			dndpz[sz]+=FT3darray[2*(sz)+1]*FT3darray[2*(sz)+1]/disprel;
		}
	      }
	  }



  // write to file
  sprintf(fname,"%s/snapshot/f_%dbins_%.2f.dat",pdirname,usebins,time);
  out.open(fname, ios::out);

  sprintf(fname,"%s/snapshot/fdpz_%.2f.dat",pdirname,time);
  outpz.open(fname, ios::out);

  for (int go=0;go<maxptot;go++)
    {
      out << go << "\t";
	for (int inner=0;inner<bins;inner++)
	{
           if (countptot[go*bins+inner]!=0)
	     out << dndp[go*bins+inner]/countptot[go*bins+inner] << "\t";
           else 
	     out << 0.0 << "\t";
        } 
      out << endl;
    }

  for (int go=1;go<=NUM;go++)
    {
      outpz << go << "\t";
      outpz << dndpz[go]<< endl;
    }

  out.close();
  outpz.close();

	if (time == 0.)
	{
		// The first time only, we also put out the information 
		// about the distribution of events in the bins:

		sprintf(fname,"%s/snapshot/f_%dbins_count.dat",pdirname,usebins);
		out.open(fname, ios::out);

		int checksum = 0;
		for (int go=0;go<maxptot;go++)
		{
			out << go << "\t";
			for (int inner=0;inner<bins;inner++)
			{
				out << countptot[go*bins+inner] << "\t";
				checksum += countptot[go*bins+inner];
			} 
			out << endl;
		}
		assert( checksum == NUMT * NUMT * NUM - 1); // We should record ALL events.
		out.close();
	};

  //clean up
  delete [] dndp;
  delete [] dndpz;
  delete [] countptot;
  delete [] FT3darray;

}

void outputSpectra(double time)
{
	SUNmatrix ***temp;

  if (!can_do_FFT_NR) return;  // safety checked in fft_init() instead of just crashing

  gaugefix3d(150);
    
  // allocate memory for temporary storage variable
	temp = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) temp[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) temp[sx][sy] = new SUNmatrix[NUM+2];
	
	for (int sx=1;sx<=NUMx;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1;sz<=NUM;sz++) { 
        temp[sx][sy][sz] = w[sx][sy][sz][0];
  }
	
  outputSpectrum(temp,"w0",time);
  outputSpectrum(pix,"ex",time);
  outputSpectrum(piy,"ey",time);
  outputSpectrum(ez,"ez",time);

  //cleanup
  delete [] temp;
  
}

double latticeK(int px, int py, int pz) {
	return 2*sqrt(sin((px+1)*M_PI/NUM)*sin((px+1)*M_PI/NUM) + sin((py+1)*M_PI/NUM)*sin((py+1)*M_PI/NUM) + sin((pz+1)*M_PI/NUM)*sin((pz+1)*M_PI/NUM))/A;
}


void outputSpectrum(SUNmatrix*** field, char* fieldName,double time) {

  fstream out;
  char fname[255];
  char *pdirname = "data";		
  double *FT3darray,*power;
  extern void FT3d(double* pArr, int nx, int ny, int nz, int sign);

	FT3darray = new double[NUMT*NUMT*NUM*2+2];
  power = new double[NUMT*NUMT*NUM+1];

  // set all elements to zero to make sure
	for (int i=0;i<NUMT*NUMT*NUM;i++) power[i] = 0;
	    
  int SUNmax = field[1][1][1].maxelement();
	for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
	{	
		// load FT3darray
	  for (int sx=0;sx<NUMT;sx++) 
		  for (int sy=0;sy<NUMT;sy++)
	      for (int sz=0; sz<NUM;sz++) { 
		      // get real part
		      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]=field[sx+1][sy+1][sz+1].element(SUNelement);
		      // set imaginary part to zero
		      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]=0;
	  }

    // take fourier transform
    FT3d(FT3darray,NUMT,NUMT,NUM,1);
    
    // add power to array for averaging over color channels
    for (int sx=0;sx<NUMT;sx++) 
		  for (int sy=0;sy<NUMT;sy++)
	      for (int sz=0; sz<NUM;sz++) { 
		      power[sx*NUMT*NUM+sy*NUM+sz] += FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)];
		      power[sx*NUMT*NUM+sy*NUM+sz] += FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1]*FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1];
	  }
	  
  }
  
  // output color-averaged distributions to file
  sprintf(fname,"%s/snapshot/%s_spectrum_%.2f.dat",pdirname,fieldName,time);
  out.open(fname, ios::out);
	for (int sx=0;sx<NUMT/2;sx++) 
	  for (int sy=0;sy<NUMT/2;sy++)
	    for (int sz=0; sz<NUM/2;sz++) {
	     	  	out << latticeK(sx,sy,sz);
	     	  	out << "\t";
	     	  	out << power[sx*NUMT*NUM+sy*NUM+sz]/(2*SUNmax*NUMT*NUMT*NUM*A*A*A);
	    	  	out << endl;
	}
	out.close();

  delete [] FT3darray;
  delete [] power;
  
	return;
	
}

void measureangle(double time)
{
  if (! can_do_FFT_NR) return;  // safety checked in fft_init() instead of just crashing

  gaugefix3d(150);

  angledumpsave(time); // Save gauge fixed fields for optional later analysis

  measureangleforbins(time, ANGLEBINS1);  // take value from params.txt
  measureangleforbins(time, ANGLEBINS2);  // take value from params.txt
  measureangleforbins(time, ANGLEBINS3);  // take value from params.txt
}

#else  // if not defined SU2

// Sorry, for SU3 and related we can not do this measurement
// so we do nothing.
void gaugefix3d(int maxiter)
{
	// at least we avoid a compiler warning
	int dummy = maxiter;
	dummy++;
}

void measureangleforbins(double time, int usebins)
{
	// at least we avoid a compiler warning
	double dummy = time; dummy++;
	double goofy = usebins; goofy++;
}

void measureangle(double time)
{
	// at least we avoid a compiler warning
	double dummy = time;
	dummy++;
}

#endif

