#include <fstream>
#include <iostream>
#include <string>
#include <stdlib.h>
#include <time.h>

#include "solve.h"

// Some libraries only needed for copyParameterFile()
#include "sunmat.h"
#include "outputangle.h"

#ifndef STANDALONE_SOLVE // just needed for copyParameterFile()
#include "solvempi.h"
#endif

using namespace std;

// this workhorse examines a key to see if corresponds to a var we are setting
// and then attempts to set the var corresponding to key by converting value to the
// appropriate type.  lots of hardcoding here
void setParameter(char *key, char *value) {
	// integer params
	if (strcmp(key,"NUM")==0) {NUM=atoi(value); return;}
	if (strcmp(key,"NUMT")==0) {NUMT=atoi(value); return;}
	if (strcmp(key,"MIMICK1D")==0) {MIMICK1D=atoi(value); return;}
	if (strcmp(key,"STEPS")==0) {STEPS=atoi(value); return;}
	if (strcmp(key,"UPDATE")==0) {UPDATE=atoi(value); return;}
	if (strcmp(key,"SNAPUPDATE")==0) {SNAPUPDATE=atoi(value); return;}
	if (strcmp(key,"ENERGYUPDATE")==0) {ENERGYUPDATE=atoi(value); return;}		
	if (strcmp(key,"ANGLEUPDATE")==0) {ANGLEUPDATE=atoi(value); return;}
	if (strcmp(key,"ANGLEBINS1")==0) {ANGLEBINS1=atoi(value); return;}
	if (strcmp(key,"ANGLEBINS2")==0) {ANGLEBINS2=atoi(value); return;}
	if (strcmp(key,"ANGLEBINS3")==0) {ANGLEBINS3=atoi(value); return;}
	if (strcmp(key,"CHECKPOINTUPDATE")==0) {CHECKPOINTUPDATE=atoi(value); return;}
	if (strcmp(key,"TIMEOUT")==0) {TIMEOUT=atoi(value); return;}
	if (strcmp(key,"STATISTICSUPDATE")==0) {STATISTICSUPDATE=atoi(value); return;}
	if (strcmp(key,"USEPOLYHEDRA")==0) {USEPOLYHEDRA=atoi(value); return;}
	if (strcmp(key,"NW")==0) {NW=atoi(value); return;}
	if (strcmp(key,"NPHI")==0) {NPHI=atoi(value); return;}
	if (strcmp(key,"NZ")==0) {NZ=atoi(value); return;}
	if (strcmp(key,"RANDOMSEED")==0) {RANDOMSEED=atoi(value); return;}
	if (strcmp(key,"PERTURB")==0) {PERTURB=atoi(value); return;}		
	if (strcmp(key,"MIMICKSUN")==0) {MIMICKSUN=atoi(value); return;}
	if (strcmp(key,"COUPLE_CURRENTS")==0) {COUPLE_CURRENTS=atoi(value); return;}
  if (strcmp(key,"FILTER_FUNC")==0) {FILTER_FUNC=atoi(value); return;}
  if (strcmp(key,"INITIAL_CONDITION_TYPE")==0) {INITIAL_CONDITION_TYPE=atoi(value); return;}
	// double/float params
	if (strcmp(key,"A")==0) {A=atof(value); return;}
	if (strcmp(key,"EPS")==0) {EPS=atof(value); return;}
	if (strcmp(key,"XI")==0) {XI=atof(value); return;}
	if (strcmp(key,"MU2")==0) {MU2=atof(value); return;}
	if (strcmp(key,"SIG")==0) {SIG=atof(value); return;}
	if (strcmp(key,"PERTURBSIG")==0) {PERTURBSIG=atof(value); return;}
	if (strcmp(key,"FILTER_WIDTH")==0) {FILTER_WIDTH=atof(value); return;}						
	if (strcmp(key,"MAXEVIOLATION")==0) {MAXEVIOLATION=atof(value); return;}

	// If we reached here, we don't know what that is. This deserves an
	// error message:
	cout << "  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << endl;
	cout << "  <<                                              <<" << endl;
	cout << "      UNKNOWN PARAMETER: " << key << endl;
	cout << "  >>                                              >>" << endl;
	cout << "  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << endl;
	return;
}

//
// This routine assumes that paramters are in text file with
// each parameter on a new line in the format 
//
// PARAMKEY	PARAMVALUE
//
// The PARAMKEY must begin the line and only tabs and spaces
// can appear between the PARAMKEY and PARAMVALUE.
// 
// Lines which begin with 'commentmarker' defined below are ignored
//
void readParameters(char *filename) {
	
	string commentmarker = "//"; 
	char space = ' '; 
	char tab = '\t';

	const int maxline = 128; // maximum line length used in the buffer for reading
	char buffer[maxline];
	ifstream paramFile(filename);
	
	while(!paramFile.eof()) {
		paramFile.getline(buffer,maxline,'\n');
		string line = buffer; int length = strlen(buffer);
		if (line.substr(0,commentmarker.length())!=commentmarker && line.length()>0) {
			char key[64]="",value[64]=""; int founddelim=0;
			for (int i=0;i<length;i++) {
				if (buffer[i]==space || buffer[i]==tab) founddelim=1;
				else {
					if (founddelim==0) key[strlen(key)] = buffer[i];
					else value[strlen(value)] = buffer[i];
				}
			}
			if (strlen(key)>0 && strlen(value)>0) {
				setParameter(key,value);
				cout << key << " = " << value << endl;
			}
		}
	}
	  
  // check nw consistency
	if ( USEPOLYHEDRA!=1 && (NW != NZ*NPHI)) {
	 	cout << "==> Warning: Inconsistent NW,NZ, and NPHI in parameter file." << endl;
	 	cout << "==> Warning: Adjusting so that NW = NZ*NPHI" << endl;
	 	NW = NZ*NPHI;
	 	cout << "NW = " << NW << endl;
	};

	// Set NUMx = NUMT
	NUMx = NUMT;
	
	// Set randomseedused accordingly
	if (RANDOMSEED < 0) 
	{
		RANDOMSEED = (unsigned int)time(0);  // timer as seed
	}
	else
	{
		RANDOMSEED = RANDOMSEED;
	};
	
	return;
}

// Copy the params.txt file into the 'data' or 'data_mpi' directory,
// adding some compiler specifications
void copyParameterFile(char *filename)
{
	fstream out;
	char fname[255];

	const int maxline = 255; // maximum line length used in the buffer for reading
	char buffer[maxline];

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

  	sprintf(fname,"%s/%s",pdirname,filename);

	out.open(fname, ios::out);

	// Add some comments at the top
	out << "////////////////////////////////////////////////////" << endl;
	out << "// Copy of " << filename << endl;
	time_t t = time(0);
	out << "// Date: " << asctime(localtime(&t));
	out << "// SU(N): " << SUNmatrix_string << endl;
	out << "// FFT: " << FFT_string << endl;
	out << "// ---------------------------------------------- //" << endl;
#ifdef STANDALONE_SOLVE
	out << "// Standalone: yes" << endl;
#else
	if (NUMx!=1) out << "// Standalone: no" << endl;
	out << "// MPI: " << (char *)((NUMx==1) ? "yes" : "no") << endl;
	out << "// Number of processors: " << MPIsize << endl;
	out << "// OPTION_DYNAMICLOADBALANCING: " << (char *)(OPTION_DYNAMICLOADBALANCING ? "true" : "false") << endl;
	out << "// OPTION_MEMORY_DROP_OW_OJ: " << (char *)(OPTION_MEMORY_DROP_OW_OJ ? "true" : "false") << endl;
#endif
	out << "// ---------------------------------------------- //" << endl;
	out << "// ACCURATESUMMATION: " << (char *)(ACCURATESUMMATION ? "true" : "false") << endl;
	out << "// ANGLE_MEASUREMENT_DNDPZ_MEAN: " << (char *)(ANGLE_MEASUREMENT_DNDPZ_MEAN ? "true" : "false") << endl;
	out << "// RANDOMSEEDUSED: " << RANDOMSEEDUSED << endl;
	out << "////////////////////////////////////////////////////" << endl;
	out << endl;
	// Now copy the original params.txt file in there:
	ifstream paramFile(filename);
	while(!paramFile.eof()) 
	{
		paramFile.getline(buffer,maxline,'\n');
		out << buffer << endl;
	};

	out.close();

};
