/* Solve MPI */

/* 
MPI extension to solve.cpp

This program complies with MPI-1.1 Standard
http://www.mpi-forum.org/docs/docs.html

Online reference for MPI commands:
http://www-unix.mcs.anl.gov/mpi/mpi-standard/mpi-report-1.1/mpi-report.htm#Node0

Original MPI implementation: Andreas Ipp
enhances solve.cpp by Paul Romatschke, Mike Strickland, and Anton Rebhan

In case of questions or bug reports don't hesitate to contact:
ipp@ect.it or andreas.ipp@gmx.at

!!Please do not distribute without permission of the author(s)!!

*/

#include <mpi.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
//#include <time.h>

using namespace std; // used by mpicore.h

#include "sunmat.h"
#include "mpicore.h"
#include "crossplatform.h"
#include "solve.h"
#include "outputroutines.h"
#include "outputangle.h"
#include "solvempi.h"

// If everything works, you may uncomment the following line
// to increase speed slighly:
// #define  NDEBUG
#include <assert.h>

/* Local functions */
static void manager(void);
static void worker(void);

// MPI variables
int myrank; // rank in MPI world (which CPU am I?)
static int namelen;
static char processor_name[MPI_MAX_PROCESSOR_NAME];
int MPIsize;

static int leftrank; // left neighbor
static int rightrank; // right neighbor


/////////////////////////////////////////////
// Buffer space for MPI
// (faster data transfer possible, if memory is allocated contiguously)
// [Note: in multi-threaded application, also float and double
//        buffers should be doubled into send and receive buffers]
static int lenBufferfloat= 0;
static int maxBufferfloat = 0;
static float *pBufferfloat = NULL;
static int lenBufferdouble;
static int maxBufferdouble;
static double *pBufferdouble;
static int lenBufferAllsend;
static int maxBufferAllsend;
static char *pBufferAllsend;
static int lenBufferAllreceive;
static int maxBufferAllreceive;
static char *pBufferAllreceive;

/*
void wait ( int seconds )
{
  clock_t endwait;
  endwait = clock () + seconds * CLOCKS_PER_SEC ;
  while (clock() < endwait) {}
};
*/


// Helper functions
// always return positive modulus
// ! We assume that b is positive !
int positivemod(int a, int b)
{
	int res;
	res = a % b;
	if (res < 0) res += b;
	return res;
}

/////////////////////////////////////////////////////////////////////
//
//  Class SimulationSlice
//
//

// Contains all state information about one
// slice (x-direction) of the simulation.
class SimulationSlice
{
private:
public:
	int varselection; // Which of the fields below are allocated

	// these hold the current values; note w[v] = a[v] wo[v] + b[v] wz[v]  
	SUNmatrix ***pix,***piy,***ez,****w;
	SUNLieMatrix ***u,***ux,***uy;

	// we don't need the following Upper case letters. 
	// Information is always contained
	// in the lower case version. But we can link those 
	// either with the global lower case letters,
	// or with the global Upper case letters. 
	// (see linkGlobalLower/UpperCase() below)

	// these hold the updated values
//	SUNmatrix ***Pix,***Piy,***Ez,****W;
//	SUNLieMatrix ***U,***Ux,***Uy;

	// these hold the old values of W used for making it's 
	// time derivative symmetric in t
	SUNmatrix ****ow;

	// j holds current at t; oj holds current at t-eps
	SUNmatrix ****j,****oj;

	//int counter_w;
	//int *pcounter_w; // Pointer to a counter for allocation of w-fields.
	//int *pcounter_ow;
	//int counter_j;
	//int *pcounter_j;
	//int *pcounter_oj;

	// stores the running integrated value of the HL contribution to the energy
	double HLenergy;  // change float to double for accuracy.

private:
	void packSUNmatrixarray(SUNmatrix *p, int max);
	void unpackSUNmatrixarray(SUNmatrix *p, int max);
	void packSUNLieMatrixarray(SUNLieMatrix *p, int max);
	void unpackSUNLieMatrixarray(SUNLieMatrix *p, int max);

public:
	SimulationSlice();
	~SimulationSlice();

    void AllocateMemory(int newvarselection = ALL_FIELDS);
	void FreeMemory(int newvarselection = ALL_FIELDS);
	void ReallocateMemory(int newvarselection = ALL_FIELDS);

	void linkGlobalLowerCase(int x, int offset = 0);
	void linkGlobalUpperCase(int x, int offset = 0);

	// we can also pack parts of it, after each calculation step
	int MPIPack(int part=ALL_PARTS); //void *buffer, int bufferlen);
    int MPIUnpack(int part=ALL_PARTS); //void *buffer, int bufferlen);

	// Rearrange pointers between two different SimulationSlices
	void swapzplane(SimulationSlice *ss, int thisy, int ssy);

	// Copy a subset from one slice to another
	void copyfrom(SimulationSlice *ss, int newvarselection = ALL_FIELDS);
};

SimulationSlice::SimulationSlice()
{
	//cout << "Create" << endl;
	varselection = NO_FIELDS;
};

SimulationSlice::~SimulationSlice() 
{
	//cout << "Destroy" << endl;
	FreeMemory(); // free used memory.
};

void SimulationSlice::AllocateMemory(int newvarselection)
{
	assert((varselection & newvarselection) == 0); // only allocate those that don't exist yet.

	// Our internal data only represent a slice in x-direction.
	// Therefore sx only takes the value 0.

	// Now build the structure that is compatible with solve.cpp
	if (newvarselection & FIELDS_PU)
	{
		pix = new SUNmatrix**[1]; 
		for (int sx=0;sx<1;sx++) pix[sx] = new SUNmatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) pix[sx][sy] = new SUNmatrix[NUM+2];

		piy = new SUNmatrix**[1]; 
		for (int sx=0;sx<1;sx++) piy[sx] = new SUNmatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) piy[sx][sy] = new SUNmatrix[NUM+2];

		ez = new SUNmatrix**[1]; 
		for (int sx=0;sx<1;sx++) ez[sx] = new SUNmatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) ez[sx][sy] = new SUNmatrix[NUM+2];

		// indices x,y,z
		u = new SUNLieMatrix**[1]; 
		for (int sx=0;sx<1;sx++) u[sx] = new SUNLieMatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) u[sx][sy] = new SUNLieMatrix[NUM+2];

		ux = new SUNLieMatrix**[1]; 
		for (int sx=0;sx<1;sx++) ux[sx] = new SUNLieMatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) ux[sx][sy] = new SUNLieMatrix[NUM+2];

		uy = new SUNLieMatrix**[1]; 
		for (int sx=0;sx<1;sx++) uy[sx] = new SUNLieMatrix*[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) uy[sx][sy] = new SUNLieMatrix[NUM+2];
	};

	if (newvarselection & FIELDS_WJ)
	{        
		// indices x,y,z,W-index
		w = new SUNmatrix***[1]; 
		for (int sx=0;sx<1;sx++) w[sx] = new SUNmatrix**[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) w[sx][sy] = new SUNmatrix*[NUM+2];
		for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) w[i][j][k] = new SUNmatrix[NW];

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			// indices x,y,z,W-index
			ow = new SUNmatrix***[1]; 
			for (int sx=0;sx<1;sx++) ow[sx] = new SUNmatrix**[NUMT+2];
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) ow[sx][sy] = new SUNmatrix*[NUM+2];
			for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) ow[i][j][k] = new SUNmatrix[NW];
		};
			
		// indices x,y,z,(x,y,z)
		j = new SUNmatrix***[1]; 
		for (int sx=0;sx<1;sx++) j[sx] = new SUNmatrix**[NUMT+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) j[sx][sy] = new SUNmatrix*[NUM+2];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) j[sx][sy][sz] = new SUNmatrix[3];

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			oj = new SUNmatrix***[1]; 
			for (int sx=0;sx<1;sx++) oj[sx] = new SUNmatrix**[NUMT+2];
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) oj[sx][sy] = new SUNmatrix*[NUM+2];
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) oj[sx][sy][sz] = new SUNmatrix[3];
		};
	};

	varselection |= newvarselection;
};

// Specify those bits that you want to be deleted
void SimulationSlice::FreeMemory(int newvarselection)
{
	//cout << "Freememory" << endl;

	assert((((~varselection) & newvarselection) == 0) 
		|| (newvarselection == ALL_FIELDS)); 
	  // only free those that exist (or ALL_FIELDS meaning all that existed).

	// Destruct specified elements
	int nvs = varselection & newvarselection; // only those bits that existed

	if (nvs & FIELDS_PU)
	{
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] pix[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] pix[sx];
		delete[] pix; 

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] piy[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] piy[sx];
		delete[] piy;

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] ez[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] ez[sx];
		delete[] ez;

		// indices x,y,z
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] u[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] u[sx];
		delete[] u;

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] ux[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] ux[sx];
		delete[] ux;

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] uy[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] uy[sx];
		delete[] uy;
	};

	if (nvs & FIELDS_WJ)
	{        
		// indices x,y,z,W-index
		for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) delete[] w[i][j][k];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] w[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] w[sx];
		delete[] w;

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			// indices x,y,z,W-index
			for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) delete[] ow[i][j][k];
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] ow[sx][sy];
			for (int sx=0;sx<1;sx++) delete[] ow[sx];
			delete[] ow;
		};
			
		// indices x,y,z,(x,y,z)
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) delete[] j[sx][sy][sz];
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] j[sx][sy];
		for (int sx=0;sx<1;sx++) delete[] j[sx];
		delete[] j;

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) delete[] oj[sx][sy][sz];
			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) delete[] oj[sx][sy];
			for (int sx=0;sx<1;sx++) delete[] oj[sx];
			delete[] oj;
		};
	};

	varselection ^= nvs;  // delete those bits (XOR)
};

void SimulationSlice::ReallocateMemory(int newvarselection)
{
	// Free those elements that are not contained in newvarselection
	int nvs = varselection & (~newvarselection);
	FreeMemory(nvs);

	// and allocate those elements that are still missing in varselection
	nvs = (~varselection) & newvarselection;
	AllocateMemory(nvs);
};

// Associate lower case letters in SimulationSlice
// with global lower case letters (for solve.cpp)
// (offset is a QUICKFIX to simulate transposed 
//  simulation slices, i.e. swap sx and sy coordinates.
//  Default: offset = 0)
void SimulationSlice::linkGlobalLowerCase(int x, int offset)
{
	if (varselection & FIELDS_PU)
	{
		::pix[x] = this->pix[0] + offset;
		::piy[x] = this->piy[0] + offset;
		::ez[x] = this->ez[0] + offset;
		::u[x] = this->u[0] + offset;
		::ux[x] = this->ux[0] + offset;
		::uy[x] = this->uy[0] + offset;
	};
	if (varselection & FIELDS_WJ)
	{
		::w[x] = this->w[0] + offset;
		if (OPTION_MEMORY_DROP_OW_OJ)
		{
			// lower case letter (previous time slice) links to j,
			// upper case letter (current time slice) links to oj:
			::j[x] = this->j[0] + offset;
		}
		else
		{
			::ow[x] = this->ow[0] + offset;
			::j[x] = this->j[0] + offset;
			::oj[x] = this->oj[0] + offset;
		};
	};
};

// Associate lower case letters in SimulationSlice
// with global Upper case letters (for solve.cpp)
void SimulationSlice::linkGlobalUpperCase(int x, int offset)
{
	if (varselection & FIELDS_PU)
	{
		::Pix[x] = this->pix[0] + offset;
		::Piy[x] = this->piy[0] + offset;
		::Ez[x] = this->ez[0] + offset;
		::U[x] = this->u[0] + offset;
		::Ux[x] = this->ux[0] + offset;
		::Uy[x] = this->uy[0] + offset;
	};
	if (varselection & FIELDS_WJ)
	{
		::W[x] = this->w[0] + offset;
		if (OPTION_MEMORY_DROP_OW_OJ)
		{
			// link ow field to current w field.
			// !W and ow link to the same:
			// ow is used, and the new value is written to W,
			// which is the same variable. => saves space.
			::ow[x] = this->w[0] + offset;

			// lower case letter (previous time slice) links to j,
			// upper case letter (current time slice) links to oj:
			::oj[x] = this->j[0] + offset;
		}
		else
		{
		};
	};
};

////////////////////////////////////////////////////////////////////////////
// SU(N) dependend code
#ifdef SU2

// Helper function:
void SimulationSlice::packSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferfloat[lenBufferfloat++] = p[sz].a;
		pBufferfloat[lenBufferfloat++] = p[sz].b;
		pBufferfloat[lenBufferfloat++] = p[sz].c;
	}
};

// Helper function:
void SimulationSlice::unpackSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a = pBufferfloat[lenBufferfloat++];
		p[sz].b = pBufferfloat[lenBufferfloat++];
		p[sz].c = pBufferfloat[lenBufferfloat++];
	}
};

void SimulationSlice::packSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = p[sz].a;
		pBufferdouble[lenBufferdouble++] = p[sz].b;
		pBufferdouble[lenBufferdouble++] = p[sz].c;
		pBufferdouble[lenBufferdouble++] = p[sz].d;
	}
};

void SimulationSlice::unpackSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a = pBufferdouble[lenBufferdouble++];
		p[sz].b = pBufferdouble[lenBufferdouble++];
		p[sz].c = pBufferdouble[lenBufferdouble++];
		p[sz].d = pBufferdouble[lenBufferdouble++];
	}
};

////////////////////////////////////////////////////////////////
#elif defined SU2double

// Helper function:
void SimulationSlice::packSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = p[sz].a;
		pBufferdouble[lenBufferdouble++] = p[sz].b;
		pBufferdouble[lenBufferdouble++] = p[sz].c;
	}
};

// Helper function:
void SimulationSlice::unpackSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a = pBufferdouble[lenBufferdouble++];
		p[sz].b = pBufferdouble[lenBufferdouble++];
		p[sz].c = pBufferdouble[lenBufferdouble++];
	}
};

void SimulationSlice::packSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = p[sz].a;
		pBufferdouble[lenBufferdouble++] = p[sz].b;
		pBufferdouble[lenBufferdouble++] = p[sz].c;
		pBufferdouble[lenBufferdouble++] = p[sz].d;
	}
};

void SimulationSlice::unpackSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a = pBufferdouble[lenBufferdouble++];
		p[sz].b = pBufferdouble[lenBufferdouble++];
		p[sz].c = pBufferdouble[lenBufferdouble++];
		p[sz].d = pBufferdouble[lenBufferdouble++];
	}
};

////////////////////////////////////////////////////////////////
#elif defined SU3

// Helper function:
void SimulationSlice::packSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferfloat[lenBufferfloat++] = p[sz].a1;
		pBufferfloat[lenBufferfloat++] = p[sz].a2;
		pBufferfloat[lenBufferfloat++] = p[sz].a3;
		pBufferfloat[lenBufferfloat++] = p[sz].a4;
		pBufferfloat[lenBufferfloat++] = p[sz].a5;
		pBufferfloat[lenBufferfloat++] = p[sz].a6;
		pBufferfloat[lenBufferfloat++] = p[sz].a7;
		pBufferfloat[lenBufferfloat++] = p[sz].a8;
	}
};

// Helper function:
void SimulationSlice::unpackSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a1 = pBufferfloat[lenBufferfloat++];
		p[sz].a2 = pBufferfloat[lenBufferfloat++];
		p[sz].a3 = pBufferfloat[lenBufferfloat++];
		p[sz].a4 = pBufferfloat[lenBufferfloat++];
		p[sz].a5 = pBufferfloat[lenBufferfloat++];
		p[sz].a6 = pBufferfloat[lenBufferfloat++];
		p[sz].a7 = pBufferfloat[lenBufferfloat++];
		p[sz].a8 = pBufferfloat[lenBufferfloat++];
	}
};

void SimulationSlice::packSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = real(p[sz].m11);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m11);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m12);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m12);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m13);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m13);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m21);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m21);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m22);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m22);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m23);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m23);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m31);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m31);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m32);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m32);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m33);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m33);
	}
};

void SimulationSlice::unpackSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].m11 = pBufferdouble[lenBufferdouble++];
		p[sz].m11 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m12 = pBufferdouble[lenBufferdouble++];
		p[sz].m12 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m13 = pBufferdouble[lenBufferdouble++];
		p[sz].m13 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m21 = pBufferdouble[lenBufferdouble++];
		p[sz].m21 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m22 = pBufferdouble[lenBufferdouble++];
		p[sz].m22 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m23 = pBufferdouble[lenBufferdouble++];
		p[sz].m23 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m31 = pBufferdouble[lenBufferdouble++];
		p[sz].m31 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m32 = pBufferdouble[lenBufferdouble++];
		p[sz].m32 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m33 = pBufferdouble[lenBufferdouble++];
		p[sz].m33 += I * pBufferdouble[lenBufferdouble++];
	}
};

////////////////////////////////////////////////////////////////
#elif defined SU3double

// Helper function:
void SimulationSlice::packSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = p[sz].a1;
		pBufferdouble[lenBufferdouble++] = p[sz].a2;
		pBufferdouble[lenBufferdouble++] = p[sz].a3;
		pBufferdouble[lenBufferdouble++] = p[sz].a4;
		pBufferdouble[lenBufferdouble++] = p[sz].a5;
		pBufferdouble[lenBufferdouble++] = p[sz].a6;
		pBufferdouble[lenBufferdouble++] = p[sz].a7;
		pBufferdouble[lenBufferdouble++] = p[sz].a8;
	}
};

// Helper function:
void SimulationSlice::unpackSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].a1 = pBufferdouble[lenBufferdouble++];
		p[sz].a2 = pBufferdouble[lenBufferdouble++];
		p[sz].a3 = pBufferdouble[lenBufferdouble++];
		p[sz].a4 = pBufferdouble[lenBufferdouble++];
		p[sz].a5 = pBufferdouble[lenBufferdouble++];
		p[sz].a6 = pBufferdouble[lenBufferdouble++];
		p[sz].a7 = pBufferdouble[lenBufferdouble++];
		p[sz].a8 = pBufferdouble[lenBufferdouble++];
	}
};

void SimulationSlice::packSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		pBufferdouble[lenBufferdouble++] = real(p[sz].m11);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m11);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m12);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m12);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m13);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m13);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m21);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m21);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m22);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m22);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m23);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m23);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m31);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m31);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m32);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m32);
		pBufferdouble[lenBufferdouble++] = real(p[sz].m33);
		pBufferdouble[lenBufferdouble++] = imag(p[sz].m33);
	}
};

void SimulationSlice::unpackSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		p[sz].m11 = pBufferdouble[lenBufferdouble++];
		p[sz].m11 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m12 = pBufferdouble[lenBufferdouble++];
		p[sz].m12 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m13 = pBufferdouble[lenBufferdouble++];
		p[sz].m13 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m21 = pBufferdouble[lenBufferdouble++];
		p[sz].m21 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m22 = pBufferdouble[lenBufferdouble++];
		p[sz].m22 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m23 = pBufferdouble[lenBufferdouble++];
		p[sz].m23 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m31 = pBufferdouble[lenBufferdouble++];
		p[sz].m31 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m32 = pBufferdouble[lenBufferdouble++];
		p[sz].m32 += I * pBufferdouble[lenBufferdouble++];
		p[sz].m33 = pBufferdouble[lenBufferdouble++];
		p[sz].m33 += I * pBufferdouble[lenBufferdouble++];
	}
};

////////////////////////////////////////////////////////////////
#elif defined SU3tvmet

// Helper function:
void SimulationSlice::packSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		for (int i=0; i<3; i++)
		{
			for (int j=0; j<3; j++)
			{
				pBufferdouble[lenBufferdouble++] = real(p[sz].mat(i,j));
				pBufferdouble[lenBufferdouble++] = imag(p[sz].mat(i,j));
			}
		}
	}
};

// Helper function:
void SimulationSlice::unpackSUNmatrixarray(SUNmatrix *p, int max)
{
	for (int sz=0;sz<max;sz++) 
	{
		for (int i=0; i<3; i++)
		{
			for (int j=0; j<3; j++)
			{
				p[sz].mat(i,j) = pBufferdouble[lenBufferdouble++];
				p[sz].mat(i,j) += I * pBufferdouble[lenBufferdouble++];
			}
		}
	}
};

void SimulationSlice::packSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	packSUNmatrixarray(p, max);
};

void SimulationSlice::unpackSUNLieMatrixarray(SUNLieMatrix *p, int max)
{
	unpackSUNmatrixarray(p, max);
};

#else
#error Please specify SU(N) code here!  // Generate compiler error
#endif // SU(N) choice
////////////////////////////////////////////////////////////////////////////

int SimulationSlice::MPIPack(int part)  //void *buffer, int bufferlen, int *position)
{
	// see packing example at
	// http://www-unix.mcs.anl.gov/mpi/mpi-standard/mpi-report-1.1/node62.htm#Node62

	// Since there is no easy way to pack classes, we first have to
	// copy information from the classes into double and float 
	// buffers.

	// Pack all information into contiguous arrays:
	lenBufferfloat = 0;
	lenBufferdouble = 0;

	if (part & PART1)
	{
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			packSUNmatrixarray(this->pix[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++)
			packSUNmatrixarray(this->piy[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++)
			packSUNmatrixarray(this->ez[sx][sy], NUM+2);
	  
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			packSUNLieMatrixarray(this->u[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			packSUNLieMatrixarray(this->ux[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			packSUNLieMatrixarray(this->uy[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) 
			packSUNmatrixarray(this->j[sx][sy][sz], 3);
	};
	if (part & PART2)
	{
		for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++)
			packSUNmatrixarray(this->w[i][j][k], NW);

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) 
				packSUNmatrixarray(this->ow[i][j][k], NW);

			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) 
				packSUNmatrixarray(this->oj[sx][sy][sz], 3);
		};

		pBufferdouble[lenBufferdouble++] = HLenergy;
	}
	if (part & ~ALL_PARTS)
	{
		cout << myrank << ": WARNING: MPIPack(): Unknown part " << part << endl;
	};

    // Some basic test:
    if (   (lenBufferfloat > maxBufferfloat) 
		|| (lenBufferdouble > maxBufferdouble) )
	{
		cout << myrank << ": WARNING! SimulationSlice::MPIPack(): Buffer size insufficient for part " << part << endl;
		// The buffer size should be calculated correctly 
		// in allocateMemoryMPI()
	}

	// Some basic test:
    if ((part == ALL_PARTS)
		&& ((lenBufferfloat != maxBufferfloat) 
		|| (lenBufferdouble != maxBufferdouble)) )
	{
		cout << myrank << ": WARNING! SimulationSlice::MPIPack(): Buffer size calculated wrongly." << endl;
		// The buffer size should be calculated correctly 
		// in allocateMemoryMPI()
	}


	return 0;
};

int SimulationSlice::MPIUnpack(int part) //void *buffer, int bufferlen)
{
	// Unpack all information into contiguous arrays:
	lenBufferfloat = 0;
	lenBufferdouble = 0;

	if (part & PART1)
	{
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			unpackSUNmatrixarray(this->pix[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++)
			unpackSUNmatrixarray(this->piy[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++)
			unpackSUNmatrixarray(this->ez[sx][sy], NUM+2);
	  
		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			unpackSUNLieMatrixarray(this->u[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			unpackSUNLieMatrixarray(this->ux[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) 
			unpackSUNLieMatrixarray(this->uy[sx][sy], NUM+2);

		for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) 
			unpackSUNmatrixarray(this->j[sx][sy][sz], 3);
	};
	if (part & PART2)
	{
		for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++)
			unpackSUNmatrixarray(this->w[i][j][k], NW);

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			for (int i=0;i<1;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) 
				unpackSUNmatrixarray(this->ow[i][j][k], NW);

			for (int sx=0;sx<1;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) 
				unpackSUNmatrixarray(this->oj[sx][sy][sz], 3);
		};

		HLenergy = pBufferdouble[lenBufferdouble++];
	};

	if (part & ~ALL_PARTS)
	{
		cout << myrank << ": WARNING: MPIUnpack(): Unknown part " << part << endl;
	};

	// Some basic test:
    if (   (lenBufferfloat > maxBufferfloat) 
		|| (lenBufferdouble > maxBufferdouble) )
	{
		cout << myrank << ": WARNING! SimulationSlice::MPIUnpack(): Buffer size insufficient." << endl;
		// The buffer size should be calculated correctly 
		// in allocateMemoryMPI()
	}
	// Some basic test:
    if ((part == ALL_PARTS)
		&& ((lenBufferfloat != maxBufferfloat) 
		|| (lenBufferdouble != maxBufferdouble)) )
	{
		cout << myrank << ": WARNING! SimulationSlice::MPIUnpack(): Buffer size calculated wrongly." << endl;
		// The buffer size should be calculated correctly 
		// in allocateMemoryMPI()
	}

	return 0;
};

template <class T>
void swappointers(T **p1, T **p2)
{
	T *tmp;
	tmp = *p1;
	*p1 = *p2;
	*p2 = tmp;
};

// QUICKFIX to reorganize SimulationSlices into a 
// transposed simulation slice (i.e. swap sx <-> sy)
void SimulationSlice::swapzplane(SimulationSlice *ss, int thisy, int ssy)
{
	swappointers( this->pix[0] + thisy, ss->pix[0] + ssy);
	swappointers( this->piy[0] + thisy, ss->piy[0] + ssy);
	swappointers( this->ez[0]  + thisy, ss->ez[0]  + ssy);
	swappointers( this->w[0]   + thisy, ss->w[0]   + ssy);
	swappointers( this->u[0]   + thisy, ss->u[0]   + ssy);
	swappointers( this->ux[0]  + thisy, ss->ux[0]  + ssy);
	swappointers( this->uy[0]  + thisy, ss->uy[0]  + ssy);
	swappointers( this->j[0]   + thisy, ss->j[0]   + ssy);
	if (! OPTION_MEMORY_DROP_OW_OJ)
	{
		swappointers( this->ow[0]  + thisy, ss->ow[0]  + ssy);
		swappointers( this->oj[0]  + thisy, ss->oj[0]  + ssy);
	};
};

void SimulationSlice::copyfrom(SimulationSlice *ss, int newvarselection)
{
	// Since this is a new function, we can be strict in checking:
	assert(((~varselection) & newvarselection)==0);
	assert(((~ss->varselection) & newvarselection)==0);


	// Our internal data only represent a slice in x-direction.
	// Therefore sx only takes the value 0.

	// Now build the structure that is compatible with solve.cpp
	if (newvarselection & FIELDS_PU)
	{
		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					pix[sx][sy][sz] = ss->pix[sx][sy][sz];

		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					piy[sx][sy][sz] = ss->piy[sx][sy][sz];

		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					ez[sx][sy][sz] = ss->ez[sx][sy][sz];

		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					u[sx][sy][sz] = ss->u[sx][sy][sz];

		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					ux[sx][sy][sz] = ss->ux[sx][sy][sz];

		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					uy[sx][sy][sz] = ss->uy[sx][sy][sz];
	};

	if (newvarselection & FIELDS_WJ)
	{        
		// indices x,y,z,W-index
		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					for (int i=0;i<NW;i++)
						w[sx][sy][sz][i] = ss->w[sx][sy][sz][i];

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			// indices x,y,z,W-index
			for (int sx=0;sx<1;sx++) 
				for (int sy=0;sy<NUMT+2;sy++) 
					for (int sz=0; sz<NUM+2; sz++)
						for (int i=0;i<NW;i++)
							ow[sx][sy][sz][i] = ss->ow[sx][sy][sz][i];
		};
			
		// indices x,y,z,(x,y,z)
		for (int sx=0;sx<1;sx++) 
			for (int sy=0;sy<NUMT+2;sy++) 
				for (int sz=0; sz<NUM+2; sz++)
					for (int i=0;i<3;i++)
						j[sx][sy][sz][i] = ss->j[sx][sy][sz][i];

		if (! OPTION_MEMORY_DROP_OW_OJ)
		{
			for (int sx=0;sx<1;sx++) 
				for (int sy=0;sy<NUMT+2;sy++) 
					for (int sz=0; sz<NUM+2; sz++)
						for (int i=0;i<3;i++)
							oj[sx][sy][sz][i] = ss->oj[sx][sy][sz][i];
		};
	};


};

//#if defined MemoryStandard
/////////////////////////////////////////////////////////////////////
// Garbage related functions

// List of allocated slices that are currently not in use:
SimulationSlice **garbagelist; 
int garbagemax; // length of list
int garbagenum; // actual number of garbage

// Obtain allocated Simulation Slice - either fresh or from garbage can.
SimulationSlice *GetNewSimulationSlice(int newvarselection = ALL_FIELDS)
{
	SimulationSlice* ss;

	if (garbagenum > 0)
	{
		if (MSG_GARBAGE) cout << myrank << ": New SimulationSlice from Garbage" 
			<< " (" << garbagenum << "/" << garbagemax << ")" << endl;

		//// old routine (if there are no measureangle)
		//garbagenum--;
		//return garbagelist[garbagenum];

		// Let us first see if there is a SimulationSlice with the same properties
		// that are required:
		int i = 0;
		for (i=garbagenum-1; i>=0; i--)
		{
			if (garbagelist[i]->varselection == newvarselection) break; // exactly what we need
		};
		if (i>=0)
		{
			assert(garbagelist[i]->varselection == newvarselection);

			// Oh, we found some suitable garbage. Let us sell this as
			// fresh simulation space
			if (MSG_GARBAGE) cout << myrank << ": Garbage: exact match" 
				<< " (" << garbagelist[i]->varselection << ")" << endl;
			ss = garbagelist[i]; // Remember this
			garbagenum--;
			garbagelist[i] = garbagelist[garbagenum]; // move last element to empty position
			return ss;
		}
		else
		{
			// Oh, we didn't find exactly the required slice,
			// but we can spice up existing garbage a bit:
			if (MSG_GARBAGE) cout << myrank << ": Garbage: reallocate"
				<< " (from " << garbagelist[garbagenum-1]->varselection
				<< " to " << newvarselection << ")" << endl;
			garbagenum--;
			ss = garbagelist[garbagenum];
			ss->ReallocateMemory(newvarselection);
			// NOTE: We give back the last item that has been added
			//       to the garbage list. It has the added advantage
			//       that it might still be in the memory cache.
			return ss;

		};
	}
	else
	{
		if (MSG_GARBAGE) cout << myrank << ": Creating new SimulationSlice" 
				<< " (" << newvarselection << ")"  << endl;

		// Garbage can empty. Let us make fresh simulation space.
		SimulationSlice *pslice = new SimulationSlice();
		pslice->AllocateMemory(newvarselection);
		return pslice;
	}
};

// Get rid of them
// NOTE: Currently there is no method implemented to collect garbage
//       from time to time. Memory can only increase.
void DisposeSimulationSlice(SimulationSlice *yuck)
{
	// nothing to throw away?
	if (yuck == NULL) return;

	// Garbage can full?
	if (garbagenum >= garbagemax)
	{
		cout << myrank << ": WARNING: YUCK! DisposeSimulationSlice(): Garbage can full!" << endl;
		yuck->FreeMemory();
		delete yuck; // bye bye..
	}
	else
	{
		if (MSG_GARBAGE) cout << myrank << ": Disposing SimulationSlice"
			<< " (" << garbagenum << "/" << garbagemax << ")" << endl;

		// Don't wanna touch waste. Away with it:
		garbagelist[garbagenum]=yuck;
		garbagenum++;
	};
	return;
};
/////////////////////////////////////////////////////////////////////
//#endif // MemoryStandard

/////////////////////////////////////////////////////////////////////
//
//  Class SliceInfo
//
//
// Information about one slice
class SliceInfo 
{
public:
	int time; // Current time
	  // Part: 0.. nothing done; 1..part1 done; 2..part2 done.
	int currentpart;
	int previouspart;
	SimulationSlice *previousslice; // Holds pointer to previous slice
	SimulationSlice *currentslice; // Holds pointer to current slice
	int currentblock; // Counter for block (0..not blocked)
	int previousblock;
	  // Block means a sending operation is pending, so this slice should not
	  // be deleted yet.

	int simulationblock; // if >0, this slice should not be simulated.

	SliceInfo();
	~SliceInfo() {};

	void reset();
/*
	// Code could be written more elegantly with these functions ?

	int exists(int probetime); // does this time exist?
	int part(int probetime); // get part at this time
	SimulationSlice *slice(int probetime);
	int block(int probetime); // get current block value at this time
	int increaseblock(int probetime);
	int decreaseblock(int probetime);
*/
};

SliceInfo::SliceInfo()
{
	currentslice = NULL;
	previousslice = NULL;
	reset();
};

void SliceInfo::reset()
{
	time = 0;
	DisposeSimulationSlice(currentslice);
	currentslice = NULL;
	DisposeSimulationSlice(previousslice);
	previousslice = NULL;
	currentpart = 0;
	previouspart = 0;
	currentblock = 0;
	previousblock = 0;
	simulationblock = 0;
};
/*
 // does this time exist?
int SliceInfo::exists(int probetime)
{
};

// get part at this time
int SliceInfo::part(int probetime)
{
};

SimulationSlice *SliceInfo::slice(int probetime)
{
};

 // get current block value at this time
int SliceInfo::block(int probetime)
{
};

int SliceInfo::increaseblock(int probetime)
{
};

int SliceInfo::decreaseblock(int probetime)
{
};
*/

/////////////////////////////////////////////////////////////////////
// Struct for send and receive buffers
//
// (This struct is also used in MPI_Send, TAG_WANTTOSEND,
//  and assumes 4 INT, starting at 'pos')
struct SendInfo
{
	int pos;  // Position of Slice
	int node; // Node we want to send it to (or receive from)
	int time; // Time of that slice
	int part; // which part to send

	void dump() { 
		cout << myrank << ": SendInfo dump: "
			<< "pos: " << pos
			<< "node: " << node
			<< "time: " << time
			<< "part: " << part
			<< endl;
	};

	void reset()
	{
		pos = 0;
		node = 0;
		time = 0;
		part = 0;
	};
};


/////////////////////////////////////////////////////////////////////
// FIFOBuffer
// (simple First-In-First-Out Buffer)
// ADDITIONAL IMPLEMENTATIONS:

// special class:
template <>
void FIFOBuffer<SliceInfo>::dumpstate()
{
	cout << myrank << ": FIFO state SliceInfo: ";
	int i = first;
	while (i != last)
	{
		cout << (pbuf+i)->time 
			<< (char)((pbuf+i)->currentpart) + 'a'  
			<< (char)((pbuf+i)->previouspart) + 'a' << ", ";
		i++;
		if (i>=maxlen) i = 0;
	};
	cout << endl;
	return;
};

template <>
void FIFOBuffer<SendInfo>::dumpstate()
{
	cout << myrank << ": FIFO state SendInfo: ";
	int i = first;
	while (i != last)
	{
		cout << " [" << "pos " << ((pbuf+i)->pos)
			//<< (char)((pbuf+i)->part) + 'a' 
			<< ", part " << ((pbuf+i)->part)
			<< ", time " << ((pbuf+i)->time)
			<< ", to " << ((pbuf+i)->node)
			<< "] ";
		i++;
		if (i>=maxlen) i = 0;
	};
	cout << endl;
	return;
};

/////////////////////////////////////////////////////////////////////
// Output Measurements
// (internally handle variables as double, in order not to loose
//  additional accuracy when forming sums)

class outputVariables
{
public:
	int pos;
	int time;
	int count; // used by manager only for data collection

	double cjxybar;
	double cjxzbar;
	double cjyzbar;
	
	double etenergy;
	double ezenergy;
	double bzenergy;
	double btenergy;
	double hlenergy;

	double energy;
	double energyDiff; // also save locally for easy transmission
	double gausslaw;

	double initialed;

#if defined SU2
	float gausslawtempa;
	float gausslawtempb;
	float gausslawtempc;
#elif defined SU2double
	double gausslawtempa;
	double gausslawtempb;
	double gausslawtempc;
#elif defined SU3
	float gausslawtemp[SUNmatrix_NUMBER_OF_FLOATS];
#elif defined SU3double
	double gausslawtemp[SUNmatrix_NUMBER_OF_DOUBLES];
#elif defined SU3tvmet
	double gausslawtemp[SUNmatrix_NUMBER_OF_DOUBLES];
#else
#error Please specify SU(N) code here.
#endif

	outputVariables();
	void reset();
	void add(outputVariables ov);
};

outputVariables::outputVariables()
{
	reset();
};

void outputVariables::reset()
{
	pos = 0;
	time = 0;
	count = 0;

	cjxybar = 0.;
	cjxzbar = 0.;
	cjyzbar = 0.;
	
	etenergy = 0.;
	ezenergy = 0.;
	bzenergy = 0.;
	btenergy = 0.;
	hlenergy = 0.;
	
	energy = 0.;
	energyDiff = 0.;
	gausslaw = 0.;

	initialed = 0.;

#if defined SU2 || defined SU2double
	gausslawtempa = 0.;
	gausslawtempb = 0.;
	gausslawtempc = 0.;
#elif defined SU3
	for (int i=0; i<SUNmatrix_NUMBER_OF_FLOATS; i++)
		gausslawtemp[i] = 0.;
#elif defined SU3double
	for (int i=0; i<SUNmatrix_NUMBER_OF_DOUBLES; i++)
		gausslawtemp[i] = 0.;
#elif defined SU3tvmet
	for (int i=0; i<SUNmatrix_NUMBER_OF_DOUBLES; i++)
		gausslawtemp[i] = 0.;
#else
#error Please specify SU(N) code here.
#endif

};

void outputVariables::add(outputVariables ov)
{
	count++; // Increment counter

	cjxybar += ov.cjxybar;
	cjxzbar += ov.cjxzbar;
	cjyzbar += ov.cjyzbar;
	
	etenergy += ov.etenergy;
	ezenergy += ov.ezenergy;
	bzenergy += ov.bzenergy;
	btenergy += ov.btenergy;
	hlenergy += ov.hlenergy;
	
	energy += ov.energy;
	energyDiff += ov.energyDiff;
	gausslaw += ov.gausslaw;

	initialed += ov.initialed;

#if defined SU2 || defined SU2double
	gausslawtempa += ov.gausslawtempa;
	gausslawtempb += ov.gausslawtempb;
	gausslawtempc += ov.gausslawtempc;
#elif defined SU3
	for (int i=0; i<SUNmatrix_NUMBER_OF_FLOATS; i++)
		gausslawtemp[i] += ov.gausslawtemp[i];
#elif defined SU3double
	for (int i=0; i<SUNmatrix_NUMBER_OF_DOUBLES; i++)
		gausslawtemp[i] += ov.gausslawtemp[i];
#elif defined SU3tvmet
	for (int i=0; i<SUNmatrix_NUMBER_OF_DOUBLES; i++)
		gausslawtemp[i] += ov.gausslawtemp[i];
#else
#error Please specify SU(N) code here.
#endif
	
};

static MPI_Datatype myMPI_outputVariables;

static void
commitoutputVariablesMPIstruct(void)
{
  /* Create the MPI types from our structs */
  /* from: Snir et al: MPI: Complete Reference, Vol I, p 153 */
  
  // Example following http://www-unix.mcs.anl.gov/mpi/mpi-standard/mpi-report-1.1/node61.htm
  
	// Define space for 1 SUNmatrix (gausslawtemp)
	int maxfloat = SUNmatrix_NUMBER_OF_FLOATS;
	int maxdouble = SUNmatrix_NUMBER_OF_DOUBLES;

//#if defined SU2
//	int maxfloat = 3;
//	int maxdouble = 0;
//#elif defined SU2double
//	int maxfloat = 0;
//	int maxdouble = 3;
//#elif defined SU3
//	int maxfloat = SUNmatrix_NUMBER_OF_FLOATS;
//	int maxdouble = 0;
//#elif defined SU3double
//	int maxfloat = 0;
//	int maxdouble = SUNmatrix_NUMBER_OF_DOUBLES;
//#elif defined SU3tvmet
//	int maxfloat = 0;
//	int maxdouble = SUNmatrix_NUMBER_OF_DOUBLES;
//#else
//#error Please specify SU(N) code here.
//#endif

  outputVariables ov;
  const int maxtype = 4;
  MPI_Datatype type[maxtype] = {MPI_INT, MPI_DOUBLE, MPI_FLOAT, MPI_DOUBLE};
  int          blocklen[maxtype] = {3, 12, maxfloat, maxdouble}; 
  MPI_Aint     disp[maxtype]; 
  MPI_Aint     base; 

  // compute displacements of structure components
  MPI_Address( &ov.pos, disp); 
  MPI_Address( &ov.cjxybar, disp+1); 
#if defined SU2
  MPI_Address( &ov.gausslawtempa, disp+2); 
  MPI_Address( &ov.gausslawtempa, disp+3); // not needed
#elif defined SU2double
  MPI_Address( &ov.gausslawtempa, disp+2); // zero length
  MPI_Address( &ov.gausslawtempa, disp+3);
#elif defined SU3
  MPI_Address( &(ov.gausslawtemp[0]), disp+2);
  MPI_Address( &(ov.gausslawtemp[0]), disp+3); // not needed 
#elif defined SU3double
  MPI_Address( &(ov.gausslawtemp[0]), disp+2); // zero length
  MPI_Address( &(ov.gausslawtemp[0]), disp+3); 
#elif defined SU3tvmet
  MPI_Address( &(ov.gausslawtemp[0]), disp+2); // zero length
  MPI_Address( &(ov.gausslawtemp[0]), disp+3); 
  // MPI_Address( ov.gausslawtemp, disp+2); // this would be equivalent.
#else
#error Please specify SU(N) code here.
#endif

  base = disp[0]; 
  for (int i=0; i < maxtype; i++) disp[i] -= base; 
 
  MPI_Type_struct( maxtype, blocklen, disp, type, &myMPI_outputVariables); 

  MPI_Type_commit(&myMPI_outputVariables);   // commit the type
}


outputVariables outputMeasurementsMPIworker(double time) {
	outputVariables ov;

	ov.cjxybar = newCjbar();
	ov.cjxzbar = Cjxzbar();
	ov.cjyzbar = Cjyzbar();
	
	ov.etenergy = etEnergyDensity();
	ov.ezenergy = ezEnergyDensity();
	ov.bzenergy = bzEnergyDensity();
	ov.btenergy = btEnergyDensity();
	ov.hlenergy = HLenergy/NUM/NUMT/NUMT/A/A/A/A;

	ov.energy = 0;
	ov.energy += ov.hlenergy;
	ov.energy += ov.ezenergy;
	ov.energy += ov.etenergy;
	ov.energy += ov.bzenergy;
	ov.energy += ov.btenergy;
	
	//energyDiff = (energy-InitialEnergy)/InitialEnergy;
	ov.energyDiff=ov.energy;
	::energyDiff = ov.energyDiff; // also save in global variable ?

	//ov.gausslaw = checkGaussLaw();
	SUNmatrix temp;
	temp = checkGaussLawPart1();
#if defined SU2 || defined SU2double
	ov.gausslawtempa = temp.a;
	ov.gausslawtempb = temp.b;
	ov.gausslawtempc = temp.c;
#elif defined SU3 || defined SU3double
	int k=0;
	ov.gausslawtemp[k++] = temp.a1;
	ov.gausslawtemp[k++] = temp.a2;
	ov.gausslawtemp[k++] = temp.a3;
	ov.gausslawtemp[k++] = temp.a4;
	ov.gausslawtemp[k++] = temp.a5;
	ov.gausslawtemp[k++] = temp.a6;
	ov.gausslawtemp[k++] = temp.a7;
	ov.gausslawtemp[k++] = temp.a8;
#elif defined SU3tvmet
	int k=0;
	for (int i=0; i<3; i++)
	{
		for (int j=0; j<3; j++)
		{
			ov.gausslawtemp[k++] = real(temp.mat(i, j));
			ov.gausslawtemp[k++] = imag(temp.mat(i, j));
		}
	};
#else
#error Please specify SU(N) code here.
#endif
	ov.gausslaw = 0.;

	// only first time:
	if (time == 0.)
	{
		ov.initialed = initialED();
	};

	return ov;
};

void outputMeasurementsMPImanager(double time, outputVariables ov) 
{
	// Update timing information
	timeval newtime;
	double outputseconds, averageseconds;
	outputcounter++;
	gettimeofday(&newtime, 0);
	outputseconds = ((double)(newtime.tv_sec - outputtime.tv_sec)) 
	  + ((double)(newtime.tv_usec - outputtime.tv_usec) * 0.000001);
	averageseconds = (((double)(newtime.tv_sec - starttime.tv_sec)) 
	  + ((double)(newtime.tv_usec - starttime.tv_usec) * 0.000001)) / outputcounter;
        outputtime = newtime;

	// output to screen

	cout.precision(5);
	int dwidth = 13;
	cout.width(dwidth); cout << time;
	cout.width(dwidth); cout << ov.ezenergy;
	cout.width(dwidth); cout << ov.etenergy;
	cout.width(dwidth); cout << ov.bzenergy;
	cout.width(dwidth); cout << ov.btenergy;
	cout.width(dwidth); cout << ov.hlenergy;
	cout.width(dwidth); cout << ov.gausslaw;
	cout.width(dwidth); cout << ov.energyDiff;
	cout.width(dwidth); cout << outputseconds;
	cout.width(dwidth); cout << averageseconds;
	cout << endl;

	// output to files

	cbar_out << time;
	cbar_out << "\t" << log10(ov.cjxybar);
	cbar_out << "\t" << log10(ov.cjxzbar);
	cbar_out << "\t" << log10(ov.cjyzbar) << endl;

	gausslaw_out << time << "\t" << log10(ov.gausslaw) << endl; 

	energy_out << time;
	energy_out << "\t" << fabs(ov.energy);
	energy_out << "\t" << ov.ezenergy;
	energy_out << "\t" << ov.etenergy;
	energy_out << "\t" << ov.bzenergy;
	energy_out << "\t" << ov.btenergy;
	energy_out << "\t" << fabs(ov.hlenergy);
	energy_out << endl;
}

void outputMeasurementsMPImanagerFirstTime(double time, outputVariables ov) 
{
	// Avoid compiler warning about unused variable
	double dummy = time; dummy++;
	
	//double phirms = btEnergyDensity();
	double phirms = ov.btenergy;
	cout << "Initial Bt RMS: " << phirms << endl;

	//double phizrms = bzEnergyDensity();
	double phizrms = ov.bzenergy;
	cout << "Initial Bz RMS:" << phizrms << endl;

	//double etrms = etEnergyDensity()*4;
	double etrms = ov.etenergy*4;
	cout << "Initial et RMS: " << etrms << endl;

	//double ezrms = ezEnergyDensity()*4;
	double ezrms = ov.ezenergy*4;
	cout << "Initial ez RMS: " << ezrms << endl;

	//InitialEnergy = initialED();
	InitialEnergy = ov.initialed;
	cout << "Initial Energy Density: " << InitialEnergy << endl;

	//double gauss = checkGaussLaw();
	double gauss = ov.gausslaw;
	cout << "Initial Gauss Law Check: " << gauss << endl;

	printDivider();

	int dwidth = 13;
	cout.width(dwidth); cout << "Time";
	cout.width(dwidth); cout << "<Ez>";
	cout.width(dwidth); cout << "<Et>";
	cout.width(dwidth); cout << "<Bz>";
	cout.width(dwidth); cout << "<Bt>";
	cout.width(dwidth); cout << "<HL>";
	cout.width(dwidth); cout << "GaussCheck";
	cout.width(dwidth); cout << "EnergyViol";
	cout.width(dwidth); cout << "Seconds";
	cout.width(dwidth); cout << "Sec. (av)";
	cout << endl;
	
	printDivider();
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// statisticsVariables
// Gather Timing and other information for profiling the code.

// Define the timers
const int MAXTIMER = 12;

const int TIMER_TOTAL = 0;
const int TIMER_IDLE = 1;
const int TIMER_PART1 = 2;
const int TIMER_PART2 = 3;
const int TIMER_SEND = 4;
const int TIMER_RECEIVE = 5;
const int TIMER_PACK = 6;
const int TIMER_UNPACK = 7;
const int TIMER_WHICHWORK = 8;
const int TIMER_BLOCK_MESSAGES = 9; // Timing of the whole block of things
const int TIMER_BLOCK_SENDRECEIVE = 10;
const int TIMER_BLOCK_WORK = 11;

const char *TIMERNAME[MAXTIMER] =
	{"Total time", "Idle time", "Part 1", "Part 2", 
	 "Send", "Receive", "Pack", "Unpack", "WhichWork",
	 "*Messages", "*Send&Recv", "*Work"}; // Stared quantities refer to a block.

class statisticsVariables
{
public:
	int node;
	int time;
	int countnodes; // used by manager only for data collection (in EvolveMPI())

	// Timers
	long count[MAXTIMER];  // count of events in current section
	long counttotal[MAXTIMER];

	double starttime[MAXTIMER]; // temporarily keeps time
	double checktime[MAXTIMER];

	double minduration[MAXTIMER]; // start to last check
	double maxduration[MAXTIMER]; // start to stop
	double mindurationtotal[MAXTIMER]; // start to last check
	double maxdurationtotal[MAXTIMER]; // start to stop
	
	// Constants

	// Methods:
	statisticsVariables();
	void reset();

	// Timers:
	void start(int id);
	void check(int id);
	void stop(int id);

	void accumulate(); // Add section values to total
	void resetsection(); // Reset section values

	void add(statisticsVariables stats);
};

statisticsVariables::statisticsVariables()
{
	reset();
};

void statisticsVariables::reset()
{	
	node = myrank; // 0; We reset to the current rank.
	time = 0;
	countnodes = 0;

	for (int i=0; i<MAXTIMER; i++)
	{
		count[i] = 0;
		counttotal[i] = 0;
		starttime[i] = -1;
		checktime[i] = -1;
		minduration[i] = 0;
		maxduration[i] = 0;
		mindurationtotal[i] = 0;
		maxdurationtotal[i] = 0;
	};
};

void statisticsVariables::start(int id)
{
	if (starttime[id] != -1 || checktime[id] != -1)
	{
		cout << myrank << ": WARNING: statisticsVariables::start() called, but previous start() had not been stopp()ed." << endl;
	};
	starttime[id] = MPI_Wtime();
	checktime[id] = starttime[id];
};

void statisticsVariables::check(int id)
{
	if (starttime[id] == -1)
	{
		cout << myrank << ": WARNING: statisticsVariables::check() called without prior start()" << endl;
	};
	checktime[id] = MPI_Wtime();
};

void statisticsVariables::stop(int id)
{
	if (starttime[id] == -1 || checktime[id] == -1)
	{
		cout << myrank << ": WARNING: statisticsVariables::stop() called without prior start()" << endl;
	};
	double stoptime = MPI_Wtime();

	// Accumulate:
	count[id]++;
	minduration[id] += checktime[id] - starttime[id];
	maxduration[id] += stoptime - starttime[id];
	starttime[id] = -1;
	checktime[id] = -1;

//	// Output:
//	cout << myrank << ": Timer " << TIMERNAME[id] << ": "
//		<< "count: " << count[id]
//		<< ", min: " << minduration[id]
//		<< ", max: " << maxduration[id] << endl;
};

 // Add section values to total
void statisticsVariables::accumulate()
{
	for (int i=0; i<MAXTIMER; i++)
	{
		counttotal[i] += count[i];
		mindurationtotal[i] += minduration[i] ;
		maxdurationtotal[i] += maxduration[i];
	};
};

 // Reset section values
void statisticsVariables::resetsection()
{
	for (int i=0; i<MAXTIMER; i++)
	{
		count[i] = 0;
		minduration[i] = 0;
		maxduration[i] = 0;
	};
};

void statisticsVariables::add(statisticsVariables stats)
{
	countnodes++; // Increment counter

	for (int i=0; i<MAXTIMER; i++)
	{
		count[i] += stats.count[i];
		counttotal[i] += stats.counttotal[i];
		minduration[i] += stats.minduration[i];
		maxduration[i] += stats.maxduration[i];
		mindurationtotal[i] += stats.mindurationtotal[i];
		maxdurationtotal[i] += stats.maxdurationtotal[i];
	};
};

static MPI_Datatype myMPI_statisticsVariables;

static void
commitstatisticsVariablesMPIstruct(void)
{
  /* Create the MPI types from our structs */
  /* from: Snir et al: MPI: Complete Reference, Vol I, p 153 */
  
  // Example following http://www-unix.mcs.anl.gov/mpi/mpi-standard/mpi-report-1.1/node61.htm
  
	// Define space for 1 SUNmatrix (gausslawtemp)
  statisticsVariables stats;
  const int maxtype = 3;
  MPI_Datatype type[maxtype] = {MPI_INT, MPI_LONG, MPI_DOUBLE};
  int          blocklen[maxtype] = {3, 2 * MAXTIMER, 6 * MAXTIMER}; 
  MPI_Aint     disp[maxtype]; 
  MPI_Aint     base; 

  // compute displacements of structure components
  MPI_Address( &stats.node, disp); 
  MPI_Address( &stats.count[0], disp+1); 
  MPI_Address( &stats.starttime[0], disp+2); 

  base = disp[0]; 
  for (int i=0; i < maxtype; i++) disp[i] -= base; 
 
  MPI_Type_struct( maxtype, blocklen, disp, type, &myMPI_statisticsVariables); 

  MPI_Type_commit(&myMPI_statisticsVariables);   // commit the type
}

// pStats points to a list of statisticVariables,
//    where [0] is the total, and [1] to [MPIsize-1] are the nodes.
void outputStatistics(double time, statisticsVariables *pStats) 
{
    fstream out;
    char filenamestatistics[255];

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	if (MSG_DISK) 
		cout << "0: Writing statistics data. time: " << time << endl;
	
	sprintf(filenamestatistics,"%s/statistics/stats%.2f.txt",pdirname,time);

	// file name defined above
	out.open(filenamestatistics, ios::out);
	out.precision(5);
	int dwidth = 13;
	out.width(dwidth); out << "Time:";
	out.width(dwidth); out << time << endl;
	out << endl;

	out << endl << " =======  TOTAL  =======" << endl << endl;

	out.width(dwidth); out << "";
	out << " | ";
	out.width(dwidth); out << "Section";
	out.width(dwidth); out << "";
	out.width(dwidth); out << "";
	out.width(dwidth); out << "";
	out << " | ";
	out.width(dwidth); out << "Sum";
	out.width(dwidth); out << "";
	out.width(dwidth); out << "";
	out.width(dwidth); out << "";
	out << endl;

	out.width(dwidth); out << "";
	out << " | ";
	out.width(dwidth); out << "Min-Duration";
	out.width(dwidth); out << "Max-Duration";
	out.width(dwidth); out << "Count";
	out.width(dwidth); out << "Max-Mean";
	out << " | ";
	out.width(dwidth); out << "Min-Duration";
	out.width(dwidth); out << "Max-Duration";
	out.width(dwidth); out << "Count";
	out.width(dwidth); out << "Max-Mean";
	out << endl;
	for (int i = 0, n = 0; i<MAXTIMER; i++)
	{
		out.width(dwidth); out << TIMERNAME[i];
		out << " | ";
		out.width(dwidth); out << pStats[n].minduration[i];
		out.width(dwidth); out << pStats[n].maxduration[i];
		out.width(dwidth); out << pStats[n].count[i];
		out.width(dwidth); 
		if (pStats[n].count[i] > 0) out << pStats[n].maxduration[i] / (double) pStats[n].count[i];
		else out << "-";
		out << " | ";
		out.width(dwidth); out << pStats[n].mindurationtotal[i];
		out.width(dwidth); out << pStats[n].maxdurationtotal[i];
		out.width(dwidth); out << pStats[n].counttotal[i];
		out.width(dwidth); 
		if (pStats[n].counttotal[i] > 0) out << pStats[n].maxdurationtotal[i] / (double) pStats[n].counttotal[i];
		else out << "-";
		out << endl;
	};

	out << endl << "(Star '*' Quantities are block timing and sum up to total. Part 1, Part 2, and Idle sum up to Work.)" << endl << endl;

	out << endl << " ======= DETAILS =======" << endl << endl;
/*
	//////////////////////////////////////
	// Arranged in a huge matrix...

	out.width(dwidth); out << "";
	for (int i = 0; i<MAXTIMER; i++)
	{
		out << " | ";
		out.width(dwidth); out << TIMERNAME[i];
		out.width(dwidth); out << "";
		out.width(dwidth); out << "";
		out.width(dwidth); out << "";
	};
	out << endl;

	out.width(dwidth); out << "Node";
	for (int i = 0; i<MAXTIMER; i++)
	{
		out << " | ";
		out.width(dwidth); out << "Min-Duration";
		out.width(dwidth); out << "Max-Duration";
		out.width(dwidth); out << "Count";
		out.width(dwidth); out << "Max-Mean";
	};
	out << endl;

	for (int n = 0; n<MPIsize; n++)
	{
		out.width(dwidth);
		if (n == 0) out << "TOTAL:";
		else out << pStats[n].node;
		for (int i = 0; i<MAXTIMER; i++)
		{
			out << " | ";
			out.width(dwidth); out << pStats[n].minduration[i];
			out.width(dwidth); out << pStats[n].maxduration[i];
			out.width(dwidth); out << pStats[n].count[i];
			out.width(dwidth); 
			if (pStats[n].count[i] > 0) out << pStats[n].maxduration[i] / (double) pStats[n].count[i];
			else out << "-";
		};
		out << endl;
	};
*/
	//////////////////////////////////////
	// Arranged by topic...

	for (int i = 0; i<MAXTIMER; i++)
	{
		out << endl;
		out << "  ---- " << TIMERNAME[i] << " ----" << endl;

		out.width(dwidth); out << "Node";
		out << " | ";
		out.width(dwidth); out << "Min-Duration";
		out.width(dwidth); out << "Max-Duration";
		out.width(dwidth); out << "Count";
		out.width(dwidth); out << "Max-Mean";
		out << endl;

		for (int n = 0; n<MPIsize; n++)
		{
			out.width(dwidth);
			if (n == 0) out << "TOTAL:";
			else out << pStats[n].node;
			out << " | ";
			out.width(dwidth); out << pStats[n].minduration[i];
			out.width(dwidth); out << pStats[n].maxduration[i];
			out.width(dwidth); out << pStats[n].count[i];
			out.width(dwidth); 
			if (pStats[n].count[i] > 0) out << pStats[n].maxduration[i] / (double) pStats[n].count[i];
			else out << "-";
			out << endl;
		};
	};
	out.close();

}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Global variables

SliceInfo *sliceinfo;
int firstslice;  // Position of first slice
int lastslice;  // Position of last slice
// NOTE: firstslice is the "leftmost" slice in a chain.
//       Since we might warp around, firstslice might be 99
//       and lastslice might be 2, if there exist slices 99, 100, 1, 2.

// Global variables for dynamic work loading:
int dynamicloadstateleft = STATE_BM_NOTWAITING; // States for dynamic load balancing
int dynamicloadstateright = STATE_BM_NOTWAITING;
int bordermoveleft = 0;  // Move the border at next time-step
int bordermoveright = 0; // (can be -1, 0, +1)

// Initial variables for data dump
int datadumpinitialtime = 0;

// Global variables for data dump of recorded measurables.
outputVariables *pOVarDataDump; // Log file of all "measureables" for a data dump
int maxOVarDataDump = STEPS / UPDATE + 1; // maximum number of measureables that we need to save.
// pOVarDataDump = new outputVariables[maxOVarDataDump];
int lenOVarDataDump = 0; // Current number of valid dumps.

double startMPItime; // Starting time of MPI

/////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
//
// Data dump save & load

// Dump data currently in the buffer to a file.
void datadumpsave(int *ppos, int *ptime, int *ppart, int state)
{
    fstream out;
    char dirname[255];
    char filename[255];
    char filenamemeasurements[255];
    char filenamemostrecent[255];
	int savetime;
	float timefloat, savetimefloat;

	if (state == DATADUMP_NOTHING) return; // Nothing to do.

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	// ((( for CHECKPOINTUPDATE, 0 and 1 are probed instead of 1 and 2 in worker() because
	// the slice t-1 is sent back to the manager.)))
	int itime = *ptime - 2;  // Definition corresponding to worker()
	if (itime%CHECKPOINTUPDATE==0)
		savetime = *ptime;
	else if (itime%CHECKPOINTUPDATE==1)
		savetime = *ptime-1; // save two consecutive time-slices in same directory.
	else
	{
		// ERROR! There should be no dump at this time!
		cout << "0: datadumpsave(): ERROR: Try to dump pos: " << *ppos << ", time: " << *ptime << " out of sync!" << endl;
		savetime = *ptime; // just in case...
	};

	if (MSG_DISK) 
		cout << "0: Dumping data. pos: " << *ppos << ", time: " << *ptime << endl;
	
	timefloat = (*ptime - 2)*EPS;
	savetimefloat = (savetime - 2)*EPS;

	sprintf(dirname, "%s/datadump/set%.2f",pdirname,savetimefloat);
	sprintf(filename,"%s/datadump/set%.2f/data_t%.2f_p%03d.dmp",pdirname,savetimefloat,timefloat,*ppos);
	sprintf(filenamemeasurements,"%s/datadump/set%.2f/measurements.dmp",pdirname,savetimefloat);
	sprintf(filenamemostrecent,"%s/datadump/mostrecent.txt",pdirname);

	//if (MSG_DISK)
	//	cout << "0: Dumping data to file " << filename
	//	<< ", directory " << dirname
	//	<< endl;

	if (state == DATADUMP_FIRST)
	{
		// Create the directory structure:
		if(mymkdir(dirname, 0711) == -1)
		{
			// Nothing to worry about: Most likely the directory already existed.
			if (MSG_DISK) cout << "Could not create directory '" << dirname << "'." << endl;
		}
		else
			cout << "Directory '" << dirname << "' successfully created." << endl;
	};

	//cout << "sizeof: " << sizeof(lenBufferfloat) << endl;
	//cout << "total: " << sizeof(*ppos) + sizeof(*ptime) + sizeof(*ppart)
	//	+ sizeof(lenBufferfloat) + lenBufferfloat * sizeof(float)
	//	+ sizeof(lenBufferdouble) + lenBufferdouble * sizeof(double) 
	//	<< endl;

	// file name defined above
	out.open(filename, ios::out|ios::binary);
	out.write((char *)ppos, sizeof(*ppos));
	out.write((char *)ptime, sizeof(*ptime));
	out.write((char *)ppart, sizeof(*ppart));
	out.write((char *)&lenBufferfloat, sizeof(lenBufferfloat));
	out.write((char *)pBufferfloat, lenBufferfloat * sizeof(float));
	out.write((char *)&lenBufferdouble, sizeof(lenBufferdouble));
	out.write((char *)pBufferdouble, lenBufferdouble * sizeof(double));
	out.close();

	if (state == DATADUMP_LAST)
	{
		// Dump the measurements
		out.open(filenamemeasurements, ios::out|ios::binary);
		out.write((char *)&lenOVarDataDump, sizeof(lenOVarDataDump));
		out.write((char *)pOVarDataDump, lenOVarDataDump * sizeof(outputVariables));
		out.close();

		// If this was the last of a series, we should also
		// write this to a file.
		// (in the current implementation, this file is changed
		//  after a slice only. So if the dump is incomplete, this
		//  pointer will be incomplete...)
		out.open(filenamemostrecent, ios::out);
		out << setiosflags(ios::fixed) << setprecision(2) << savetimefloat << endl;
		out.close();
	};
};

// Checks whether a data dump exists.
bool datadumpexists(int *ptime)
{
	// Return true if data dump exists, and the time
    ifstream in;
    char filenamemostrecent[255];
	float savetimefloat;
	bool returnval = false; // default.

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	sprintf(filenamemostrecent,"%s/datadump/mostrecent.txt",pdirname);

	in.open(filenamemostrecent);
	if (in.fail() == 1)
	{
		// File does not exist
	}
	else
	{
		if (!in.eof())
		{
			in >> savetimefloat;
			if (savetimefloat > 0)
			{
				*ptime = (int) floor(2.0 + (savetimefloat / EPS) + 0.5);
				cout << "Master: Found old data at time " << savetimefloat 
					<< " (" << *ptime << "). We can continue the work. " << endl;
				returnval = true;
			};
		};
	};
	in.close();

	return returnval;
};

// Load data dump measurements.
void datadumploadmeasurements(int *psavetime, int *ptime)
{
    ifstream in;
    char filenamemeasurements[255];
	int savetime;
	float timefloat, savetimefloat;
	
	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	//int itime = *ptime - 2;  // Definition corresponding to worker()
	savetime = *psavetime;

	timefloat = (*ptime - 2)*EPS;
	savetimefloat = (savetime - 2)*EPS;

	// cout << "obtain " << *psavetime << ", " << savetimefloat << endl;

	sprintf(filenamemeasurements,"%s/datadump/set%.2f/measurements.dmp",pdirname,savetimefloat);

	if (MSG_DISK) 
		cout << "0: Loading measurements from data dump. time: " << *ptime
			 << ", (" << filenamemeasurements << ")" << endl;

	// file name defined above
	in.open(filenamemeasurements, ios::out|ios::binary);
	in.read((char *)&lenOVarDataDump, sizeof(lenOVarDataDump));
	in.read((char *)pOVarDataDump, lenOVarDataDump * sizeof(outputVariables));
	in.close();
};

// Load data dump from file.
void datadumpload(int *ppos, int *psavetime, int *ptime, int *ppart)
{
    ifstream in;
    char filename[255];
	int savetime;
	float timefloat, savetimefloat;
	
	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	// int itime = *ptime - 2;  // Definition corresponding to worker()
	savetime = *psavetime;

	timefloat = (*ptime - 2)*EPS;
	savetimefloat = (savetime - 2)*EPS;

	sprintf(filename,"%s/datadump/set%.2f/data_t%.2f_p%03d.dmp",pdirname,savetimefloat,timefloat,*ppos);

	if (MSG_DISK) 
		cout << "0: Loading data from data dump."
			 << "pos: " << *ppos << ", time: " << *ptime 
			 << ", (" << filename << ")" << endl;

	// file name defined above
	in.open(filename, ios::out|ios::binary);
	in.read((char *)ppos, sizeof(*ppos));
	in.read((char *)ptime, sizeof(*ptime));
	in.read((char *)ppart, sizeof(*ppart));
	in.read((char *)&lenBufferfloat, sizeof(lenBufferfloat));
	in.read((char *)pBufferfloat, lenBufferfloat * sizeof(float));
	in.read((char *)&lenBufferdouble, sizeof(lenBufferdouble));
	in.read((char *)pBufferdouble, lenBufferdouble * sizeof(double));
	in.close();
};

/////////////////////////////////////////////////////////////////////
// Broadcast initial parameters to all workers
void broadcastParametersMPI()
{
	MPI_Comm comm = MPI_COMM_WORLD; 
    int root=0; 

	MPI_Bcast( &NUM, 1, MPI_INT, root, comm); 
	MPI_Bcast( &NUMT, 1, MPI_INT, root, comm); 
	MPI_Bcast( &STEPS, 1, MPI_INT, root, comm); 
	MPI_Bcast( &UPDATE, 1, MPI_INT, root, comm); 
	MPI_Bcast( &SNAPUPDATE, 1, MPI_INT, root, comm); 
	MPI_Bcast( &ANGLEUPDATE, 1, MPI_INT, root, comm); 
	MPI_Bcast( &ANGLEBINS1, 1, MPI_INT, root, comm); 
	MPI_Bcast( &ANGLEBINS2, 1, MPI_INT, root, comm); 
	MPI_Bcast( &ANGLEBINS3, 1, MPI_INT, root, comm); 
	MPI_Bcast( &CHECKPOINTUPDATE, 1, MPI_INT, root, comm); 
	MPI_Bcast( &TIMEOUT, 1, MPI_INT, root, comm); 
	MPI_Bcast( &STATISTICSUPDATE, 1, MPI_INT, root, comm); 
	MPI_Bcast( &MIMICK1D, 1, MPI_INT, root, comm); 
	MPI_Bcast( &USEPOLYHEDRA, 1, MPI_INT, root, comm); 
	MPI_Bcast( &NW, 1, MPI_INT, root, comm); 
	MPI_Bcast( &NPHI, 1, MPI_INT, root, comm); 
	MPI_Bcast( &NZ, 1, MPI_INT, root, comm); 
	MPI_Bcast( &RANDOMSEED, 1, MPI_INT, root, comm); 
	MPI_Bcast( &RANDOMSEEDUSED, 1, MPI_INT, root, comm); 
	MPI_Bcast( &MIMICKSUN, 1, MPI_INT, root, comm); 

	// just in case someone wants to change the type later 
	// and forgets to update here as well...
	assert(sizeof(A) == sizeof(double));

	MPI_Bcast( &A, 1, MPI_DOUBLE, root, comm); 
	MPI_Bcast( &EPS, 1, MPI_DOUBLE, root, comm); 
	MPI_Bcast( &XI, 1, MPI_DOUBLE, root, comm); 
	MPI_Bcast( &MU2, 1, MPI_DOUBLE, root, comm); 
	MPI_Bcast( &SIG, 1, MPI_DOUBLE, root, comm); 
	MPI_Bcast( &MAXEVIOLATION, 1, MPI_DOUBLE, root, comm); 

	MPI_Bcast( &NUMx, 1, MPI_INT, root, comm); 
}

void allocateMemoryMPI()
{
	printDivider();
	if (myrank == 0) 
		cout << "==> Manager: Allocating memory" << endl;
	else
		cout << "==> " << myrank << ": Allocating memory" << endl;

	// Calculate the buffer size needed to contain all float variables for a slice
	// Number of SUNmatrix
	int maxSUNmatrix = 
		(1)*(NUMT+2)*(NUM+2) // x, y, z directions
		* 3 // applies to pix, piy, ez
		+
		(1)*(NUMT+2)*(NUM+2)*NW // x, y, z, NW directions
		* (OPTION_MEMORY_DROP_OW_OJ ? 1 : 2)  // applies to w, (ow)
		+
		(1)*(NUMT+2)*(NUM+2)*3 // x, y, z, 3 directions
		* (OPTION_MEMORY_DROP_OW_OJ ? 1 : 2)  // applies to j, (oj)
		;

	// Number of SUNLieMatrix
	int maxSUNLieMatrix =
		(1)*(NUMT+2)*(NUM+2) // x, y, z directions
		* 3 // applies to u, ux, uy
		;

	maxBufferfloat = 
		maxSUNmatrix * SUNmatrix_NUMBER_OF_FLOATS
		+
		maxSUNLieMatrix * SUNLieMatrix_NUMBER_OF_FLOATS
		;
	lenBufferfloat = 0;
	pBufferfloat = new float[maxBufferfloat];

	maxBufferdouble = 
		maxSUNmatrix * SUNmatrix_NUMBER_OF_DOUBLES
		+
		maxSUNLieMatrix * SUNLieMatrix_NUMBER_OF_DOUBLES
		+
		1 // HLenergy
		;
	lenBufferdouble = 0;
	pBufferdouble = new double[maxBufferdouble];

	// A structure that can hold everything for MPI transfer:
	maxBufferAllsend = 10 * sizeof(int)/sizeof(char) // 10.. extra space
		+ maxBufferfloat * sizeof(float)/sizeof(char)
		+ maxBufferdouble * sizeof(double)/sizeof(char);
	maxBufferAllreceive = maxBufferAllsend; // same size
	lenBufferAllsend = 0;
	lenBufferAllreceive = 0;
	pBufferAllsend = new char[maxBufferAllsend];
	pBufferAllreceive = new char[maxBufferAllreceive];

	if (MSG_INFO) cout << myrank << ": Size of maxBufferAll: " << maxBufferAllsend << endl;

	// Set the x-direction to 1 (as solvempi.cpp will take care of the rest)
	NUMx = 1;

	int myNUMx = NUMx; // Usually we only need all variables at sx=0, 1, and 2.
	if (myrank == 0)
	{
		// For the master only, we need to reserve more space
		// (because of snapshot()).
		// (It doesn't matter so much, as only the pointer table is larger)
		myNUMx = NUMT; // For snapshot(), sx can have sx = 0 .. NUMT+1
	};
	// Allocate only pointers for global variables
	::pix = new SUNmatrix**[myNUMx+2]; 
	::piy = new SUNmatrix**[myNUMx+2]; 
	::ez = new SUNmatrix**[myNUMx+2]; 
	::w = new SUNmatrix***[myNUMx+2]; 
	::u = new SUNLieMatrix**[myNUMx+2]; 
	::ux = new SUNLieMatrix**[myNUMx+2]; 
	::uy = new SUNLieMatrix**[myNUMx+2]; 
	::Pix = new SUNmatrix**[myNUMx+2]; 
	::Piy = new SUNmatrix**[myNUMx+2]; 
	::Ez = new SUNmatrix**[myNUMx+2]; 
	::W = new SUNmatrix***[myNUMx+2]; 
	::U = new SUNLieMatrix**[myNUMx+2]; 
	::Ux = new SUNLieMatrix**[myNUMx+2]; 
	::Uy = new SUNLieMatrix**[myNUMx+2]; 
	::ow = new SUNmatrix***[myNUMx+2]; 
	::j = new SUNmatrix***[myNUMx+2]; 
	::oj = new SUNmatrix***[myNUMx+2]; 

	// Allocate the rest (these are not transfered between nodes):

    // indices nw,(x,y,z)
	v = new double*[NW]; for (int i=0;i<NW;i++) v[i] = new double[3];

    // indices nw
	a = new double[NW];
	b = new double[NW];
		
	// Allocate structures to contain pointers to all information
	sliceinfo = new SliceInfo[NUMT]; // no border padding necessary
	firstslice = -1; // no slice yet
	lastslice = -1;

	garbagemax = 2 * NUMT; // we would never need more than 2*NUMT
	if (myrank == 0) garbagemax = 3 * NUMT; // Master needs more for all slices.
	garbagelist = new SimulationSlice*[garbagemax];
	garbagenum = 0;

	if (myrank == 0)
	{
		// For master only: 
		// Also allocate space for measurements data dumps.
		pOVarDataDump = new outputVariables[maxOVarDataDump];
	};
}

// Packs information from pBufferfloat and pBufferdouble
// into pBufferAllsend and sends it via MPI
//
// position and time specify the location of the slice.
void sendBufferToNode(int node, int position, int time, int part)
{
   /* SENDER CODE */
  const int maxtype = 7;
  int len[maxtype]; 
  MPI_Aint disp[maxtype]; 
  MPI_Datatype type[maxtype], newtype; 
 
  /* build datatype for pBufferfloat[0..lenBufferfloat-1] */ 
 
  MPI_Address( &position, disp); 
  type[0] = MPI_INT; 
  len[0] = 1; 
  MPI_Address( &time, disp+1); 
  type[1] = MPI_INT; 
  len[1] = 1; 
  MPI_Address( &part, disp+2); 
  type[2] = MPI_INT; 
  len[2] = 1; 
  MPI_Address( &lenBufferfloat, disp+3); 
  type[3] = MPI_INT; 
  len[3] = 1; 
  MPI_Address( pBufferfloat, disp+4); 
  type[4] = MPI_FLOAT; 
  len[4] = lenBufferfloat; 
  MPI_Address( &lenBufferdouble, disp+5); 
  type[5] = MPI_INT; 
  len[5] = 1; 
  MPI_Address( pBufferdouble, disp+6); 
  type[6] = MPI_DOUBLE; 
  len[6] = lenBufferdouble;

//  MPI_Aint base = disp[0]; 
//  for (int i=0; i < maxtype; i++) disp[i] -= base; 

  MPI_Type_struct( maxtype, len, disp, type, &newtype); 
  MPI_Type_commit( &newtype); 
 
  /* Pack lenBufferfloat followed by pBufferfloat[0..lenBufferfloat-1]*/ 
 
  lenBufferAllsend = 0; 
  MPI_Pack( MPI_BOTTOM, 1, newtype, pBufferAllsend, maxBufferAllsend, &lenBufferAllsend, MPI_COMM_WORLD); 

  MPI_Type_free( &newtype); 

  // Basic protection:
  if (lenBufferAllsend >= maxBufferAllsend)
  {
	  // pBufferAll size was insufficient
	  cout << myrank << ": WARNING: sendBufferToNode(int node): pBufferAllsend size was insufficient!" << endl;
  }

  /* Send */ 
  MPI_Send( pBufferAllsend, 
	  lenBufferAllsend, // length of actual data
	  MPI_PACKED, 
	  node, // destination process rank
	  TAG_DATA, // message tag
      MPI_COMM_WORLD) ;
};

// Receives information into pBufferfloat and pBufferdouble
// 
// position and time specify the location of the slice within
// the simulation.
void receiveBufferFromNode(int node, int *pposition, int *ptime, int *ppart)
{
	 MPI_Status status; 
 
	/* Receive */

  MPI_Recv( pBufferAllreceive, 
	  maxBufferAllreceive, 
	  MPI_PACKED, 
	  node, // rank of source
	  TAG_DATA, // message tag
	  MPI_COMM_WORLD, // communicator (handle)
	  &status); 
 
  /* Unpack float and double */
  int MPIposition = 0; 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pposition, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, ptime, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, ppart, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, &lenBufferfloat, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pBufferfloat, lenBufferfloat, MPI_FLOAT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, &lenBufferdouble, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pBufferdouble, lenBufferdouble, MPI_DOUBLE, MPI_COMM_WORLD); 
};

// Need static variables for non-blocking sending:
// -> actually not really necessary, as all is packed into BufferAll..
static int isend_position;
static int isend_time;
static int isend_part;
const int isend_maxtype = 7;
static int isend_len[isend_maxtype]; 
static MPI_Aint isend_disp[isend_maxtype]; 
static MPI_Datatype isend_type[isend_maxtype], isend_newtype; 
static int type_commited = false;

// Isend: Immediate Send (non-blocking) 
void IsendBufferToNode(int node, int position, int time, int part, MPI_Request *request)
{
   /* SENDER CODE */ 
 
	// copy local variables to static ones
	isend_position = position;
	isend_time = time;
	isend_part = part;

	// Free the old type if commited before
	if (type_commited)
	{
		  MPI_Type_free( &isend_newtype); 
	}
	else
	{
		type_commited = true; // Starting next time we can free something.
	};

  /* build datatype for pBufferfloat[0..lenBufferfloat-1] */ 
 
  MPI_Address( &isend_position, isend_disp); 
  isend_type[0] = MPI_INT; 
  isend_len[0] = 1; 
  MPI_Address( &isend_time, isend_disp+1); 
  isend_type[1] = MPI_INT; 
  isend_len[1] = 1; 
  MPI_Address( &isend_part, isend_disp+2); 
  isend_type[2] = MPI_INT; 
  isend_len[2] = 1; 
  MPI_Address( &lenBufferfloat, isend_disp+3); 
  isend_type[3] = MPI_INT; 
  isend_len[3] = 1; 
  MPI_Address( pBufferfloat, isend_disp+4); 
  isend_type[4] = MPI_FLOAT; 
  isend_len[4] = lenBufferfloat; 
  MPI_Address( &lenBufferdouble, isend_disp+5); 
  isend_type[5] = MPI_INT; 
  isend_len[5] = 1; 
  MPI_Address( pBufferdouble, isend_disp+6); 
  isend_type[6] = MPI_DOUBLE; 
  isend_len[6] = lenBufferdouble; 

//  MPI_Aint base = isend_disp[0]; 
//  for (int i=0; i < isend_maxtype; i++) isend_disp[i] -= base; 

  MPI_Type_struct( isend_maxtype, isend_len, isend_disp, isend_type, &isend_newtype); 
  MPI_Type_commit( &isend_newtype); 
 
  /* Pack lenBufferfloat followed by pBufferfloat[0..lenBufferfloat-1]*/ 
 
  lenBufferAllsend = 0; 
  MPI_Pack( MPI_BOTTOM, 1, isend_newtype, pBufferAllsend, maxBufferAllsend, &lenBufferAllsend, MPI_COMM_WORLD); 
 
  // Basic protection:
  if (lenBufferAllsend >= maxBufferAllsend)
  {
	  // pBufferAllsend size was insufficient
	  cout << myrank << ": WARNING: sendBufferToNode(int node): pBufferAllsend size was insufficient!" << endl;
  }

  /* Send */ 

  MPI_Isend( pBufferAllsend, 
	  lenBufferAllsend, // length of actual data
	  MPI_PACKED, 
	  node, // destination process rank
	  TAG_DATABLOCK, // message tag
      MPI_COMM_WORLD,
	  request) ; // pointer to communication request
};



// Ireceive: Immediate receive (non-blocking).
// This function should be called after MPI_Test() call, with its status argument.
void IreceiveBufferFromNode(int node, int *pposition, int *ptime, int *ppart, MPI_Status *status)
{
	// Avoid compiler warning about unused variables:
	int dummy = node; dummy++;
	MPI_Status * dummy2 = status; dummy2++;

	// MPI_Status status; 
 
	/* Receive */
/*
  I-recv of this kind should have been called before:
  MPI_Recv( pBufferAllreceive, 
	  maxBufferAllreceive, 
	  MPI_PACKED, 
	  node, // rank of source
	  TAG_DATA, // message tag
	  MPI_COMM_WORLD, // communicator (handle)
	  status); 
	  */
 
  /* Unpack float and double */
  int MPIposition = 0; 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pposition, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, ptime, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, ppart, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, &lenBufferfloat, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pBufferfloat, lenBufferfloat, MPI_FLOAT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, &lenBufferdouble, 1, MPI_INT, MPI_COMM_WORLD); 
  MPI_Unpack(pBufferAllreceive, maxBufferAllreceive, &MPIposition, pBufferdouble, lenBufferdouble, MPI_DOUBLE, MPI_COMM_WORLD); 
};



// enforces boundary conditions by explicitly copying values into the padding elements
// Only for one slice of simulation
void enforceBCsforslice(int sx)
{

  //fill in z=0,NUM+1 values for main x,y
//  for (int sx=1;sx<=NUMx;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 
	pix[sx][sy][0] = pix[sx][sy][NUM]; 	pix[sx][sy][NUM+1] = pix[sx][sy][1];
	piy[sx][sy][0] = piy[sx][sy][NUM]; 	piy[sx][sy][NUM+1] = piy[sx][sy][1];
	ez[sx][sy][0] = ez[sx][sy][NUM]; 	ez[sx][sy][NUM+1] = ez[sx][sy][1];
	u[sx][sy][0] = u[sx][sy][NUM]; 		u[sx][sy][NUM+1] = u[sx][sy][1];
	ux[sx][sy][0] = ux[sx][sy][NUM]; 	ux[sx][sy][NUM+1] = ux[sx][sy][1];
	uy[sx][sy][0] = uy[sx][sy][NUM]; 	uy[sx][sy][NUM+1] = uy[sx][sy][1];
	for (int i=0;i<NW;i++)  {
	  w[sx][sy][0][i] = w[sx][sy][NUM][i]; 	w[sx][sy][NUM+1][i] = w[sx][sy][1][i];
	}
    }
  
  //now fill in y=0,NUMT+1 values for main x and all z
//  for (int sx=1;sx<=NUMx;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      	pix[sx][0][sz] = pix[sx][NUMT][sz]; 	pix[sx][NUMT+1][sz] = pix[sx][1][sz];
	piy[sx][0][sz] = piy[sx][NUMT][sz]; 	piy[sx][NUMT+1][sz] = piy[sx][1][sz];
	ez[sx][0][sz] = ez[sx][NUMT][sz]; 	ez[sx][NUMT+1][sz] = ez[sx][1][sz];
	u[sx][0][sz] = u[sx][NUMT][sz]; 	u[sx][NUMT+1][sz] = u[sx][1][sz];
	ux[sx][0][sz] = ux[sx][NUMT][sz]; 	ux[sx][NUMT+1][sz] = ux[sx][1][sz];
	uy[sx][0][sz] = uy[sx][NUMT][sz]; 	uy[sx][NUMT+1][sz] = uy[sx][1][sz];
	for (int i=0;i<NW;i++)  {
	  w[sx][0][sz][i] = w[sx][NUMT][sz][i]; 	w[sx][NUMT+1][sz][i] = w[sx][1][sz][i];
  	}
    }
  
  //finally fill in x=0,NUMT+1 values for general y,z
/*
  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      pix[0][sy][sz] = pix[NUMx][sy][sz]; 	pix[NUMx+1][sy][sz] = pix[1][sy][sz];
      piy[0][sy][sz] = piy[NUMx][sy][sz]; 	piy[NUMx+1][sy][sz] = piy[1][sy][sz];
      ez[0][sy][sz] = ez[NUMx][sy][sz]; 	ez[NUMx+1][sy][sz] = ez[1][sy][sz];
      u[0][sy][sz] = u[NUMx][sy][sz]; 		u[NUMx+1][sy][sz] = u[1][sy][sz];
      ux[0][sy][sz] = ux[NUMx][sy][sz]; 	ux[NUMx+1][sy][sz] = ux[1][sy][sz];
      uy[0][sy][sz] = uy[NUMx][sy][sz]; 	uy[NUMx+1][sy][sz] = uy[1][sy][sz];
      for (int i=0;i<NW;i++)  {
	w[0][sy][sz][i] = w[NUMx][sy][sz][i]; 	w[NUMx+1][sy][sz][i] = w[1][sy][sz][i];
      }
    }
*/
}

// Version of setInitialConditions() that distributes to other computers.
void setInitialConditionsMPI()
{
	int initialtime = 1; // Initial time for slices
	int finaltime = 1;
	int prevtime = 0;
	int part = ALL_PARTS;
	bool usedump = false;
	outputVariables ov;
	float timefloat;

	usedump = datadumpexists(&initialtime);

	if (usedump)
	{
		// Old data exist.
	    cout << "==> Initializing variables from data dump" << endl;
		cout << "Starting from time " << initialtime << endl;
		finaltime = initialtime + 1; // two time slices.
		prevtime = initialtime - 1;
		datadumpinitialtime = initialtime; // set global initial time, in order to prevent immediate saving of loaded dump.

		for (int i = 1; i < MPIsize; i++)
		{
			// Send new time to all nodes first:
			MPI_Bsend(&prevtime, 1, MPI_INT,
			   i, // dest
			   TAG_SETTIME, // tag
			   MPI_COMM_WORLD);
		};

		// Load old data dumps:
		datadumploadmeasurements(&initialtime, &initialtime);

		cout << "Recovering old measurements:" << endl;

		// Output old data dumps:
		ov = pOVarDataDump[0];
		timefloat = (ov.time-2) * EPS;
		outputMeasurementsMPImanagerFirstTime(timefloat, ov);

		for (int i = 1; i<lenOVarDataDump; i++)
		{
			ov = pOVarDataDump[i];
			if (ov.time < initialtime+2) // adjust properly, so that no measurements is generated twice or omitted.
			{
				timefloat = (ov.time-2) * EPS;
				outputMeasurementsMPImanager(timefloat, ov);
			};
		};

		cout << "Continuing calculation..." << endl;
	}
	else
	{
	    cout << "==> Initializing variables\n";
		datadumpinitialtime = 0;
	};

	//srand((unsigned int)time(0));
	//srand(19691123);
	srand(RANDOMSEEDUSED);

	// Reserve some space:
	SimulationSlice slice;
	slice.AllocateMemory();

	//cout << "M: Allocating\n";

	// associate its memory to pix{0].., piy[0].., etc.
	if (OPTION_MEMORY_DROP_OW_OJ)
	{
		// We first link to UpperCase, so that ow and w will
		// finally link to the same, and accessing ow will not
		// raise an error.
		slice.linkGlobalUpperCase(0);
	};
	slice.linkGlobalLowerCase(0);
	slice.HLenergy = 0.;

	//cout << "M: Setting InitialConditions\n" << endl;

	if (MPIsize > NUMT)
	{
		cout << "Master: WARNING: setInitialConditionsMPI(): There must be at least as many transverse slices as computer nodes (better much more!)" << endl;
	}

	// Split the slices among all nodes.
	// NOTE: pos is the internal position and runs from 0 to NUMT-1
	//       sx is the position in solve.cpp and runs from 1 to NUMT
	//       In all calls to solve.cpp, pos is converted to sx = pos+1
	int leftnode;
	int thisnode;
	int rightnode;
	for (int time = initialtime; time <= finaltime; time++)
	{
		if(usedump)
		{
			// Reset the slice pointers to avoid error messages
			for (int i = 1; i < MPIsize; i++)
			{
				// Send new time to all nodes first:
				MPI_Bsend(NULL, 0, MPI_INT,
				   i, // dest
				   TAG_RESETSLICEPOINTERS, // tag
				   MPI_COMM_WORLD);
			};
		};

		for (int pos=0;pos<NUMT;pos++) 
		{
			if (usedump)
			{
				// Use data dump
				datadumpload(&pos, &initialtime, &time, &part);
			}
			else
			{
				// Generate initial conditions
				//cout << "M: setInitialConditionsForSlice, pos=" << pos << "\n" << endl;
				setInitialConditionsForSlice(0);

				//cout << "M: enforceBCsforslice, pos=" << pos << "\n" << endl;
				enforceBCsforslice(0);

				// Now send this block to a worker:
				slice.MPIPack();
			};
			leftnode = 1 + (MPIsize - 1) * positivemod(pos - 1,NUMT) / NUMT;
			thisnode = 1 + (MPIsize - 1) * pos / NUMT;
			rightnode = 1 + (MPIsize - 1) * positivemod(pos + 1,NUMT) / NUMT;
			if (MSG_SENDDATA) cout << "Manager: Sending slice " << pos + 1 << " to node " << 
				thisnode << endl;
			sendBufferToNode(
				thisnode,  // node we send this to.
				pos, // position
				time,  // Initial time.
				ALL_PARTS // part
				);
			if (leftnode != thisnode)
			{
				// Since we are just at the border, send this
				// also to the left node:
				sendBufferToNode(
					leftnode,  // node we send this to.
					pos, // position
					time,  // Initial time.
					ALL_PARTS // part
					);
				if (leftnode == MPIsize - 1)
				{
					// Special treatment, as last node will not 
					// get all slices as neighbors:
					MPI_Send(NULL, 0, MPI_INT,
					   leftnode, // dest
					   TAG_RESETFIRSTSLICEPOINTER, // tag
					   MPI_COMM_WORLD);
				};
			};
			if (rightnode != thisnode)
			{
				// Also send this to right node
				sendBufferToNode(
					rightnode,  // node we send this to.
					pos, // position
					time,  // Initial time.
					ALL_PARTS // part
					);
			};
		}; // for (pos..)
	}; // for (time..)

	// Special treatment for last node, as last node will not 
	// get all slices as neighbors:
	MPI_Send(NULL, 0, MPI_INT,
	   MPIsize - 1, // dest
	   TAG_INCREMENTLASTSLICEPOINTER, // tag
	   MPI_COMM_WORLD);

	// Finished distributing work, but we have to wait until all
	// have received this. For this reason, we tell all workers
	// to notify the manager when they are done:
	for (int i = 1; i < MPIsize; i++)
	{
	   MPI_Send(
		   NULL, // buffer
		   0, // count
		   MPI_INT, // MPI_Datatype
		   i, // dest
		   TAG_ALLSLICESSENT, // tag
		   MPI_COMM_WORLD);
	}

}

/////////////////////////////////////////////////////////////////////
// Class for storing snapshot info
//
// We can not assume that the complete memory information of all
// parallel processes will fit into the manager memory.
// Therefore, we only store information in 2 orthogonal slices:
// * Information at sx == NUMT/2 and arbitrary sy, sz is 
//   stored in 'slice'.
// * Information at sy == NUMT/2 and arbitrary sx, sz is 
//   stored in 'sliceT' (transposed Slice).
// (Currently this is implemented with a number of
//  difficult to read hacks.)
class snapshotInfo
{
public:
	int time;
	int count;
	SimulationSlice* slice;
	SimulationSlice* sliceT; // transposed slice
	// Transposed sliceT contains
	// information with sx and sy swapped.
	// so that we can provide information in sx direction.

	// The following is a quick-fix for the 
	// snapshot output b_x, b_y, etc.
	SimulationSlice* slice2; // also probe at h/2 + 1
	SimulationSlice* sliceT2; 

	snapshotInfo();
	void reset();
	void add(SimulationSlice **ppslice, int pos, int time);
	  // pointer to pointer slice, as we might swap the whole thing.
};

snapshotInfo::snapshotInfo()
{
	slice = NULL;
	sliceT = NULL;
	slice2 = NULL;
	sliceT2 = NULL;

	reset();
};

void snapshotInfo::reset()
{
	time = 0;
	count = 0;

	DisposeSimulationSlice(slice);
	slice = NULL;

	DisposeSimulationSlice(sliceT);
	sliceT = NULL;

	DisposeSimulationSlice(slice2);
	slice2 = NULL;

	DisposeSimulationSlice(sliceT2);
	sliceT2 = NULL;
};

void snapshotInfo::add(SimulationSlice **ppslice, int pos, int time)
{
	// Avoid compiler warning about unused variable
	int dummy = time; dummy++;
	
	if (slice == NULL) slice = GetNewSimulationSlice();
	if (sliceT == NULL) sliceT = GetNewSimulationSlice();
	if (slice2 == NULL) slice2 = GetNewSimulationSlice();
	if (sliceT2 == NULL) sliceT2 = GetNewSimulationSlice();

	count++;  // Increase counter

	// Don't forget: the 'pos' and 'sx' are off by one:
	// (pos = 0 .. NUMT; sx = 0 .. NUMT+2)
	int sx = pos + 1;
	if( sx == NUMT/2)
	{
		// This slice is in the probing plane: 
		// We want to save the whole slice
		// QUICKFIX: Instead of copying, we just
		//           swap the pointers:
		swappointers( &slice, ppslice);
	}
	else if( sx == NUMT/2 + 1)
	{
		// This slice is in the probing plane: 
		// We want to save the whole slice
		// QUICKFIX: Instead of copying, we just
		//           swap the pointers:
		swappointers( &slice2, ppslice);
	}
	else
	{
		// In all other cases, we just save the
		// sy == NUMT/2 component:
		sliceT->swapzplane(*ppslice, sx, NUMT/2);
		// or the sy == NUMT/2 + 1 component
		sliceT2->swapzplane(*ppslice, sx, NUMT/2 + 1);
	};
};

/////////////////////////////////////////////////////////////////////
// Class for storing angle measurement info
//
// Store only the fields pix, piy, ez, ux, uy, u
class angleInfo
{
public:
	int time;
	int count;

	// We store the WHOLE simulation field,
	// but only P (pix, piy, ez) and U (ux, uy, u) fields.
	SimulationSlice** slicelist; // contains pos = 0..NUMT-1

	angleInfo();
	~angleInfo();
	void reset();
	void add(SimulationSlice *pslice, int pos, int time);
	void getAllSimulationSpacePU();
	void linkGlobalLowerCasePU();
};

angleInfo::angleInfo()
{
	slicelist = NULL;
	reset();
};

angleInfo::~angleInfo()
{
	reset();
};

void angleInfo::reset()
{
	time = 0;
	count = 0;

	if (slicelist != NULL)
	{
		for (int i=0; i<NUMT; i++)
		{
			DisposeSimulationSlice(slicelist[i]);
			slicelist[i] = NULL;
		};
		delete [] slicelist;
		slicelist = NULL;
	};
};

void angleInfo::add(SimulationSlice *pslice, int pos, int time)
{
	// Avoid compiler warning about unused variable
	int dummy = time; dummy++;
	
	// Allocate memory only for P (pix, piy, ez) and U (ux, uy, u) fields:
	if (slicelist == NULL) 
	{
		slicelist = new SimulationSlice*[NUMT];
		assert(slicelist != NULL);
	
		for (int i=0; i<NUMT; i++)
			slicelist[i] = NULL;
	};

	// Don't forget: the 'pos' and 'sx' are off by one:
	// (pos = 0 .. NUMT; sx = 0 .. NUMT+2)
	// int sx = pos + 1;

	if (slicelist[pos] == NULL) slicelist[pos] = GetNewSimulationSlice(FIELDS_PU);
	assert(slicelist[pos] != NULL);

	count++;  // Increase counter

	//cout << "pos: " << pos 
	//	<< ", slicelist[pos].selection: " << slicelist[pos]->varselection
	//	<< ", pslice.selection: " << pslice->varselection << endl;

	// We copy all required information to the new location:
	slicelist[pos]->copyfrom(pslice, FIELDS_PU);
};

// Allocates enough simulation space to contain the complete P and U fields of a simulation:
void angleInfo::getAllSimulationSpacePU()
{
	// Allocate memory only for P (pix, piy, ez) and U (ux, uy, u) fields:
	if (slicelist == NULL) 
	{
		slicelist = new SimulationSlice*[NUMT];
		assert(slicelist != NULL);
	};

	for (int i=0; i<NUMT; i++)
	{
		slicelist[i] = GetNewSimulationSlice(FIELDS_PU);
	};
};

void angleInfo::linkGlobalLowerCasePU()
{
	// Link the fields
	for (int i = 0; i<NUMT; i++)
	{
		// Don't forget: the 'pos' and 'sx' are off by one:
		// (pos = 0 .. NUMT; sx = 0 .. NUMT+2)
		// sx = pos + 1;
		assert(slicelist[i] != NULL);
		slicelist[i]->linkGlobalLowerCase(i+1);
	};

	// Also link the periodic boundaries correctly:
	slicelist[NUMT-1]->linkGlobalLowerCase(0);
	slicelist[0]->linkGlobalLowerCase(NUMT+1);
};

/////////////////////////////////////////////////////////////////////

// Parallel version of Evolve()
// Part of the code from Evolve() is directly in worker() below!
void EvolveMPI() {
	bool dowork = true;

	int flag;
	MPI_Status status;
	MPI_Status status1;
	int pos;  // internal position (0 to NUMT-1)
	int time; // time step
	int part; // part of calculation (1 or 2)
	SimulationSlice *pslice;
	int sx;

	outputVariables ov;
	outputVariables *pOVarBuffer; // Contains a list for many elements
	int maxOVarBuffer = NUMT * 2; // Should be enough space to also catch delayed elements
	pOVarBuffer = new outputVariables[maxOVarBuffer];
	int OVpos; // local variable for buffer position.

	angleInfo *pAInfoBuffer;  // Contains list for many elements
	int maxAInfoBuffer = NUMT * 2; // Should be enough space 
	pAInfoBuffer = new angleInfo[maxAInfoBuffer];
	int AIpos; // local variable for buffer position.

	snapshotInfo *pSInfoBuffer; // Contains list for many elements
	int maxSInfoBuffer = NUMT * 2; // Should be enough space 
	pSInfoBuffer = new snapshotInfo[maxSInfoBuffer];
	int SIpos; // local variable for buffer position.
	// Special treatment for output of b_x: 
	// Need two values for u and ux:
	SUNLieMatrix **snapshotu = new SUNLieMatrix*[2*(NUMT+2)];
	SUNLieMatrix **snapshotux = new SUNLieMatrix*[2*(NUMT+2)];

	statisticsVariables stats;
	statisticsVariables *pStatsBuffer; // Contains a list for many elements
	int maxStatsBuffer = MPIsize * NUMT; // Should be enough space to also catch delayed elements
	pStatsBuffer = new statisticsVariables[maxStatsBuffer];
	int statspos; // local variable for buffer position.

	int datadumpstate = DATADUMP_NOTHING; // First, middle, or last (const, defined above)
	int datadumpcurrent = -1; // Current dump time stamp.
	int datadumpnumber = 0; // Number of slices saved so far.

	SendInfo sinfo; // temporary space.
	SendInfo mostrecentinfo; // Most recent information
	mostrecentinfo.reset();

	double eps = EPS;
	float timefloat = 0; // original 'time' from solve.cpp

	int workersslicesreceived = 0; // Number of workers who received their initial slices
	int workersfinished = 0; // Number of workers who finished their work

	int receivestate = STATE_NOTRECEIVING;
	MPI_Request receiverequest; // Handle for asynchronous receive

	bool timeoutnotreachedyet = true; // We only need to send signal once.

	if (MSG_MESSAGE) cout << myrank << ": Start to work" << endl;
	while (dowork || (receivestate != STATE_NOTRECEIVING))
	{
		// Probe for messages from anybody:
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		while (flag)
		{
			if (MSG_MESSAGE) cout << "Manager: New message from " << status.MPI_SOURCE << ", TAG: " << status.MPI_TAG << endl;
			switch(status.MPI_TAG)
			{
			case TAG_ALLSLICESRECEIVED:
				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, 
				  status.MPI_SOURCE, // rank of source
				  status.MPI_TAG, // message tag
				  MPI_COMM_WORLD, // communicator (handle)
				  &status1);

				workersslicesreceived++;

				if (workersslicesreceived >= MPIsize - 1) // subtract one for the manager
				{
					// Now all workers have acknowledged that they have
					// received all initial fields.

					if (MPIsize > 2)
					{
						// Now we can start the calculation:
						if (MSG_MESSAGE) cout << "Manager: STARTING SIMULATION! " << endl;
						
						for (int i = 1; i < MPIsize; i++)
						{
							MPI_Send(
							   NULL,// buffer
							   0, // count
							   MPI_INT, // MPI_Datatype
							   i, // dest
							   TAG_STARTSIMULATION, // tag
							   MPI_COMM_WORLD);	
						};
					}
					else
					{
						// Special case: only 1 manager + 1 worker?
						// We would need extra treatment, but this is not implemented.

						// Special case: only 1 manager? 
						cout << "MANAGER: WARNING: setInitialConditionsMPI(): Not enough workers!" << endl;
					}
				};
				break;
			case TAG_REACHEDSIMULATIONEND:
				// let us hope no worker finishes twice
				// (otherwise the counting gets messed up)

				// assert (status.MPI_SOURCE has not sent this tag before..)

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, 
				  status.MPI_SOURCE, // rank of source
				  status.MPI_TAG, // message tag
				  MPI_COMM_WORLD, // communicator (handle)
				  &status1);

				workersfinished++;
				if (workersfinished >= MPIsize - 1) // subtract one for the manager
				{
					if (MSG_MESSAGE) cout << "Manager: All work done!" << endl;
					// Work done! Phew! :-)
					dowork = false;


				};
				break;

			case TAG_WANTTOSEND:
				// A node wants to send a data block to us:

				if (MSG_MESSAGE) cout << myrank << ": received TAG_WANTTOSEND"
					<< " from " << status.MPI_SOURCE << endl;

				// Acknowledge message
			    MPI_Recv(&sinfo.pos, 4, MPI_INT, 
					status.MPI_SOURCE, TAG_WANTTOSEND, MPI_COMM_WORLD, &status1);

				if (receivestate == STATE_NOTRECEIVING)
				{
					// Great, I like to have the data:
					if (MSG_MESSAGE) cout << myrank << ": TAG_WANTTOSEND: Accepted. Please send. " << endl;
					
					// We prepare for receiving data 
					// asynchronously: First receive, then send:
					MPI_Irecv( pBufferAllreceive, 
						  maxBufferAllreceive, 
						  MPI_PACKED, 
						  status.MPI_SOURCE, // rank of source
						  TAG_DATABLOCK, // message tag
						  MPI_COMM_WORLD, // communicator (handle)
						  &receiverequest);

					// Now we send back that the other party can start
					// to send data:
					MPI_Send(NULL, 0, MPI_INT,
					   status.MPI_SOURCE, // dest
					   TAG_PLEASESEND, // tag
					   MPI_COMM_WORLD);

					// cout << myrank << ": TAG_WANTTOSEND: send PLEASESEND " << endl;

					receivestate = STATE_RECEIVING;
				}
				else
				{
					// Sorry, currently I have somehing else to do
					if (MSG_MESSAGE) cout << myrank << ": TAG_WANTTOSEND: Failed. Please don't send. " << endl;

					MPI_Send(NULL, 0, MPI_INT,
					   status.MPI_SOURCE, // dest
					   TAG_DONOTSEND, // tag
					   MPI_COMM_WORLD);
				};
				break;

			case TAG_OUTPUTVARIABLES:
				// Receive struct from worker
			    MPI_Recv(
					&ov,
					1, 
					myMPI_outputVariables,
					status.MPI_SOURCE, // rank of source
					status.MPI_TAG, // message tag
					MPI_COMM_WORLD, // communicator (handle)
					&status1);

				if (MSG_OUTPUT) cout << "Master: Receiving Output Measurement "
					<< " from worker " << status.MPI_SOURCE
					<< " at pos " << ov.pos
					<< ", time " << ov.time << endl;

				// Simple Cache algorithm: time mod BufSize
				OVpos = positivemod(ov.time, maxOVarBuffer);
				
				if (pOVarBuffer[OVpos].count == 0)
				{
					// Great, buffer space exists and it is empty:

					// Add what we received:
					pOVarBuffer[OVpos].time = ov.time;
					pOVarBuffer[OVpos].add(ov);
				}
				else if (pOVarBuffer[OVpos].time == ov.time)
				{
					// Great, a new element to an existing place:
					pOVarBuffer[OVpos].add(ov);
				}
				else
				{
					// Buffer size already in use at a different time.
					// Probably we could handle this more gracefully,
					// but it is easier to just increase the buffer size
					// when necessary.
					cout << "Master: WARNING: Buffer size for outputVariables is too small! " 
						<< "Measurement at time " << pOVarBuffer[OVpos].time << " lost." << endl;
					
					// We have to throw away the old measurement.
					pOVarBuffer[OVpos].reset();

					pOVarBuffer[OVpos].time = ov.time;
					pOVarBuffer[OVpos].add(ov);
				};

				if (pOVarBuffer[OVpos].count >= NUMT) // Number of slices in x-direction 
				{
					// Output of the measurement:
					ov = pOVarBuffer[OVpos];

					// First complete Gausslaw:
					SUNmatrix temp;
#if defined SU2 || defined SU2double
					temp.a = ov.gausslawtempa;
					temp.b = ov.gausslawtempb;
					temp.c = ov.gausslawtempc;
#elif defined SU3 || defined SU3double
					int k=0;
					temp.a1 = ov.gausslawtemp[k++];
					temp.a2 = ov.gausslawtemp[k++];
					temp.a3 = ov.gausslawtemp[k++];
					temp.a4 = ov.gausslawtemp[k++];
					temp.a5 = ov.gausslawtemp[k++];
					temp.a6 = ov.gausslawtemp[k++];
					temp.a7 = ov.gausslawtemp[k++];
					temp.a8 = ov.gausslawtemp[k++];
#elif defined SU3tvmet
					int k=0;
					for (int i=0; i<3; i++)
					{
						for (int j=0; j<3; j++)
						{
							temp.mat(i, j) = ov.gausslawtemp[k++];
							temp.mat(i, j) += I * ov.gausslawtemp[k++];
						}
					};
#else
#error Please specify SU(N) code here.
#endif
					ov.gausslaw = checkGaussLawPart2(temp);

					// Set time:
					timefloat = (ov.time - 2)*eps;
					
					// Output:
					if (ov.time - 2 == 0)
					{
						// First time only measurement:
						outputMeasurementsMPImanagerFirstTime(timefloat, ov);
					}
					else
					{
						// all other times the usual measurement:
						outputMeasurementsMPImanager(timefloat, ov);

						/*
						// Simulate twice output of data at the beginning
						// THIS IS A QUICK FIX to simulate the original output of solve.cpp
						if (pOVarBuffer[OVpos].time-1 == 2)
							outputMeasurementsMPImanager(timefloat, ov);
						*/
					};

					// Save the data for later use as a data-dump
					pOVarDataDump[lenOVarDataDump] = ov;
					lenOVarDataDump++;
					if (lenOVarDataDump > maxOVarDataDump)
					{
						// Error! Too many measurements for dump
						cout << "ERROR: EvolveMPI(): Too many measurements for data dump!" << endl;
						
						// Some ad-hoc behavior to save the world:
						lenOVarDataDump = 0;
					};

					// Free the space for following times
					pOVarBuffer[OVpos].reset();
				};
				break;
			case TAG_STATISTICSVARIABLES:
				// Receive struct from worker
			    MPI_Recv(
					&stats,
					1, 
					myMPI_statisticsVariables,
					status.MPI_SOURCE, // rank of source
					status.MPI_TAG, // message tag
					MPI_COMM_WORLD, // communicator (handle)
					&status1);

				if (MSG_OUTPUT) cout << "Master: Receiving Statistics information "
					<< " from worker " << status.MPI_SOURCE
					<< ", time " << stats.time << endl;

				//cout << "time: " << stats.time
				//	<< ", maxStatsBuffer: " << maxStatsBuffer
				//	<< ", MPIsize: " << MPIsize
				//	<< ", stats.node: " << stats.node
				//	<< endl;

				// Simple Cache algorithm: time mod BufSize
				// Data are organized as time * MPIsize + node
				statspos = positivemod(stats.time, maxStatsBuffer / MPIsize) * MPIsize;
				assert(statspos + (MPIsize-1)< maxStatsBuffer);

				if (pStatsBuffer[statspos + 0].countnodes == 0)
				{
					// Great, buffer space exists and it is empty:

					// Add what we received to the global sum:
					pStatsBuffer[statspos + 0].time = stats.time;
					pStatsBuffer[statspos + 0].add(stats);

					// Also, save the single node's data separately
					pStatsBuffer[statspos + stats.node] = stats;
				}
				else if (pStatsBuffer[statspos + 0].time == stats.time)
				{
					// Great, a new element to an existing place:
					pStatsBuffer[statspos + 0].add(stats);

					// Also, save the single node's data separately
					pStatsBuffer[statspos + stats.node] = stats;
				}
				else
				{
					// Buffer size already in use at a different time.
					// Probably we could handle this more gracefully,
					// but it is easier to just increase the buffer size
					// when necessary.
					cout << "Master: WARNING: Buffer size for statisticsVariables is too small! " 
						<< "Measurement at time " << pStatsBuffer[statspos * MPIsize + 0].time << " lost." << endl;
					
					// We have to throw away the old measurement.
					pStatsBuffer[statspos + 0].reset();

					pStatsBuffer[statspos + 0].time = stats.time;
					pStatsBuffer[statspos + 0].add(stats);
				};

				if (pStatsBuffer[statspos + 0].countnodes >= MPIsize-1) // Number of worker nodes
				{
					// Output of the measurement:
					stats = pStatsBuffer[statspos + 0];

					// Set time:
					timefloat = (stats.time - 2)*eps;
					
					// all other times the usual measurement:
					outputStatistics(timefloat, &pStatsBuffer[statspos + 0]);

					// Free the space for following times
					pStatsBuffer[statspos + 0].reset();
				};

				break;
			case TAG_FINISHEDSLICE:
				// A node wants to tell us about the completion of a part:

				if (MSG_MESSAGE) cout << myrank << ": received TAG_FINISHEDSLICE"
					<< " from " << status.MPI_SOURCE << endl;

				// Acknowledge message
			    MPI_Recv(&sinfo.pos, 4, MPI_INT, 
					status.MPI_SOURCE, TAG_FINISHEDSLICE, MPI_COMM_WORLD, &status1);

				// If we wanted, we could collect all the data to keep
				// track of the current state of the whole simulation.

				// For now we are only interested in keeping the latest
				// time (in order to coordinate the time-out).
				if ((sinfo.time > mostrecentinfo.time)
					|| ((sinfo.time == mostrecentinfo.time) &&
					    (sinfo.part > mostrecentinfo.part)))
				{
					mostrecentinfo = sinfo;
					mostrecentinfo.node = status.MPI_SOURCE;
				};

				break;
			default:
				// Completely unexpected message!
				cout << "Manager: WARNING: EvolveMPI(): Unexpected message tag " << status.MPI_TAG 
					<< " sent from node " << status.MPI_SOURCE << endl;
			}; // switch(status.MPI_TAG)
			
			// More messages to proceed?
			MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		}; // while (flag)

		// At this point (right after we checked for messages),
		// we can check for the time-out:
		if (timeoutnotreachedyet && (MPI_Wtime() - startMPItime > TIMEOUT))
		{
			// send TAG_TIMEOUT to all: [ ]
			cout << "0: Time-out! All activity will stop soon." << endl;

			time = mostrecentinfo.time + 2; // Two time-steps more than most-recent one should be enough for graceful exit.
			for (int i = 1; i < MPIsize; i++)
			{
				// Send new time to all nodes first:
				MPI_Bsend(&time, 1, MPI_INT,
				   i, // dest
				   TAG_TIMEOUT, // tag
				   MPI_COMM_WORLD);
			};

			// Change global parameters accordingly. 
			STEPS = time; // Set limit for number of steps.
			CHECKPOINTUPDATE = STEPS - 3; // Set new checkpoint to start saving.

			timeoutnotreachedyet = false;
		};

		switch(receivestate)
		{
		case STATE_RECEIVING:
			MPI_Test(&receiverequest, &flag, &status);
			//MPI_Wait(&receiverequest, &status); flag = true;
			//cout << myrank << ": receive flag " << flag << endl;

			if (! flag)
			{
				// Not finished asynchronous communication yet.
				break; // Exit the switch() command.
			}
			else
			{
				// Finished receiving.
				// Receive the data set:
				IreceiveBufferFromNode(
					status.MPI_SOURCE,  // receive from worker
					&pos, // position
					&time,  // Initial time.
					&part,
					&status
					);
				if (MSG_SENDDATA) cout << myrank << ": ==> IReceiving data from " << status.MPI_SOURCE
					<< " at pos " << pos 
					<< ", time " << time 
					<< ", part " << part << endl;

				// Small error check
				// (can be removed if code works correctly:
				assert(pos == positivemod(pos, NUMT)); // worker should send valid position

				int itime = time - 2;  // Definition corresponding to worker()
				if ((itime > datadumpinitialtime + 1)
						     && ((itime%CHECKPOINTUPDATE==0)
						         || (itime%CHECKPOINTUPDATE==1)))
				{
					// Do a data dump to disk.
					// For this we don't need to unpack, just dump to a file:

					switch(datadumpstate)
					{
					case DATADUMP_NOTHING:
					case DATADUMP_LAST: // we had finished the last round.
						// This is the first element:
						datadumpcurrent = time;
						if (itime%CHECKPOINTUPDATE==1) datadumpcurrent = time-1; // in case elements arrive in wrong order
						datadumpnumber = 1; // reset counter
						datadumpstate = DATADUMP_FIRST;
						break;
					case DATADUMP_FIRST:
					case DATADUMP_MIDDLE:
						if ((time != datadumpcurrent) && (time-1 != datadumpcurrent))
						{
							// Out of sync?!? Data dump is not complete...
							cout << "ERROR: EvolveMPI() datadump out of sync." << endl;
							datadumpstate = DATADUMP_NOTHING;
						}
						else
						{
							datadumpnumber++;
							datadumpstate = DATADUMP_MIDDLE;
							if (datadumpnumber == 2 * NUMT) // One dump contains 2 full time steps (i.e. all slices)
							{
								datadumpstate = DATADUMP_LAST;
							};
						};
						break;
					default:
						// Should never happen.
						cout << "ERROR: EvolveMPI() - switch(datadumpstate)" << endl;
					};
					datadumpsave(&pos, &time, &part, datadumpstate);
				};

				if ((itime+1)%ANGLEUPDATE==0) // Note that worker sends time-1
				{
					// prepare everything for measureangle(time);
					// We use the data for a snapshot.

					// Unpack
					pslice = GetNewSimulationSlice();
					pslice->MPIUnpack(part);

					// Currently, the only reason for the manager to obtain data
					// is to snapshot them to write them to a file:

					// Simple Cache algorithm: time mod BufSize
					AIpos = positivemod(time, maxAInfoBuffer);
					
					if (pAInfoBuffer[AIpos].count == 0)
					{
						pAInfoBuffer[AIpos].time = time;
						pAInfoBuffer[AIpos].add(pslice, pos, time);
					}
					else if (pAInfoBuffer[AIpos].time == time)
					{
						// Great, a new element to an existing place:
						pAInfoBuffer[AIpos].add(pslice, pos, time);
					}
					else
					{
						// Buffer size already in use at a different time.
						// Probably we could handle this more gracefully,
						// but it is easier to just increase the buffer size
						// when necessary.
						cout << "Master: WARNING: Buffer size for angle measurement is too small! " 
							<< "Measurement at time " << pAInfoBuffer[AIpos].time << " lost." << endl;
						
						// We have to throw away the old measurement.
						pAInfoBuffer[AIpos].reset();

						pAInfoBuffer[AIpos].add(pslice, pos, time);
					};

					if (pAInfoBuffer[AIpos].count >= NUMT) // Number of slices in x-direction 
					{
						// Output of the measurement:
						//for (int i=0; i<NUMT; i++)
						//	cout << (char *) (pAInfoBuffer[AIpos].slicelist[i] == NULL ? "." : "X");
						//cout << endl;
						if (MSG_OUTPUT) cout << myrank << ": Angle-Measurement" << endl;

						// Do the angle measurement
						
						// Set the correct time:
						// (since we received the 'previous slice'
						//  we have to add 1)
						timefloat = (time + 1 - 2)*eps;

						// Link the fields
						pAInfoBuffer[AIpos].linkGlobalLowerCasePU();
						
						// Do the angle measurement:
						measureangle(timefloat);

						// Free the space for following times
						pAInfoBuffer[AIpos].reset();
					};

					// Free the resources again
					DisposeSimulationSlice(pslice);

				}; // if ((itime+1)%ANGLEUPDATE==0)

				// The check for SNAPUPDATE must be last,
				// as the original data will be changed.
				if ((itime+1)%SNAPUPDATE==0) // Note that worker sends time-1
				{
					// We use the data for a snapshot.

					// Unpack
					pslice = GetNewSimulationSlice();
					pslice->MPIUnpack(part);

					// Currently, the only reason for the manager to obtain data
					// is to snapshot them to write them to a file:

					// Simple Cache algorithm: time mod BufSize
					SIpos = positivemod(time, maxSInfoBuffer);
					
					if (pSInfoBuffer[SIpos].count == 0)
					{
						pSInfoBuffer[SIpos].time = time;
						pSInfoBuffer[SIpos].add(&pslice, pos, time);
					}
					else if (pSInfoBuffer[SIpos].time == time)
					{
						// Great, a new element to an existing place:
						pSInfoBuffer[SIpos].add(&pslice, pos, time);
					}
					else
					{
						// Buffer size already in use at a different time.
						// Probably we could handle this more gracefully,
						// but it is easier to just increase the buffer size
						// when necessary.
						cout << "Master: WARNING: Buffer size for snapshot is too small! " 
							<< "Measurement at time " << pSInfoBuffer[SIpos].time << " lost." << endl;
						
						// We have to throw away the old measurement.
						pSInfoBuffer[SIpos].reset();

						pSInfoBuffer[SIpos].add(&pslice, pos, time);
					};

					if (pSInfoBuffer[SIpos].count >= NUMT) // Number of slices in x-direction 
					{
						// Output of the measurement:

						// Don't forget: the 'pos' and 'sx' are off by one:
						// (pos = 0 .. NUMT; sx = 0 .. NUMT+2)
						sx = pos + 1;

						// Do the snapshot
						
						// Set the correct time:
						// (since we received the 'previous slice'
						//  we have to add 1)
						timefloat = (time + 1 - 2)*eps;


						// Link to transposed fields at sy == NUMT/2
						for (int i = 0; i<=NUMT+1; i++)
						{
							// QUICKFIX: This is unreadable code,
							// don't try to understand it (but it works!)
							pSInfoBuffer[SIpos].sliceT->linkGlobalLowerCase(i, -NUMT/2 + i);

							// Ok, here the explanation:
							// snapshot() below will probe these fields at
							// [i][NUMT/2], but this will just point at
							// [0][NUMT/2-NUMT/2+i]==[0][i] in sliceT, the transposed slice.
							// {(c) Andi}

							// The same in the plane NUMT/2 + 1
							// pSInfoBuffer[SIpos].sliceT2->linkGlobalLowerCase(i, -(NUMT/2+1) + i);
						};

						// Also link the periodic boundaries correctly:
						pSInfoBuffer[SIpos].sliceT->linkGlobalLowerCase(0, -NUMT/2 + 0 + NUMT);
						pSInfoBuffer[SIpos].sliceT->linkGlobalLowerCase(NUMT+1, -NUMT/2 + NUMT+1 - NUMT);

						// The transposed versions have to be linked more carefully:
						for (int i = 0; i<=NUMT+1; i++)
						{
							// Periodic boundary conditions:
							int ii = i;
							if (ii == 0) ii = NUMT;
							if (ii == NUMT+1) ii = 1;

							// QUICKFIX: This is again unreadable code,
							// but it works as above.
							::u[i] = snapshotu + (2 * i) + (-NUMT/2);
							// The difference is that for one sx value
							// we have to save two sy values: NUMT/2 and NUMT/2+1
							// In order to do so, we save the information
							// in a list snapshotu, which contains pointers
							// to sy == NUMT/2 and NUMT/2+1 alternatively.
							snapshotu[2 * i] = pSInfoBuffer[SIpos].sliceT->u[0][ii];
							snapshotu[2 * i + 1] = pSInfoBuffer[SIpos].sliceT2->u[0][ii];

							// The same for ux:
							::ux[i] = snapshotux + (2 * i) + (-NUMT/2);
							snapshotux[2 * i] = pSInfoBuffer[SIpos].sliceT->ux[0][ii];
							snapshotux[2 * i + 1] = pSInfoBuffer[SIpos].sliceT2->ux[0][ii];
						};

						// Link to slice at sx == NUMT/2
						// Snapshot uses NUMT/2
						pSInfoBuffer[SIpos].slice->linkGlobalLowerCase(NUMT/2);

						// and also at sx == NUMT/2 + 1
						pSInfoBuffer[SIpos].slice2->linkGlobalLowerCase(NUMT/2+1);

						// Do the snapshot:
						snapshot(timefloat);

						// Free the space for following times
						pSInfoBuffer[SIpos].reset();
					};

					// Free the resources again
					DisposeSimulationSlice(pslice);
				}; // if ((itime+1)%SNAPUPDATE==0)

				// We can free all resources and send the next.
				receivestate = STATE_NOTRECEIVING;
			};

			/*
			if (!receivebuffer.empty())
			{
				// receive something
			};
			*/
			break;
		case STATE_NOTRECEIVING:
			// Nothing we can do right now...
			break;
		default:
			assert(false); // Should never happen
			break;
		}; // switch(receivestate)

		// Quick check if a message is waiting in the mean-time
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		if (!flag)
		{
			// Idle time for the manager.
			// sit(back) && relax();

			// Don't waste CPU-time:
			mysleep(DEFAULT_SLEEP_MS);
		};
	}; // while (dowork)
};

void createDirectoryStructureMPI()
{
	// Create the directory structure:
	if(mymkdir("data_mpi", 0711) == -1)
	{
		// Nothing to worry about: Most likely the directory already existed.
		if (MSG_DISK) cout << "Could not create directory 'data_mpi'." << endl;
	}
	else
		cout << "Directory 'data_mpi' successfully created." << endl;

	if(mymkdir("data_mpi/snapshot", 0711) == -1)
	{
		if (MSG_DISK) cout << "Could not create directory 'data_mpi/snapshot'." << endl;
	}
	else
		cout << "Directory 'data_mpi/snapshot' successfully created." << endl;

	if(mymkdir("data_mpi/datadump", 0711) == -1)
	{
		if (MSG_DISK) cout << "Could not create directory 'data_mpi/datadump'." << endl;
	}
	else
		cout << "Directory 'data_mpi/datadump' successfully created." << endl;

	if(mymkdir("data_mpi/statistics", 0711) == -1)
	{
		if (MSG_DISK) cout << "Could not create directory 'data_mpi/statistics'." << endl;
	}
	else
		cout << "Directory 'data_mpi/statistics' successfully created." << endl;
};

void manager()
{
	extern void readParameters(char*);
	extern void copyParameterFile(char *filename);
	extern void printDivider();

        // Set initial time for timing output
        gettimeofday(&starttime, 0);
	outputtime = starttime;
	outputcounter = 0;

	printDivider();
	
	readParameters("params.txt");

	// send global parameters to all workers
	broadcastParametersMPI();
	
	printDivider();

	allocateMemoryMPI();

	fft_init();

#if defined REDO_ANGLE_FROM_DUMP
	// first set up some global simulation space like in EvolveMPI()
	angleInfo AI;
	AI.getAllSimulationSpacePU();
	AI.linkGlobalLowerCasePU();

	// Don't do the simulation, just read 
	// from angle dump and do angle measurement
	measureanglefromdump();
#else

	createDirectoryStructureMPI();

	copyParameterFile("params.txt");

	// open files to put the data in first	
	cbar_out.open("data_mpi/cbar.dat", ios::out);
	energy_out.open("data_mpi/energy.dat", ios::out);
	gausslaw_out.open("data_mpi/gausslaw.dat", ios::out);

	initializeVelocitiesAndCoefficients();
		
	// setInitialConditions();
	setInitialConditionsMPI();
		
	if (MSG_MESSAGE) cout << myrank << ": Print initial conditions" << endl;

	printDivider();

	cout << "Simulation of " << SUNmatrix_string << endl;

	cout << "Spatial Step Size (A): " << A << endl;

	cout << "Temporal Step Size (EPS): " << EPS << endl;

	cout << "Standard Deviation of Initial Distribution (SIG): " << SIG << endl;

	cout << "Longitudinal Box Size (A*" << NUM << "): " << A*NUM << endl;

	cout << "Transverse Box Size (A*" << NUMT << "): " << A*NUMT << endl;

	printDivider();
/*
	double phirms = btEnergyDensity();
	cout << "Initial Bt RMS: " << phirms << endl;

	double phizrms = bzEnergyDensity();
	cout << "Initial Bz RMS:" << phizrms << endl;

	double etrms = etEnergyDensity()*4;
	cout << "Initial et RMS: " << etrms << endl;

	double ezrms = ezEnergyDensity()*4;
	cout << "Initial ez RMS: " << ezrms << endl;

	InitialEnergy = initialED();
	cout << "Initial Energy Density: " << InitialEnergy << endl;

	double gauss = checkGaussLaw();
	cout << "Initial Gauss Law Check: " << gauss << endl;

	printDivider();

	int dwidth = 13;
	cout.width(dwidth); cout << "Time";
	cout.width(dwidth); cout << "<Ez>";
	cout.width(dwidth); cout << "<Et>";
	cout.width(dwidth); cout << "<Bz>";
	cout.width(dwidth); cout << "<Bt>";
	cout.width(dwidth); cout << "<HL>";
	cout.width(dwidth); cout << "GaussCheck";
	cout.width(dwidth); cout << "EnergyViol";
	cout << endl;
	
	printDivider();
*/

	if (MSG_MESSAGE) cout << myrank << ": Call EvolveMPI()" << endl;

	//Evolve();
	EvolveMPI();

	printDivider();
	cout << "Done." << endl;
	printDivider();

	// finally close files
	cbar_out.close();
	energy_out.close();
	gausslaw_out.close();

#endif //  defined REDOANGLE

	fft_end();

	// Finally let all workers die:
	for (int i = 1; i < MPIsize; i++)
	{
	   MPI_Send(NULL, 0, MPI_INT,
		   i, // dest
		   TAG_DIE, // tag
		   MPI_COMM_WORLD);
	};

};

// for error checking purposes
void dumptimestate()
{
	char *partchar = ".-|+****";
	char *blockchar = ".B234";
	char *borderchar = " <>X";
	char *loadstatechar = " B234";
	char *bmchar = "=- +#";

	cout << myrank << ": Dump: ";
	//if (MSG_DYNAMIC) cout << dynamicloadstateleft - STATE_BM_NOTWAITING << bordermoveleft << "] ";
	if (MSG_DYNAMIC) cout << loadstatechar[dynamicloadstateleft - STATE_BM_NOTWAITING] << bmchar[bordermoveleft+2] << "] ";

	for (int i=0; i<NUMT; i++)
	{
		cout << borderchar[((firstslice == i)?1:0) | ((lastslice == i)?2:0)
		       | ((firstslice == i-1)?1:0) | ((lastslice == i-1)?2:0)];
		//if (firstslice == i) cout << "< ";
		//if (lastslice == i) cout << "> ";
		cout << sliceinfo[i].time;
		cout << partchar[sliceinfo[i].currentpart + 4*sliceinfo[i].simulationblock]
			<< blockchar[sliceinfo[i].currentblock]
			<< "";
		/*
		if (sliceinfo[i].time == 0)
		{
			cout << "-- ";
		}
		else
		{
			cout << partchar[sliceinfo[i].currentpart]
			// cout << (char) (sliceinfo[i].currentpart + 'a')
			// << (char) (sliceinfo[i].previouspart + 'a')
			// << (char) (sliceinfo[i].currentblock + 'A')
			<< blockchar[sliceinfo[i].currentblock]
			// << (char) (sliceinfo[i].previousblock + 'A')
			<< " ";
		};
		*/
		//if (firstslice == i) cout << "< ";
		//if (lastslice == i) cout << "> ";
	};
	cout << borderchar[((firstslice == NUMT-1)?1:0) | ((lastslice == NUMT-1)?2:0)];

	//if (MSG_DYNAMIC) cout << "[" << dynamicloadstateright - STATE_BM_NOTWAITING << bordermoveright << "";
	if (MSG_DYNAMIC) cout << "[" << loadstatechar[dynamicloadstateright - STATE_BM_NOTWAITING] << bmchar[bordermoveright+2] << "";

	cout << endl;
};

// Check whether calculation is possible at a specific position
int workpossible(int pos)
{
    // cout << myrank << ": workpossible(" << pos << ")" << endl;
	assert(pos >= 0 && pos < NUMT);  // Caller is responsible for correct position.

	// Restrict input to valid position
	pos = positivemod(pos,NUMT);
	int posleft = positivemod(pos-1,NUMT);
	int posright = positivemod(pos+1,NUMT);

	// current time
	int time = sliceinfo[pos].time;

	// current part
	int part = sliceinfo[pos].currentpart;

	if (sliceinfo[pos].simulationblock > 0)
	{
		// Simulation on this slice blocked.
		return false;
	};

	switch (part)
	{
	case 0:
		// No part here yet, we can not work.
		return false;
		break; // actually don't need break after return, but ok..
	case PART1:
		// We finished the first part of the calculation.
		// Now let us check, if we have all prerequisites to finish the 
		// second part:

		// We require:
		// *) All parts finished in the three slices at the previous timestep
		// *) First part finished in all adjacent neighbors at same time.

		// data available at this position?
		if (sliceinfo[pos].currentslice == NULL) return false;

		if (sliceinfo[pos].previousslice == NULL) return false;
		if (sliceinfo[pos].previouspart != ALL_PARTS) return false;

		// data for time 'time' available at left position?
		if (sliceinfo[posleft].time == time)
		{
			// part 1 should be available in current step:
			if (sliceinfo[posleft].currentslice == NULL) return false;
			if (!(sliceinfo[posleft].currentpart & PART1)) return false;

			// all parts should be available in previous step:
			if (sliceinfo[posleft].previousslice == NULL) return false;
			if (sliceinfo[posleft].previouspart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posleft].time == time - 1)
		{
			// left side is not up-to-date
			return false;
		}
		else
		{
			// Internal inconsistency: Neighbouring slices should
			// only be apart one step.
			cout << myrank << ": WARNING: workpossible(): timing inconsistency with left neighbor" << endl;
			dumptimestate();
		};

		// data for time 'time' available at right position?
		if (sliceinfo[posright].time == time)
		{
			// part 1 should be available in current step:
			if (sliceinfo[posright].currentslice == NULL) return false;
			if (!(sliceinfo[posright].currentpart & PART1)) return false;

			// all parts should be available in previous step:
			if (sliceinfo[posright].previousslice == NULL) return false;
			if (sliceinfo[posright].previouspart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posright].time == time - 1)
		{
			// right side is not up-to-date
			return false;
		}
		else
		{
			// Internal inconsistency: Neighbouring slices should
			// only be apart one step.
			cout << myrank << ": WARNING: workpossible(): timing inconsistency with right neighbor" << endl;
			dumptimestate();
		};

		break;
	case PART2:
		// we have to wait until PART1 arrives:
		return false;
		break;
	case ALL_PARTS:
		// All steps finished, and we want to continue to the next time-step

		// We require:
		// *) End of simulation not reached yet
		// *) All parts finished in all three slices of the same timestep

		// Still time until the end of the simulation:
		if (time >= STEPS+1) return false; // last step reached

		// as long as previous data is blocked, we also can not advance:
		if (sliceinfo[pos].previousblock > 0) return false; 

		// data available at this position?
		if (sliceinfo[pos].currentslice == NULL) return false;

		// data for time 'time' available at left position?
		if (sliceinfo[posleft].time == time)
		{
			// The slice of the same time should be completed:
			if (sliceinfo[posleft].currentslice == NULL) return false;
			if (sliceinfo[posleft].currentpart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posleft].time == time + 1)
		{
			// The slice of the same time should be completed:
			if (sliceinfo[posleft].previousslice == NULL) return false;
			if (sliceinfo[posleft].previouspart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posleft].time == time - 1)
		{
			// left side is not up-to-date
			return false;
		}
		else
		{
			// Internal inconsistency: Neighbouring slices should
			// only be apart one step.
			cout << myrank << ": WARNING: workpossible(): timing inconsistency with left neighbor" << endl;
			dumptimestate();
		};

		// data for time 'time' available at right position?
		if (sliceinfo[posright].time == time)
		{
			// The slice of the same time should be completed:
			if (sliceinfo[posright].currentslice == NULL) return false;
			if (sliceinfo[posright].currentpart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posright].time == time + 1)
		{
			// The slice of the same time should be completed:
			if (sliceinfo[posright].previousslice == NULL) return false;
			if (sliceinfo[posright].previouspart != ALL_PARTS) return false;
		}
		else if (sliceinfo[posright].time == time - 1)
		{
			// right side is not up-to-date
			return false;
		}
		else
		{
			// Internal inconsistency: Neighbouring slices should
			// only be apart one step.
			cout << myrank << ": WARNING: workpossible(): timing inconsistency with right neighbor" << endl;
			dumptimestate();
		};
		break;
	default:
		cout << myrank << ": WARNING: workpossible(): part unknown: " << part << endl;
		dumptimestate();

	};

	// None of the negative checks failed, so it is possible to do some work!
	return true;
};

// Returns, on which slice one can do a part of a work
// returns -1 if no work possible, otherwise number of slice
// for which work can be done.
//
// This routine first checks if either of the border-slices can
// be worked on, and then continues to look for the oldest 
// element for which work can be done from outside to inside
//
// [ ] This routine could be further optimized: 
//     Don't need to go through all elements each time...
static int lastworkslice = -1;
int whichwork()
{
	// cout << myrank << ": Which work?" << endl;

	int workslice = -1; // Default: no slice possible.
	int pos;

	if (firstslice == -1) return -1; // No slice - no work.

	// Can we continue last time's work?
	if ((lastworkslice != -1) && (workpossible(lastworkslice)))
	{
		// Yes! Let's just do that
		workslice = lastworkslice;
		lastworkslice = -1; // For sure not possible twice...
		return workslice;
	};

	// Can we do work on the first-but-one?
	pos = positivemod(firstslice+1,NUMT);
	if (workpossible(pos))
	{
		workslice = pos;
		return workslice;
	};

	// Can we do work on the last-but-one?
	pos = positivemod(lastslice-1,NUMT);
	if (workpossible(pos))
	{
		workslice = pos;
		return workslice;
	};

	// we do work on the oldest that is possible,
	// checking from outside to inside:
	int time = STEPS + 2; // Maximum possible
	int distance = positivemod(lastslice - firstslice, NUMT);
	int halfdistance = distance / 2 + 1;
	//cout << myrank << ": distance " << distance << ", halfdistance " << halfdistance << endl;
	for (int i = 1; i < halfdistance; i++)
	{
		// Check from left side:
		pos = positivemod(firstslice + i, NUMT);
		if (workpossible(pos)
			&& (sliceinfo[pos].time < time))
		{
			// found an older element for which we can 
			// do the calculation
			workslice = pos;
			time = sliceinfo[pos].time;
		};

		// and check from right side:
		pos = positivemod(lastslice - i, NUMT);
		if (workpossible(pos)
			&& (sliceinfo[pos].time < time))
		{
			// found an older element for which we can 
			// do the calculation
			workslice = pos;
			time = sliceinfo[pos].time;
		};
		//cout << myrank << ": check pos1 " << pos1 << ", pos2 " << pos2 << endl;
	};
	if ((workslice != -1) && sliceinfo[workslice].currentpart == ALL_PARTS)
	{
		// Since we will start a new time-step, next time we can continue
		// working on the same slice:
		lastworkslice = workslice;
	};

	return workslice;
};

void worker()
{
	// set neighbors for direct communication:
	leftrank = myrank - 1;
	if (leftrank < 1) leftrank = MPIsize - 1;
	rightrank = myrank + 1;
	if (rightrank >= MPIsize) rightrank = 1;

	if (MSG_INFO) cout << "I'm worker " << myrank << " and my neighbors are " << leftrank << " and " << rightrank << endl;

	   /* RECEIVER CODE */ 

	// Receive global parameters
	broadcastParametersMPI();

	allocateMemoryMPI();

	initializeVelocitiesAndCoefficients();

	// At this point we start a big loop
	// in which we check if there are new messages.
	// If yes, then process them. If not, do some other work
	// and check later again.

	bool dowork = true;
	int startsimulation = 0; // has to increase to 2 to start simulation
	bool simulationendreached = false;

	int flag;
	MPI_Status status;
	MPI_Status status1;
	int pos;  // internal position (0 to NUMT-1)
	int time; // time step
	int part;  // part of calculation (1 or 2)
	SimulationSlice *pslice;
	SimulationSlice *sscur;
	SimulationSlice *ssprev;
	
	FIFOBuffer<SendInfo> sendbuffer(NUMT * 3);  // should be enough space
//	FIFOBuffer<SendInfo> receivebuffer(NUMT * 3);
	int sendstate = STATE_NOTSENDING;
	int receivestate = STATE_NOTRECEIVING;
	SendInfo currentsend;
	SendInfo sinfo; // temporary send info

	MPI_Request sendrequest;    // Handle for asynchronous send
	MPI_Request receiverequest; // & receive

	// Variables for dynamic work load:
	int width; // temporary variable for the number of simulation slices

	int workslice;
	
	double eps = EPS;
	float timefloat = 0; // original 'time' from solve.cpp

	int itime;  // time - 2

	double idletimer = MPI_Wtime();
	bool idlestate = false;

	statisticsVariables stats;
	int currentmaximumtime = 0; // current max time for statistics.

	// int testdata = 100*myrank; // for testing purposes

	stats.start(TIMER_TOTAL);

	if (MSG_MESSAGE) cout << myrank << ": Start to work" << endl;
	while (dowork)
	{
		stats.start(TIMER_BLOCK_MESSAGES);
		// cout << myrank << ".";
		// Probe for messages from anybody:
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
		while (flag)
		{
			if (MSG_MESSAGE) cout << myrank << ": New message from " << status.MPI_SOURCE << ", TAG: " << status.MPI_TAG << endl;
			switch(status.MPI_TAG)
			{

			case TAG_DATA:
				// Receive the data set:
				receiveBufferFromNode(
					status.MPI_SOURCE,  // receive from manager or worker
					&pos, // position
					&time,  // Initial time.
					&part
					);
				if (MSG_SENDDATA) cout << myrank << ": Receiving data from " << status.MPI_SOURCE
					<< " at pos " << pos 
					<< ", time " << time
					<< ", part " << part << endl;
				if (MSG_STATEDUMP) dumptimestate();

				// Unpack
				pslice = GetNewSimulationSlice();
				pslice->MPIUnpack(part);

				// Small error check
				// (can be removed if code works correctly:
				if (pos != positivemod(pos, NUMT))
				{
					// sender is supposed to send valid position!
					cout << myrank << ": WARNING: worker(): " 
						<< " Received invalid position " << pos
						<< " from node " << status.MPI_SOURCE << endl;
				};
				
				// put this into right position
				// Check that the new slice is one time-step 
				// ahead than the one we currently have:
				if (time == sliceinfo[pos].time + 1)
				{
					// OK, we get a new piece of slice

					// previous slice is of no use -> garbage can.
					DisposeSimulationSlice(sliceinfo[pos].previousslice);
					
					// current slice gets old
					sliceinfo[pos].previousslice = sliceinfo[pos].currentslice;
					sliceinfo[pos].previouspart = sliceinfo[pos].currentpart;
					sliceinfo[pos].previousblock = sliceinfo[pos].currentblock;
				
					// and new slice gets the current one
					sliceinfo[pos].currentslice = pslice;
					sliceinfo[pos].currentpart = part; 
					sliceinfo[pos].currentblock = 0;
					sliceinfo[pos].time = time; // increment time
				}
				else
				{
					// Uh oh.. are we running out of sync?
					cout << myrank << ": WARNING: worker() "
						<< ": Expected slice at position " << pos 
						<< " with time " << sliceinfo[pos].time + 1
						<< " but received time " << time << endl;
					dumptimestate();
				};

				// Adjust pointers to first and last slice
				// CHECK: is the new slice a neighbor of the existing slices?
				if (firstslice == -1)
				{
					assert(lastslice == -1);

					// This is the first slice ever:
					firstslice = pos;
					lastslice = pos;
				}
				else if (pos == positivemod(firstslice - 1, NUMT))
				{
					// add to left
					firstslice = pos;
				}
				else if (pos == positivemod(lastslice + 1, NUMT))
				{
					// add to right
					lastslice = pos;
				}
				else if ((pos == firstslice) || (pos == lastslice))
				{
					// Ok, we received an update of an existing slice.
				}
				else
				{
					// we have received a slice that is neither
					// a left neighbor nor a right neighbor of
					// existing slices!					
					cout << myrank << ": WARNING: worker() "
						<< ": Received slice at position " << pos 
						<< " which is neither neighbor of " << firstslice
						<< " nor neighbor of " << lastslice << endl;
				};

				// cout << myrank << ": Firstslice: " << firstslice << " Lastslice: " << lastslice << endl;

				// just for test: we put our message to newsx
				// slice.linkGlobalLowerCase(newsx);
			   	//slice.MPIUnpack();
				// cout << myrank << ": Receiving data from manager (END). " << endl;

				if (MSG_STATEDUMP) dumptimestate();

				break;
			case TAG_RESETFIRSTSLICEPOINTER:
				// special treatment for last node
				assert(myrank == MPIsize - 1); // only applies to last node

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				firstslice = -1;
				lastslice = -1;
				break;
			case TAG_INCREMENTLASTSLICEPOINTER:
				// special treatment for last node
				assert(myrank == MPIsize - 1); // only applies to last node
				assert(lastslice == NUMT - 1); // current position should be on last slice.

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				lastslice = 0; // has been sent at the very beginning
				break;
			case TAG_RESETSLICEPOINTERS:
				// to be used when loading data dump

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				firstslice = -1;
				lastslice = -1;
				break;
			case TAG_DIE:
				// We finish all work
				dowork = false;

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				if (MSG_MESSAGE) cout << myrank << ": I finished all work." << endl;
				break;
			case TAG_ALLSLICESSENT:
				if (MSG_MESSAGE) cout << myrank << ": I'm ready to start!" << endl;

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				if (MSG_STATEDUMP) dumptimestate();

				// Send back to manager that we are ready:
				MPI_Send(NULL, 0, MPI_INT,
				   0, // dest
				   TAG_ALLSLICESRECEIVED, // tag
				   MPI_COMM_WORLD);

				break;
			case TAG_STARTSIMULATION:
				if (MSG_MESSAGE) cout << myrank << ": I start simulation!" << endl;

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);
				
				if (MSG_STATEDUMP) dumptimestate();

				startsimulation++;  // Let's go!
				break;

			case TAG_SETTIME:
				if (MSG_MESSAGE) cout << myrank << ": I receive new starting time" << endl;

				// Acknowledge message
			    MPI_Recv(&sinfo.time, 1, MPI_INT, 
					status.MPI_SOURCE, 
					TAG_SETTIME, MPI_COMM_WORLD, &status1);
				
				datadumpinitialtime = sinfo.time + 1; // global variable to prevent immediate saving of loaded dump data.
				for (int i=0; i<NUMT; i++)
				{
					sliceinfo[i].time = sinfo.time;
				};
				break;

			case TAG_TIMEOUT:
				if (MSG_MESSAGE) cout << myrank << ": I receive time-out" << endl;

				// Acknowledge message
			    MPI_Recv(&sinfo.time, 1, MPI_INT, 
					status.MPI_SOURCE, 
					TAG_TIMEOUT, MPI_COMM_WORLD, &status1);
				
				STEPS = sinfo.time; // Set limit for number of steps.
				CHECKPOINTUPDATE = STEPS - 3; // Set new checkpoint to start saving.
				break;

			case TAG_WANTTOSEND:
				// A node wants to send a data block to us:

				if (MSG_MESSAGE) cout << myrank << ": received TAG_WANTTOSEND from " << status.MPI_SOURCE << endl;

				// Receive message
			    MPI_Recv(&sinfo.pos, 4, MPI_INT, 
					status.MPI_SOURCE, 
					TAG_WANTTOSEND, MPI_COMM_WORLD, &status1);

				if ((receivestate == STATE_NOTRECEIVING) // I'm currently free
					&& ((sliceinfo[sinfo.pos].previousblock == 0) // and the slice to be deleted is not blocked for pending sending
					    ||(sliceinfo[sinfo.pos].time >= sinfo.time))) // unless we don't expect to advance
				{
					// Great, I like to have the data:
					if (MSG_MESSAGE) cout << myrank << ": TAG_WANTTOSEND: Accepted. Please send. " << endl;

					stats.start(TIMER_RECEIVE);

					// We prepare for receiving data 
					// asynchronously: First receive, then send:
					MPI_Irecv( pBufferAllreceive, 
						  maxBufferAllreceive, 
						  MPI_PACKED, 
						  status.MPI_SOURCE, // rank of source
						  TAG_DATABLOCK, // message tag
						  MPI_COMM_WORLD, // communicator (handle)
						  &receiverequest);

					// Now we send back that the other party can start
					// to send data:
					MPI_Bsend(NULL, 0, MPI_INT,
					   status.MPI_SOURCE, // dest
					   TAG_PLEASESEND, // tag
					   MPI_COMM_WORLD);

					// cout << myrank << ": TAG_WANTTOSEND: send PLEASESEND " << endl;

					receivestate = STATE_RECEIVING;
				}
				else
				{
					// Sorry, currently I have somehing else to do
					if (MSG_MESSAGE) cout << myrank << ": TAG_WANTTOSEND: Failed. I can not accept right now. " << endl;

					MPI_Bsend(NULL, 0, MPI_INT,
					   status.MPI_SOURCE, // dest
					   TAG_DONOTSEND, // tag
					   MPI_COMM_WORLD);
				};
				break;

			case TAG_DONOTSEND:
				// We have to cancel the send procedure,
				// because the other node is not ready to
				// receive:
				if (MSG_SENDBUFFER) cout << myrank << ": received TAG_DONOTSEND. Ok, I cancel my request.";

				assert(sendstate == STATE_WAITINGTOSEND); // only makes sense if we have been waiting to send
				assert(currentsend.node == status.MPI_SOURCE); // and the message should only come from the node that we want to send the data to.

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				//For now, push data to the back of the stack for later sending
				if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				sendbuffer.push(currentsend);
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				sendstate = STATE_NOTSENDING;
				break;

			case TAG_PLEASESEND:
				// Ok, now we can send the data:
				if (MSG_MESSAGE) cout << myrank << ": received TAG_PLEASESEND"
					<< " from " << status.MPI_SOURCE << endl;
				assert(sendstate == STATE_WAITINGTOSEND); // only makes sense if we have been waiting to send
				assert(currentsend.node == status.MPI_SOURCE); // and the message should only come from the node that we want to send the data to.

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				// So, start the asynchronous send:
				if (MSG_SENDDATA) cout << myrank << ": I start asynchronous send: "
					<< "to node " << currentsend.node
					<< " at pos " << currentsend.pos
					<< " and time " << currentsend.time 
					<< " and part " << currentsend.part << endl;

				// After we packed the data, we can decrease the 
				// block on this slice (i.e. free it)
				pos = currentsend.pos;
				if (sliceinfo[pos].time == currentsend.time)
				{
					pslice = sliceinfo[pos].currentslice;

					sliceinfo[pos].currentblock --;
					if (sliceinfo[pos].currentblock < 0)
					{
						cout << myrank << ": WARNING: worker(): A currentblock decreased below 0" << endl;
					};
				}
				else if (sliceinfo[pos].time - 1 == currentsend.time)
				{
					pslice = sliceinfo[pos].previousslice;

					sliceinfo[pos].previousblock --;
					if (sliceinfo[pos].previousblock < 0)
					{
						cout << myrank << ": WARNING: worker(): A previousblock decreased below 0" << endl;
					};
				}
				else
				{
					cout << myrank << ": WARNING!! TAG_PLEASESEND: OUT OF SYNC?" 
						<< " currentsend.time: " << currentsend.time 
						<< ", sliceinfo[currentsend.pos] " << sliceinfo[currentsend.pos].time << endl;
					pslice = NULL; // set anything to avoid compiler warning, but the program will crash soon anyway.
				};

				//pslice->w[0][1][1][1].a = (float) testdata++;//rand();
				//cout << myrank << ": Isending to node " << currentsend.node
				//	<< ", pos " << currentsend.pos 
				//	<< ", time " << currentsend.time 
				//	<< ", create rand(): " << pslice->w[0][1][1][1].a << endl;

				stats.start(TIMER_PACK);
				// cout << myrank << ": MPIPack()" << endl;
				pslice->MPIPack(currentsend.part);
				stats.stop(TIMER_PACK);

				stats.start(TIMER_SEND);
				// cout << myrank << ": IsendBufferToNode()" << endl;
				IsendBufferToNode(
					currentsend.node,  // node we send this to.
					currentsend.pos, // position
					currentsend.time,  // Initial time.
					currentsend.part,
					&sendrequest
				); 

				// cout << myrank << ": IsendBufferToNode() done" << endl;
				sendstate = STATE_SENDING;

				break;

				// ******* Dynamic work load related tags *******

			case TAG_BM_PLEASESENDTWO_TOLEFT:
				//if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_PLEASESENDTWO " << endl;

				// Acknowledge message
			    MPI_Recv(&time, 1, MPI_INT, status.MPI_SOURCE, 
					TAG_BM_PLEASESENDTWO_TOLEFT, MPI_COMM_WORLD, &status1);

				// Received notification for border move.
				// If both send this tag at the same time, then the 
				// border is not moved.
				if (status.MPI_SOURCE != leftrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_PLEASESENDTWO_TOLEFT: Received message from "
						<< status.MPI_SOURCE << " which is not my left neighbor!" << endl;
				};

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_PLEASESENDTWO_TOLEFT from left"
					<< " at time " << time << endl;
				if (MSG_STATEDUMP) dumptimestate();
				// Received this from leftrank, so I am the right rank with priority.
				switch (dynamicloadstateleft)
				{
				case STATE_BM_NOTWAITING:
					if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: I have not been waiting." << endl;
					// Ok, I have to admit, left really wins.
					// Now I will send the two slices to left and 
					// move my left border back by one.

					if ((bordermoveleft != 0) || (bordermoveright > 0))
					{
						cout << myrank << ": WARNING: TAG_BM_PLEASESENDTWO_TOLEFT: bordermove inconsistent!"
							<< " left: " << bordermoveleft << ", right " << bordermoveright << endl;
						assert(false);
					};

					// But first, let me check what my current width is:
					// (including the coming bordermove)
					width = positivemod(lastslice - 1 + bordermoveright - (firstslice + 1 + bordermoveleft), NUMT);
					if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: width: " << width << endl;
					
					workslice = positivemod(firstslice + 1, NUMT);

					// Either the width is not enough, or I have
					// already finished and sent you my border slice:
					if (width < 3)
					{
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: I don't have enough space left. I can only send one." << endl;
						// I don't have enough space to further decrease
						// my size. I want to keep at least 1 slice!
						
						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TOLEFT, // tag
						   MPI_COMM_WORLD);

						// So, let us proceed as the border did not move:

						// Since I have not calculated my border slice yet,
						// there is nothing more for me to do now.
					}
					else if ((sliceinfo[workslice].time > time)
						     || ((sliceinfo[workslice].time == time)
						     && (sliceinfo[workslice].currentpart == ALL_PARTS)))
					{
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: I have already sent you the part, so I only send one." << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TOLEFT, // tag
						   MPI_COMM_WORLD);

						// So, let us proceed as the border did not move:

						// Since I have not calculated my border slice yet,
						// there is nothing more for me to do now.
					
					}
					else if ((sliceinfo[workslice].currentblock > 0)
						|| (sliceinfo[workslice].previousblock > 0)
						|| (sliceinfo[workslice].simulationblock > 0))
					{
						// Element is blocked, sorry, I can't send it to you
						// right now.
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: Element is blocked, so I only send one." << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TOLEFT, // tag
						   MPI_COMM_WORLD);
					}
					else
					{
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: Ok, I can send two slices to you." << endl;
						// Ok, I have enough slices left. I can give you 
						// one, so that you can calculate more, and 
						// I have to calculate less:

						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: Before send."
							<< " to " << status.MPI_SOURCE << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGTWO_TOLEFT, // tag
						   MPI_COMM_WORLD);

						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: After send." << endl;

						// You were first, I have not calculated my 
						// border cell yet. Therefore, I can not send
						// you any data immediately.

						// But I can set a flag, that as soon as I finish
						// calculating my border cell, I will move
						// the border closer to me by 1 step.
						bordermoveleft = +1;
					};
					break;
				case STATE_BM_WAITINGFORRESPONSE:
					if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: I had also claimed to win." << endl;
					// Obviously both of us have sent the same request
					// at the same time, so the border is not moved.

					// I can assume, that you receive my TAG_BM_PLEASESENDTWO
					// request while you are also in the waiting state.
					
					// I just send the last slice I finished:
					workslice = positivemod(firstslice + 1, NUMT);

					// Set the flag for completion of the current part
					sliceinfo[workslice].currentpart = PART1 | PART2;
					sliceinfo[workslice].simulationblock--;
					if (sliceinfo[workslice].simulationblock != 0)
					{
						cout << myrank << ": WARNING: worker(): TAG_BM_PLEASESENDTWO_TOLEFT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
					};

					// Send data to left neighbor
					// Put this into the sendbuffer:
					sinfo.pos = workslice; // what to send
					sinfo.node = leftrank; // node to send it to
					sinfo.time = sliceinfo[workslice].time;
					sinfo.part = PART2;
					
					if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
					if (MSG_SENDBUFFER) sendbuffer.dumpstate();
					sendbuffer.push(sinfo);
					if (MSG_SENDBUFFER) sendbuffer.dumpstate();

					sliceinfo[workslice].currentblock++; // block the slice during send

					dynamicloadstateleft = STATE_BM_NOTWAITING;
					break;
				default:
					cout << myrank << ": WARNING: TAG_BM_PLEASESENDTWO: Unexpected state for dynamicloadstateleft " << dynamicloadstateleft << endl;
				};
				if (MSG_STATEDUMP) dumptimestate();
				break;
			case TAG_BM_PLEASESENDTWO_TORIGHT:
				//if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_PLEASESENDTWO " << endl;

				// Acknowledge message
			    MPI_Recv(&time, 1, MPI_INT, status.MPI_SOURCE, 
					TAG_BM_PLEASESENDTWO_TORIGHT, MPI_COMM_WORLD, &status1);

				// Received notification for border move.
				// If both send this tag at the same time, then the 
				// border is not moved.
				if (status.MPI_SOURCE != rightrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_PLEASESENDTWO_TORIGHT: Received message from "
						<< status.MPI_SOURCE << " which is not my right neighbor!" << endl;
				};

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_PLEASESENDTWO_TORIGHT from right, "
					<< " at time " << time << endl;
				if (MSG_STATEDUMP) dumptimestate();
				// Received this from rightrank.
				switch (dynamicloadstateright)
				{
				case STATE_BM_NOTWAITING:
					if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TORIGHT: I have not been waiting." << endl;
					// Ok, you have been first. I send two slices to you
					// and move my border back by one.

					// Now I will send the two slices to left and 
					// move my left border back by one.

					if ((bordermoveright != 0) || (bordermoveleft < 0))
					{
						cout << myrank << ": WARNING:  TAG_BM_PLEASESENDTWO_TORIGHT: bordermove inconsistent!"
							<< " left: " << bordermoveleft << ", right " << bordermoveright << endl;
					};

					// But first, let me check what my current width is:
					// (including the coming bordermove)
					width = positivemod(lastslice - 1 + bordermoveright - (firstslice + 1 + bordermoveleft), NUMT);
					if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TORIGHT: width: " << width << endl;

					workslice = positivemod(firstslice + 1, NUMT);
					// Either the width is not enough, or I have
					// already finished and sent you my border slice:
					
					if (width < 2)
					{
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TORIGHT: I don't have enough space left. I can only send one." << endl;
						// I don't have enough space to further decrease
						// my size. I want to keep at least 1 slice!
						
						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TORIGHT, // tag
						   MPI_COMM_WORLD);

						// So, let us proceed as the border did not move:

						// Since I have not calculated my border slice yet,
						// there is nothing more for me to do now.
					}
					else if ((sliceinfo[workslice].time > time)
						     || ((sliceinfo[workslice].time == time)
						     && (sliceinfo[workslice].currentpart == ALL_PARTS)))
					{
						if (MSG_DYNAMIC) cout << myrank << ":  TAG_BM_PLEASESENDTWO_TORIGHT: I have already sent you the part, so I only send one." << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TORIGHT, // tag
						   MPI_COMM_WORLD);

						// So, let us proceed as the border did not move:

						// Since I have not calculated my border slice yet,
						// there is nothing more for me to do now.
					}
					else if ((sliceinfo[workslice].currentblock > 0)
						|| (sliceinfo[workslice].previousblock > 0)
						|| (sliceinfo[workslice].simulationblock > 0))
					{
						// Element is blocked, sorry, I can't send it to you
						// right now.
						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TAG_BM_SENDINGONE_TORIGHT: Element is blocked, so I only send one." << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGONE_TORIGHT, // tag
						   MPI_COMM_WORLD);
					}
					else
					{
						if (MSG_DYNAMIC) cout << myrank << ":  TAG_BM_PLEASESENDTWO_TORIGHT: Ok, I can send two slices to you." << endl;

						// Ok, I have enough slices left. I can give you 
						// one, so that you can calculate more, and 
						// I have to calculate less:

						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: Before send." << endl;

						MPI_Bsend(NULL, 0, MPI_INT,
						   status.MPI_SOURCE, // dest
						   TAG_BM_SENDINGTWO_TORIGHT, // tag
						   MPI_COMM_WORLD);

						if (MSG_DYNAMIC) cout << myrank << ": TAG_BM_PLEASESENDTWO_TOLEFT: After send." << endl;

						// You were first, I have not calculated my 
						// border cell yet. Therefore, I can not send
						// you any data immediately.

						// But I can set a flag, that as soon as I finish
						// calculating my border cell, I will move
						// the border closer to me by 1 step.
						bordermoveright = -1;
					};
					break;
				case STATE_BM_WAITINGFORRESPONSE:
					if (MSG_DYNAMIC) cout << myrank << ":  TAG_BM_PLEASESENDTWO_TORIGHT: I had also claimed to win." << endl;
					// Obviously both of us have sent the same request
					// at the same time, so the border is not moved.

					// I can assume, that you receive my TAG_BM_PLEASESENDTWO
					// request while you are also in the waiting state.
					
					// I just send the last slice I finished:
					workslice = positivemod(lastslice - 1, NUMT);

					// Set the flag for completion of the current part
					sliceinfo[workslice].currentpart = PART1 | PART2;
					sliceinfo[workslice].simulationblock--;
					if (sliceinfo[workslice].simulationblock != 0)
					{
						cout << myrank << ": WARNING: worker(): TAG_BM_PLEASESENDTWO_TORIGHT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
					};

					// Send data to right neighbor
					// Put this into the sendbuffer:
					sinfo.pos = workslice; // what to send
					sinfo.node = rightrank; // node to send it to
					sinfo.time = sliceinfo[workslice].time;
					sinfo.part = PART2;
					
					if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
					if (MSG_SENDBUFFER) sendbuffer.dumpstate();
					sendbuffer.push(sinfo);
					if (MSG_SENDBUFFER) sendbuffer.dumpstate();
					sliceinfo[workslice].currentblock++; // block the slice during send

					dynamicloadstateright = STATE_BM_NOTWAITING;
					
					break;
				default:
					cout << myrank << ": WARNING:  TAG_BM_PLEASESENDTWO_TORIGHT: Unexpected state for dynamicloadstateright " << dynamicloadstateright << endl;
				};
				if (MSG_STATEDUMP) dumptimestate();
				break;
			case TAG_BM_SENDINGONE_TORIGHT:
				// The other party did not accept my request for sending two
				// (Probably he has not enough slices left)
				// So we proceed as usual, you send one, and I send one.

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);

				if (status.MPI_SOURCE != leftrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TORIGHT: Received message from "
						<< status.MPI_SOURCE << " which is not my left neighbor!" << endl;
				};

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_SENDINGONE_TORIGHT from left." << endl;
				if (MSG_STATEDUMP) dumptimestate();
				// In this case, we proceed as if the border
				// did not move: I send one to you, you send one to me:

				if (dynamicloadstateleft != STATE_BM_WAITINGFORRESPONSE)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TORIGHT: unexpected state dynamicloadstateleft: " << dynamicloadstateleft << endl;
				};
				
				// I just send the last slice I finished:
				workslice = positivemod(firstslice + 1, NUMT);

				// Set the flag for completion of the current part
				sliceinfo[workslice].currentpart = PART1 | PART2;
				sliceinfo[workslice].simulationblock--;
				if (sliceinfo[workslice].simulationblock != 0)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TORIGHT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
				};


				// Send data to left neighbor
				// Put this into the sendbuffer:
				sinfo.pos = workslice; // what to send
				sinfo.node = leftrank; // node to send it to
				sinfo.time = sliceinfo[workslice].time;
				sinfo.part = PART2;
				
				if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				sendbuffer.push(sinfo);
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();

				sliceinfo[workslice].currentblock++; // block the slice during send

				dynamicloadstateleft = STATE_BM_NOTWAITING;
				if (MSG_STATEDUMP) dumptimestate();					
				break;
			case TAG_BM_SENDINGONE_TOLEFT:
				// The other party did not accept my request for sending two
				// (Probably he has not enough slices left)
				// So we proceed as usual, you send one, and I send one.

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);
				if (status.MPI_SOURCE != rightrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TOLEFT: Received message from "
						<< status.MPI_SOURCE << " which is not my right neighbor!" << endl;
				};

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_SENDINGONE_TOLEFT from right." << endl;
				if (MSG_STATEDUMP) dumptimestate();
				// In this case, we proceed as if the border
				// did not move: I send one to you, you send one to me:

				if (dynamicloadstateright != STATE_BM_WAITINGFORRESPONSE)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TOLEFT: unexpected state dynamicloadstateright: " << dynamicloadstateright << endl;
				};

				// I just send the last slice I finished:
				workslice = positivemod(lastslice - 1, NUMT);

				// Set the flag for completion of the current part
				sliceinfo[workslice].currentpart = PART1 | PART2;
				sliceinfo[workslice].simulationblock--;
				if (sliceinfo[workslice].simulationblock != 0)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGONE_TOLEFT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
				};

				// Send data to right neighbor
				// Put this into the sendbuffer:
				sinfo.pos = workslice; // what to send
				sinfo.node = rightrank; // node to send it to
				sinfo.time = sliceinfo[workslice].time;
				sinfo.part = PART2;
				
				if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				sendbuffer.push(sinfo);
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				sliceinfo[workslice].currentblock++; // block the slice during send

				dynamicloadstateright = STATE_BM_NOTWAITING;
				if (MSG_STATEDUMP) dumptimestate();				
				break;
			case TAG_BM_SENDINGTWO_TORIGHT:
				// Wow! The other party really accepted that I have won,
				// and he can give me more work.

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_SENDINGTWO_TORIGHT from left." << endl;
				if (MSG_STATEDUMP) dumptimestate();

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);


				// Since you have to send two slices, I don't have to send
				// anything now.

				// It suffices, if I mark my last border slice as completed
				// and then move the border accordingly.

				// Since I have a new border element, I have to update
				// the timing there properly.
				if (status.MPI_SOURCE != leftrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TORIGHT: Received message from "
						<< status.MPI_SOURCE << " which is not my left neighbor!" << endl;
				};

				if (dynamicloadstateleft != STATE_BM_WAITINGFORRESPONSE)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TORIGHT: unexpected state dynamicloadstateleft: " << dynamicloadstateleft << endl;
				};
				
				// Mark the last border slice as completed:
				workslice = positivemod(firstslice + 1, NUMT);

				// Set the flag for completion of the current part
				sliceinfo[workslice].currentpart = PART1 | PART2;
				sliceinfo[workslice].simulationblock--;
				if (sliceinfo[workslice].simulationblock != 0)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TORIGHT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
				};

				// Move the border to the left
				firstslice = positivemod(firstslice - 1, NUMT);

				// Fill the new empty slice with preliminary information:
				sliceinfo[firstslice].time = sliceinfo[workslice].time - 1;
				sliceinfo[firstslice].currentpart = 0;
				sliceinfo[firstslice].currentblock = 0;
				sliceinfo[firstslice].previouspart = 0;
				sliceinfo[firstslice].previousblock = 0;
				assert( sliceinfo[firstslice].currentslice == NULL ); // should be free!
				assert( sliceinfo[firstslice].previousslice == NULL );

				// This will be filled properly by the neighbor as soon
				// as he is done.

				// There might be concerns that we will be tempted to
				// continue the calculation at the new border cell
				// (firstslice + 1), resulting in calculation done
				// twice, but this can not happen, before our
				// neighbor sent both (firstslice+1) and (firstslice).

				dynamicloadstateleft = STATE_BM_NOTWAITING;
				if (MSG_STATEDUMP) dumptimestate();
				break;
			case TAG_BM_SENDINGTWO_TOLEFT:
				// Wow! The other party really accepted that I have won,
				// and he can give me more work.

				if (MSG_DYNAMIC) cout << myrank << ": receive TAG_BM_SENDINGTWO_TOLEFT from right." << endl;
				if (MSG_STATEDUMP) dumptimestate();

				// Acknowledge message
			    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, &status1);


				// Since you have to send two slices, I don't have to send
				// anything now.

				// It suffices, if I mark my last border slice as completed
				// and then move the border accordingly.

				// Since I have a new border element, I have to update
				// the timing there properly.

				if (status.MPI_SOURCE != rightrank)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TOLEFT: Received message from "
						<< status.MPI_SOURCE << " which is not my right neighbor!" << endl;
				};

				if (dynamicloadstateright != STATE_BM_WAITINGFORRESPONSE)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TOLEFT: unexpected state dynamicloadstateright: " << dynamicloadstateright << endl;
				};

				// Mark the last border slice as completed:
				workslice = positivemod(lastslice - 1, NUMT);

				// Set the flag for completion of the current part
				sliceinfo[workslice].currentpart = PART1 | PART2;
				sliceinfo[workslice].simulationblock--;
				if (sliceinfo[workslice].simulationblock != 0)
				{
					cout << myrank << ": WARNING: worker(): TAG_BM_SENDINGTWO_TOLEFT: inconsistent slice pointer: " << sliceinfo[workslice].simulationblock << endl;
				};

				// Move the border to the right
				lastslice = positivemod(lastslice + 1, NUMT);

				// Fill the new empty slice with preliminary information:
				sliceinfo[lastslice].time = sliceinfo[workslice].time - 1;
				sliceinfo[lastslice].currentpart = 0;
				sliceinfo[lastslice].currentblock = 0;
				sliceinfo[lastslice].previouspart = 0;
				sliceinfo[lastslice].previousblock = 0;
				assert( sliceinfo[lastslice].currentslice == NULL ); // should be free!
				assert( sliceinfo[lastslice].previousslice == NULL );

				// This will be filled properly by the neighbor as soon
				// as he is done.

				// There might be concerns that we will be tempted to
				// continue the calculation at the new border cell
				// (lastslice - 1), resulting in calculation done
				// twice, but this can not happen, before our
				// neighbor sent both (lastslice-1) and (lastslice).

				dynamicloadstateright = STATE_BM_NOTWAITING;
				if (MSG_STATEDUMP) dumptimestate();
				break;
			default:
				// Completely unexpected message!
				cout << myrank << ": WARNING: worker(): Unexpected message tag " << status.MPI_TAG 
					<< " sent from node " << status.MPI_SOURCE << endl;
			}; // switch (status.MPI_TAG)

			// more messages waiting?
			MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);

		}; // while (flag)
		stats.stop(TIMER_BLOCK_MESSAGES);

		stats.start(TIMER_BLOCK_SENDRECEIVE);
		// Now let us check if we can send or receive big blocks of data
		// (this is done asynchronously)
		switch(sendstate)
		{
		case STATE_SENDING:
			MPI_Test(&sendrequest, &flag, &status);
			//cout << myrank << ": send flag " << flag << endl;

			if (! flag)
			{
				// Not finished asynchronous communication yet.
				stats.check(TIMER_SEND);
				break; // Exit the switch() command.
			}
			else
			{
				// Finished sending.
				// We can free all resources and send the next.
				if (MSG_SENDDATA) cout << myrank << ": Isend to node " << currentsend.node
					<< ", pos " << currentsend.pos 
					<< ", time " << currentsend.time 
					<< ", part " << currentsend.part 
					<< " finished." << endl;

				stats.stop(TIMER_SEND);
				sendstate = STATE_NOTSENDING;
			};
			// NO break;  // we fall through to the next block on purpose!
		case STATE_NOTSENDING:
			if (!sendbuffer.empty())
			{
				// We can send something new
				if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.pull():" << endl;
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();
				currentsend = sendbuffer.pull();
				if (MSG_SENDBUFFER) sendbuffer.dumpstate();

				if (MSG_SENDDATA) cout << myrank << ": Isend request to node " << currentsend.node
					<< ", pos " << currentsend.pos 
					<< ", time " << currentsend.time 
					<< ", part " << currentsend.part << endl;
				
				// Send to node that we want to send something:
				// That node should prepare receive:
				MPI_Bsend(&currentsend.pos, 4, MPI_INT,
				   currentsend.node, // dest
				   TAG_WANTTOSEND, // tag
				   MPI_COMM_WORLD);

				sendstate = STATE_WAITINGTOSEND;
			};
			break;
		case STATE_WAITINGTOSEND:
			// Nothing we can do right now.
			break;
		default:
			assert(false); // Should never happen
			break;
		}; // switch(sendstate)

		switch(receivestate)
		{
		case STATE_RECEIVING:
			MPI_Test(&receiverequest, &flag, &status);
			//MPI_Wait(&receiverequest, &status); flag = true;
			//cout << myrank << ": receive flag " << flag << endl;

			if (! flag)
			{
				// Not finished asynchronous communication yet.
				stats.check(TIMER_RECEIVE);
				break; // Exit the switch() command.
			}
			else
			{
				// Finished receiving.
				// Receive the data set:
				IreceiveBufferFromNode(
					status.MPI_SOURCE,  // receive from manager or worker
					&pos, // position
					&time,  // Initial time.
					&part,
					&status
					);
				stats.stop(TIMER_RECEIVE);
				if (MSG_SENDDATA) cout << myrank << ": ==> IReceiving data from " << status.MPI_SOURCE
					<< " at pos " << pos 
					<< ", time " << time 
					<< ", part " << part << endl;
				if (MSG_STATEDUMP) dumptimestate();

				// We unpack the data below...

				//cout << myrank << ": ==> IReceiving data from " << status.MPI_SOURCE
				//	<< " at pos " << pos 
				//	<< ", time " << time 
				//	<< ", rand():" << pslice->w[0][1][1][1].a << endl;


				// Small error check
				// (can be removed if code works correctly:
				if (pos != positivemod(pos, NUMT))
				{
					// sender is supposed to send valid position!
					cout << myrank << ": WARNING: worker(): " 
						<< " Received invalid position " << pos
						<< " from node " << status.MPI_SOURCE << endl;
				};
				
				// put this into right position
				// Check that the new slice is one time-step 
				// ahead than the one we currently have:
				if (time == sliceinfo[pos].time + 1)
				{
					if (MSG_SENDDATA) cout << myrank << ": New slice at time+1" << endl;

					// OK, we get a new piece of slice

					// quick check:
					if (sliceinfo[pos].previousblock > 0)
					{
						// should not happen
						cout << myrank << ": WARNING: worker(): STATE_RECEIVING: Previous slice at position " << pos
							<< ", time " << time
							<< " is still blocked with " << sliceinfo[pos].previousblock
							<< " but we have received a new slice!" << endl;
						dumptimestate();
					};
					if (OPTION_MEMORY_DROP_OW_OJ)
					{
						// Instead of disposing SimulationSlice,
						// we move it to the new position.
						// In this way, we get a free copy of the old w
						// to the new w. 
						// The new w will be called ow and W and will be
						// overwritten automatically.

						// save previous slice with old w field:
						pslice = sliceinfo[pos].previousslice;
						//cout << myrank << ": Old slice pointer: " << (int) pslice << endl;
						
						// current slice gets old
						sliceinfo[pos].previousslice = sliceinfo[pos].currentslice;
						sliceinfo[pos].previouspart = sliceinfo[pos].currentpart;
						sliceinfo[pos].previousblock = sliceinfo[pos].currentblock;

						// Use the old slice as current slice.
						// (this contains the w field, which will be linked 
						//  to both, ow, and W)

						if (pslice == NULL)
						{
							// an old slice did not exist yet:
							// We have to create a new one:
							pslice = GetNewSimulationSlice();

							// and copy manually the w and j fields into the new slice:
							// NOTE that ow from the previous slice is stored in the new slice as w
							sscur = pslice;
							ssprev = sliceinfo[pos].previousslice;
							//for (int sx=1;sx<=NUMx;sx++) 
								for (int sy=0;sy<=NUMT+1;sy++)
									for (int sz=0; sz<=NUM+1;sz++)
										for (int i=0;i<NW;i++)  
											sscur->w[0][sy][sz][i]=ssprev->w[0][sy][sz][i];

							// We also need to copy j from the previous slice to 
							// oj in the current slice:
							// NOTE that oj from the previous slice is stored in the new slice as j.
							//for (int sx=0;sx<=NUMx+1;sx++)
								for (int sy=0;sy<=NUMT+1;sy++)
									for (int sz=0;sz<=NUM+1;sz++)
										for (int i=0;i<3;i++) 
											sscur->j[0][sy][sz][i] = ssprev->j[0][sy][sz][i];
						};

						// Unpack
						stats.start(TIMER_UNPACK);
						pslice->MPIUnpack(part);
						stats.stop(TIMER_UNPACK);

						// and new slice gets the current one
						sliceinfo[pos].currentslice = pslice;
						sliceinfo[pos].time = time; // increment time
						sliceinfo[pos].currentpart = part;
						sliceinfo[pos].currentblock = 0;
					}
					else
					{
						// previous slice is of no use -> garbage can.
						DisposeSimulationSlice(sliceinfo[pos].previousslice);
						
						// current slice gets old
						sliceinfo[pos].previousslice = sliceinfo[pos].currentslice;
						sliceinfo[pos].previouspart = sliceinfo[pos].currentpart;
						sliceinfo[pos].previousblock = sliceinfo[pos].currentblock;
					
						// Unpack
						stats.start(TIMER_UNPACK);
						pslice = GetNewSimulationSlice();
						pslice->MPIUnpack(part);
						stats.stop(TIMER_UNPACK);

						// and new slice gets the current one
						sliceinfo[pos].currentslice = pslice;
						sliceinfo[pos].time = time; // increment time
						sliceinfo[pos].currentpart = part;
						sliceinfo[pos].currentblock = 0;
					};
				}
				else if (time == sliceinfo[pos].time)
				{
					if (MSG_SENDDATA) cout << myrank << ": New part for current slice" << endl;

					// OK, we get a new piece of slice at current position
					if (sliceinfo[pos].currentslice != NULL)
					{
						// Add to existing slice:
						pslice = sliceinfo[pos].currentslice;
					}
					else
					{
						// This should never happen!
						cout << myrank << ": WARNING: worker(): If the time exists, there should be a slice!" << endl;
						pslice = GetNewSimulationSlice();
					};

					// Unpack
					stats.start(TIMER_UNPACK);
					pslice->MPIUnpack(part);
					stats.stop(TIMER_UNPACK);

					// and new slice gets the current one
					sliceinfo[pos].currentslice = pslice;
					sliceinfo[pos].time = time; // increment time
					sliceinfo[pos].currentpart |= part; // bitwise OR: add parts
					// sliceinfo[pos].currentblock untouched
				}
				else if (time == sliceinfo[pos].time + 2)
				{
					if (MSG_SENDDATA) cout << myrank << ": New slice at time+2 +++++++++++++++" << endl;
					// Oh, due to asynchronous sending,
					// we received a slice first and 
					// skipped one.

					if ((sliceinfo[pos].previousblock > 0)
						|| (sliceinfo[pos].currentblock > 0))
					{
						// should not happen
						cout << myrank << ": WARNING: Previous slice or current slice at position " << pos
							<< " is still blocked with " << sliceinfo[pos].previousblock << " or " << sliceinfo[pos].previousblock
							<< " but we have received a new slice!" << endl;
					};


					// previous slice is of no use -> garbage can.
					DisposeSimulationSlice(sliceinfo[pos].previousslice);
					sliceinfo[pos].previousslice = NULL;
					sliceinfo[pos].previouspart = 0;
					sliceinfo[pos].previousblock = 0;

					// current slice is also old -> garbage can.
					DisposeSimulationSlice(sliceinfo[pos].currentslice);
					
					// Unpack
					stats.start(TIMER_UNPACK);
					pslice = GetNewSimulationSlice();
					pslice->MPIUnpack(part);
					stats.stop(TIMER_UNPACK);

					// and new slice gets the current one
					sliceinfo[pos].currentslice = pslice;
					sliceinfo[pos].time = time; // increment time
					sliceinfo[pos].currentpart = part;
					sliceinfo[pos].currentblock = 0;
				}
				else if (time == sliceinfo[pos].time - 1)
				{
					if (MSG_SENDDATA) cout << myrank << ": New slice or add to existing slice at time-1" << endl;
					// We fill in the slice into an existing place:

					if (sliceinfo[pos].previousslice != NULL)
					{
						// Add to existing slice:
						pslice = sliceinfo[pos].previousslice;
					}
					else
					{
						// Or create a new simulation slice:
						pslice = GetNewSimulationSlice();
					};
					
					// Unpack
					stats.start(TIMER_UNPACK);
					pslice->MPIUnpack(part);
					stats.stop(TIMER_UNPACK);
					
					// and new slice gets the current one
					sliceinfo[pos].previousslice = pslice;
					sliceinfo[pos].previouspart |= part; // bitwise OR with existing part

					// In this case no need to change time or block.
				}
				else
				{
					// Uh oh.. are we running out of sync?
					cout << myrank << ": WARNING: worker() "
						<< ": Expected slice at position " << pos 
						<< " with time " << sliceinfo[pos].time + 1
						<< " but received time " << time << endl;
					dumptimestate();
				};

				// Adjust pointers to first and last slice
				// CHECK: is the new slice a neighbor of the existing slices?
				if (firstslice == -1)
				{
					assert(lastslice == -1);

					// This is the first slice ever:
					firstslice = pos;
					lastslice = pos;
				}
				else if (pos == positivemod(firstslice - 1, NUMT))
				{
					// add to left
					firstslice = pos;
				}
				else if (pos == positivemod(lastslice + 1, NUMT))
				{
					// add to right
					lastslice = pos;
				}
				else if ((pos == firstslice) || (pos == lastslice))
				{
					// Ok, we received an update of an existing slice.
				}
				else if (OPTION_DYNAMICLOADBALANCING
					&& ((pos  == positivemod(firstslice + 1, NUMT))
					     || (pos == positivemod(lastslice - 1, NUMT))))
				{
					// Ok, this can happen, as we move around borders.
				}
				else
				{
					// we have received a slice that is neither
					// a left neighbor nor a right neighbor of
					// existing slices!					
					cout << myrank << ": WARNING: worker() "
						<< ": Received slice at position " << pos 
						<< " which is neither neighbor of " << firstslice
						<< " nor neighbor of " << lastslice << endl;
				};

				// just for test: we put our message to newsx
				// slice.linkGlobalLowerCase(newsx);
			   	//slice.MPIUnpack();
				// cout << myrank << ": Receiving data from manager (END). " << endl;

				if (MSG_STATEDUMP) dumptimestate();

				if (MSG_SENDDATA) cout << myrank << ": ==> IReceiving data from " << status.MPI_SOURCE
					<< " at pos " << pos 
					<< ", time " << time 
					<< ", part " << part 
					<< " finished. " << endl;
				// We can free all resources and send the next.
				receivestate = STATE_NOTRECEIVING;
			};

			/*
			if (!receivebuffer.empty())
			{
				// receive something
			};
			*/
			break;
		case STATE_NOTRECEIVING:
			// Nothing we can do right now...
			break;
		default:
			assert(false); // Should never happen
			break;
		}; // switch(receivestate)
		stats.stop(TIMER_BLOCK_SENDRECEIVE);

		// If there are no messages, let us do some work:

		stats.start(TIMER_BLOCK_WORK);
		if (startsimulation >= 1)
		{
			// First we determine which is the next slice that can do some work:
			// returns -1 if no work possible, otherwise number of slice
			// for which work can be done.
			stats.start(TIMER_WHICHWORK);
			workslice = whichwork();
			stats.stop(TIMER_WHICHWORK);

			// if (myrank == 1) workslice =-1;

			// We simulate the environment that is needed by solve.cpp
			int worksliceleft;
			int worksliceright;
			// we map this to sx running from 0 to NUMx+1 == 2
			int sx = 1; // don't forget the discrepancy in index.
			int sxleft = 0;
			int sxright = 2;
			// In this way, solve.cpp thinks it has a grid
			// with x-extent from 0 to 2.

			if (workslice >= 0)
			{
				if (idlestate)
				{
					// if we had been in idle state, we loose the idle state.
					idlestate = false;
					stats.stop(TIMER_IDLE);
				};
				assert(workslice != firstslice); // these are the borders that we never calculate.
				assert(workslice != lastslice);

				// Let us update slice 'workslice'
				// The following part is along the lines of Evolve()

				if (MSG_PROGRAM) cout << myrank << ": Evolve() at pos " << workslice << ", time " << sliceinfo[workslice].time << endl;

				// Do work according to the part we can do:
				switch (sliceinfo[workslice].currentpart)
				{
				case ALL_PARTS: // ---------------------------- ALL PARTS ----
					stats.start(TIMER_PART1);

					// We go one time step ahead:
					if (MSG_PROGRAM) cout << myrank << ": Simulate environment at pos " << workslice 
						<< ", time " << sliceinfo[workslice].time
						<< ", part " << sliceinfo[workslice].currentpart << endl;

					// quick check:
					if (sliceinfo[workslice].previousblock > 0)
					{
						// should not happen
						cout << myrank << ": WARNING: worker(): Simulation ALL_PARTS: Previous slice at position " << pos
							<< ", time " << time
							<< " is still blocked with " << sliceinfo[pos].previousblock
							<< " but we want to advance in the calculation!" << endl;
						dumptimestate();
					};
					if (OPTION_MEMORY_DROP_OW_OJ)
					{
						// Instead of disposing SimulationSlice,
						// we move it to the new position.
						// In this way, we get a free copy of the old w
						// to the new w. 
						// The new w will be called ow and W and will be
						// overwritten automatically.

						// save previous slice with old w field:
						pslice = sliceinfo[workslice].previousslice;
						//cout << myrank << ": Old slice pointer: " << (int) pslice << endl;
						
						// current slice gets old
						sliceinfo[workslice].previousslice = sliceinfo[workslice].currentslice;
						sliceinfo[workslice].previouspart = sliceinfo[workslice].currentpart;
						sliceinfo[workslice].previousblock = sliceinfo[workslice].currentblock;

						// Use the old slice as current slice.
						// (this contains the w field, which will be linked 
						//  to both, ow, and W)

						if (pslice == NULL)
						{
							// an old slice did not exist yet:
							// We have to create a new one:
							pslice = GetNewSimulationSlice();

							// and copy manually the w and j fields into the new slice:
							// NOTE that ow from the previous slice is stored in the new slice as w
							sscur = pslice;
							ssprev = sliceinfo[workslice].previousslice;
							//for (int sx=1;sx<=NUMx;sx++) 
								for (int sy=0;sy<=NUMT+1;sy++)
									for (int sz=0; sz<=NUM+1;sz++)
										for (int i=0;i<NW;i++)  
											sscur->w[0][sy][sz][i]=ssprev->w[0][sy][sz][i];

							// We also need to copy j from the previous slice to 
							// oj in the current slice:
							// NOTE that oj from the previous slice is stored in the new slice as j.
							//for (int sx=0;sx<=NUMx+1;sx++)
								for (int sy=0;sy<=NUMT+1;sy++)
									for (int sz=0;sz<=NUM+1;sz++)
										for (int i=0;i<3;i++) 
											sscur->j[0][sy][sz][i] = ssprev->j[0][sy][sz][i];
						};

						sliceinfo[workslice].currentslice = pslice;
					}
					else
					{
						// previous slice is of no use -> garbage can.
						DisposeSimulationSlice(sliceinfo[workslice].previousslice);
						
						// current slice gets old
						sliceinfo[workslice].previousslice = sliceinfo[workslice].currentslice;
						sliceinfo[workslice].previouspart = sliceinfo[workslice].currentpart;
						sliceinfo[workslice].previousblock = sliceinfo[workslice].currentblock;

						// Reserve new space for the simulation
						sliceinfo[workslice].currentslice = GetNewSimulationSlice();
					};

					if (MSG_STATEDUMP) dumptimestate();
					sliceinfo[workslice].time++; // increment time
					sliceinfo[workslice].currentpart = PART1; // This is the new part we fill.
					sliceinfo[workslice].currentblock = 0;
					if (MSG_STATEDUMP) dumptimestate();

					// Now we simulate the environment that is needed by solve.cpp
					worksliceleft = positivemod(workslice - 1, NUMT);
					worksliceright = positivemod(workslice + 1, NUMT);

					// cout << myrank << ": Prev " << (sliceinfo[workslice].previousslice == NULL) << ", cur " << (sliceinfo[workslice].currentslice == NULL) << endl;

					// First we link the previous slice at working position to 
					// lower case variables, and the current one to Upper case.
					sliceinfo[workslice].previousslice->linkGlobalLowerCase(sx);
					sliceinfo[workslice].currentslice->linkGlobalUpperCase(sx);

					if (MSG_PROGRAM) cout << myrank << ": Linked current at pos " << workslice << ", time " << sliceinfo[workslice].time << endl;

					// Now we link the left neighbor to lower case letters
					// using either current or previous state, depending on
					// whether it has evolved already.
					//
					// OPTION_MEMORY_DROP_OW_OJ:
					//    The field oj for left or right is either stored
					//    in the previous slice, or if it does not exist, 
					//    in the current slice.
					if (sliceinfo[worksliceleft].time == sliceinfo[workslice].time)
					{
						sliceinfo[worksliceleft].previousslice->linkGlobalLowerCase(sxleft);
						if (OPTION_MEMORY_DROP_OW_OJ) ::oj[sxleft] = sliceinfo[worksliceleft].currentslice->j[0];
					}
					else if (sliceinfo[worksliceleft].time == sliceinfo[workslice].time - 1)
					{
						sliceinfo[worksliceleft].currentslice->linkGlobalLowerCase(sxleft);
						if (OPTION_MEMORY_DROP_OW_OJ)
						{
							if (sliceinfo[worksliceleft].previousslice != NULL)
							{
								::oj[sxleft] = sliceinfo[worksliceleft].previousslice->j[0];
							}
							else
							{
								::oj[sxleft] = sliceinfo[worksliceleft].currentslice->j[0];
							};
						};
					}
					else
					{
						// out of sync! but should not be!
						cout << myrank << ": WARNING: worker(): timing inconsistency between neighbors" << endl;
					};

					// Now we do the same for the right neighbor
					if (sliceinfo[worksliceright].time == sliceinfo[workslice].time)
					{
						sliceinfo[worksliceright].previousslice->linkGlobalLowerCase(sxright);
						if (OPTION_MEMORY_DROP_OW_OJ) ::oj[sxright] = sliceinfo[worksliceright].currentslice->j[0];
					}
					else if (sliceinfo[worksliceright].time == sliceinfo[workslice].time - 1)
					{
						sliceinfo[worksliceright].currentslice->linkGlobalLowerCase(sxright);
						if (OPTION_MEMORY_DROP_OW_OJ) 
						{
							if (sliceinfo[worksliceright].previousslice != NULL)
							{
								::oj[sxright] = sliceinfo[worksliceright].previousslice->j[0];
							}
							else
							{
								::oj[sxright] = sliceinfo[worksliceright].currentslice->j[0];
							};
						};
					}
					else
					{
						// out of sync! but should not be!
						cout << myrank << ": WARNING: worker(): timing inconsistency between neighbors" << endl;
					};

					// Note that we do not assign Upper case letters to 
					// sxleft or sxright. 
					// => solve.cpp MUST NEVER write to Uppercase letters
					//    at sx = 0 or sx = 2.

					// Only the first time we call loadjs() and savejs():
					if (sliceinfo[workslice].time == 2)
					{
						if (MSG_PROGRAM) cout << myrank << ": loadjs()" << endl;
						loadjs();

						// savejs means to copy from j to oj, which is completely
						// fine here 
						if (MSG_PROGRAM) cout << myrank << ": savejs()" << endl;
						if (false)//  && OPTION_MEMORY_DROP_OW_OJ)
						{
							// We copy the j from the previous time-step to the current one.
							// We only have to do this for the current slice,
							// not for right and left neighbors.
							// (the j from the current time-step is the oj from the previous one.)

							//for (int sx=0;sx<=NUMTx1;sx++)
								for (int sy=0;sy<=NUMT+1;sy++)
								  for (int sz=0;sz<=NUM+1;sz++)
									for (int i=0;i<3;i++) 
										sliceinfo[workslice].currentslice->j[0][sy][sz][i] = 
											sliceinfo[workslice].previousslice ->j[0][sy][sz][i];
						}
						else
						{
							savejs();
						};
					};
/*
					// For now, we have to manually enforce boundary conditions
					// for the W:
					// First link to lower case letters, but then back again
					enforceBCsforslice(sx);
					enforceBCsforslice(sxleft);
					enforceBCsforslice(sxright);
					sliceinfo[workslice].currentslice->linkGlobalLowerCase(sx);
					enforceBCsforslice(sx);
					sliceinfo[workslice].previousslice->linkGlobalLowerCase(sx);
*/

					// evolve fields eps forward in time storing updated fields in captial vars
					if (MSG_PROGRAM) cout << myrank << ": doIncPart1()" <<endl;
					doIncPart1(eps); 

					if (OPTION_MEMORY_DROP_OW_OJ)
					{
						// The usual position for updateHLenergy() is below in PART1.
						// With OPTION_MEMORY_DROP_OW_OJ, we move it here to ALL_PARTS above.
						if (MSG_PROGRAM) cout << myrank << ": updateHLenergy()" <<endl;
						// update hard loop energy; integrated during run
						::HLenergy = sliceinfo[workslice].previousslice->HLenergy;
						updateHLenergy(eps); 
						sliceinfo[workslice].currentslice->HLenergy = ::HLenergy;
					};

					//// Print j and oj for check:
					//cout << myrank << ": "
					//	<< workslice << ", "
					//	<< sliceinfo[workslice].time << ", "
					//	<< "oj: " 
					//	<< ::oj[sxleft][1][1][1].a << ", "
					//	<< ::oj[sx][1][1][1].a << ", "
					//	<< ::oj[sxright][1][1][1].a << ", "
					//	<< endl;
					//cout << myrank << ": "
					//	<< workslice << ", "
					//	<< sliceinfo[workslice].time << ", "
					//	<< " j: " 
					//	<< ::j[sxleft][1][1][1].a << ", "
					//	<< ::j[sx][1][1][1].a << ", "
					//	<< ::j[sxright][1][1][1].a << ", "
					//	<< endl;

					// For now, we have to manually enforce boundary conditions
					// for the Capital letter variables:
					// First link to lower case letters, but then back again
					sliceinfo[workslice].currentslice->linkGlobalLowerCase(sx);
					enforceBCsforslice(sx);
					sliceinfo[workslice].previousslice->linkGlobalLowerCase(sx);

/*
					if ((sliceinfo[workslice].time-2)%SNAPUPDATE==0)
					{ 
						// snapshot(timefloat);
						
						// Let us send back the previous slice to the 
						// manager (as snapshot() refers to lower case letters)
						sinfo.pos = workslice; // what to send
						sinfo.node = 0; // node to send it to
						sinfo.time = sliceinfo[workslice].time;
						sinfo.part = PART1; //ALL_PARTS;
						
						cout << myrank << ": sendbuffer.push():" << endl;
						sendbuffer.dumpstate();
						sendbuffer.push(sinfo);
						sendbuffer.dumpstate();

						sliceinfo[workslice].currentblock++; // block the slice during send
					};
*/

					sinfo.pos = workslice; // what to send
					sinfo.time = sliceinfo[workslice].time;
					sinfo.part = PART1;

					// We send the updated data to the neighbor:
					if (workslice == positivemod(firstslice + 1, NUMT))
					{
						// talk to left neighbor:
						// Send data to left neighbor
						// Put this into the sendbuffer:
						sinfo.node = leftrank; // node to send it to
						
						if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();
						sendbuffer.push(sinfo);
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();

						sliceinfo[workslice].currentblock++; // block the slice during send
					}; // if
						
					if (workslice == positivemod(lastslice - 1, NUMT))
					{
						// talk to right neighbor
						// Send data to right neighbor
						// Put this into the sendbuffer:
						sinfo.node = rightrank; // node to send it to
						
						if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();
						sendbuffer.push(sinfo);
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();

						sliceinfo[workslice].currentblock++; // block the slice during send
					}; // if

					sinfo.node = 0; // node to send it to (Master)

					// Send to Master information about completion
					MPI_Bsend(&sinfo.pos, 4, MPI_INT,
					   sinfo.node, // dest
					   TAG_FINISHEDSLICE, // tag
					   MPI_COMM_WORLD);

					stats.stop(TIMER_PART1);
					break;
				case PART1: // ---------------------------- PART1 ----
					stats.start(TIMER_PART2);
					// We have already done PART 1, so now let us do PART 2:
					if (MSG_PROGRAM) cout << myrank << ": Simulate environment at pos " << workslice 
						<< ", time " << sliceinfo[workslice].time
						<< ", part " << sliceinfo[workslice].currentpart << endl;

					// Now we simulate the environment that is needed by solve.cpp
					worksliceleft = positivemod(workslice - 1, NUMT);
					worksliceright = positivemod(workslice + 1, NUMT);

					// cout << myrank << ": Prev " << (sliceinfo[workslice].previousslice == NULL) << ", cur " << (sliceinfo[workslice].currentslice == NULL) << endl;

					// First we link the previous slice at working position to 
					// lower case variables, and the current one to Upper case.
					sliceinfo[workslice].previousslice->linkGlobalLowerCase(sx);
					sliceinfo[workslice].currentslice->linkGlobalUpperCase(sx);

					if (MSG_PROGRAM) cout << myrank << ": Linked current at pos " << workslice << ", time " << sliceinfo[workslice].time << endl;

					// Now we link the left neighbor to lower case letters
					// and upper case letters.
					if (sliceinfo[worksliceleft].time == sliceinfo[workslice].time)
					{
						sliceinfo[worksliceleft].previousslice->linkGlobalLowerCase(sxleft);
						sliceinfo[worksliceleft].currentslice->linkGlobalUpperCase(sxleft);
					}
					else
					{
						// out of sync! but should not be!
						cout << myrank << ": WARNING: worker(): timing inconsistency between neighbors" << endl;
					};

					// Now we do the same for the right neighbor
					if (sliceinfo[worksliceright].time == sliceinfo[workslice].time)
					{
						sliceinfo[worksliceright].previousslice->linkGlobalLowerCase(sxright);
						sliceinfo[worksliceright].currentslice->linkGlobalUpperCase(sxright);
					}
					else
					{
						// out of sync! but should not be!
						cout << myrank << ": WARNING: worker(): timing inconsistency between neighbors" << endl;
					};

					// !! Instead of calling doIncPart2(eps) here,
					//    we first do the outputMeasurements.
					//    This is, because in the OPTION_MEMORY_DROP_OW_OJ
					//    mode, the ow field will be over-written by
					//    W field, and checkGaussLawPart1() uses ow.

					if (OPTION_MEMORY_DROP_OW_OJ)
					{
						// We have already called updateHLenergy() in ALL_PARTS,
						// so let us copy back this value here.
						::HLenergy = sliceinfo[workslice].currentslice->HLenergy;
					}
					else
					{
						// The usual position for updateHLenergy() is here.
						// With OPTION_MEMORY_DROP_OW_OJ, we move it to ALL_PARTS above.
						if (MSG_PROGRAM) cout << myrank << ": updateHLenergy()" <<endl;
						// update hard loop energy; integrated during run
						::HLenergy = sliceinfo[workslice].previousslice->HLenergy;
						updateHLenergy(eps); 
						sliceinfo[workslice].currentslice->HLenergy = ::HLenergy;
					};

					// Note that we have to set timefloat
					// (= 'time' in solve.cpp) before the following steps
					// as it might be different for each slice
					timefloat = (sliceinfo[workslice].time - 2)*eps;
		/*
					// measurements and data dump
					if (i==2) outputMeasurements(timefloat);
					if ( (i>1 && (i-1)%UPDATE==0)) {
						outputMeasurements(timefloat);
						// Disabled for now
						//if (fabs(energyDiff)>=MAXEVIOLATION) {
						//snapshot(timefloat);
						//break;
						//}
					}
		*/
					// 'i' in Evolve() corresponds to
					// i = sliceinfo[workslice].time - 1
					//if ((i-1)%SNAPUPDATE==0)			
					itime = sliceinfo[workslice].time-2;
					if ((itime%SNAPUPDATE==0)
						|| (itime%ANGLEUPDATE==0)
						|| ( (itime > datadumpinitialtime + 1)
						     && ((itime%CHECKPOINTUPDATE==1)
						         || (itime%CHECKPOINTUPDATE==2))))
					{ 
						// Either we do a snapshot, angle measurement,
						// or we do a data dump to disk.
						// (for the dump we need two consecutive times
						//  such that the simulation can re-start properly)

						// snapshot(timefloat);
						
						// Let us send back the previous slice to the 
						// manager (as snapshot() refers to lower case letters)
						sinfo.pos = workslice; // what to send
						sinfo.node = 0; // node to send it to
						sinfo.time = sliceinfo[workslice].time-1;
						sinfo.part = ALL_PARTS; //ALL_PARTS;
						
						if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();
						sendbuffer.push(sinfo);
						if (MSG_SENDBUFFER) sendbuffer.dumpstate();

						sliceinfo[workslice].previousblock++; // block the slice during send
					};

					// 'i' in Evolve() corresponds to
					// i = sliceinfo[workslice].time - 1
					//if ( (i>1 && (i-1)%UPDATE==0)) {
					if ((sliceinfo[workslice].time-1 == 1) // First time initial conditions
						||(sliceinfo[workslice].time-1 == 2)
						||(sliceinfo[workslice].time-1>1 && (sliceinfo[workslice].time-2)%UPDATE==0)) {
						if (MSG_OUTPUT) cout << myrank << ": Output Measurement "
							<< " at pos " << workslice
							<< ", time " << sliceinfo[workslice].time << endl;
						
						outputVariables ov;
						ov = outputMeasurementsMPIworker(timefloat);
						ov.pos = workslice;
						ov.time = sliceinfo[workslice].time;

						// Send this struct to the master:
						// (We send in Standard mode and provide buffer)
						MPI_Send(
							&ov,
							1, 
							myMPI_outputVariables,
							0, // dest
							TAG_OUTPUTVARIABLES, // tag
							MPI_COMM_WORLD);


						// Disabled for now
						//if (fabs(energyDiff)>=MAXEVIOLATION) {
						//snapshot(timefloat);
						//break;
						//}
					};

					if (MSG_PROGRAM) cout << myrank << ": Calling solve.cpp routines at pos " << workslice << ", time " << sliceinfo[workslice].time << endl;

					// Now we can call the usual solve.cpp routines:

					if (MSG_PROGRAM) cout << myrank << ": doIncPart2()" <<endl;
					// evolve fields eps forward in time storing updated fields in captial vars
					doIncPart2(eps); 

					if (MSG_STATEDUMP) dumptimestate();
					sliceinfo[workslice].currentpart = PART1 | PART2; // This is the new part we fill.
					if (MSG_STATEDUMP) dumptimestate();


					// increment time
				//	timefloat = i*eps;
					// copy fields from capital vars (updated) down to lowercase vars (current)
					//	 copyDown();
					// Instead of copyDown(),
					// we associate lower-case vars with updated time slice
					// We don't even need to do this explicitly, as we will
					// not use the lower case letters right now.
					// sliceinfo[workslice].currentslice->linkGlobalLowerCase(sx);


					// For now, we have to manually enforce boundary conditions
					// for the W:
					// First link to lower case letters, but then back again
					sliceinfo[workslice].currentslice->linkGlobalLowerCase(sx);
					enforceBCsforslice(sx);
					sliceinfo[workslice].previousslice->linkGlobalLowerCase(sx);

					// save currents for use in energy computation
					// This savejs is not needed explicitly, as it is contained
					// when linking the new time to lower case letters.
					// --- savejs();

					if (MSG_PROGRAM) cout << myrank << ": save w and j" <<endl;

					if (OPTION_MEMORY_DROP_OW_OJ)
					{
						// w->ow and j->oj are automatically
						// enforced if the time-step is increased.
						// (old slice gets into the position of the new one!)
					}
					else
					{
						// All that is left, is to copy the w from the previous slice
						// to ow in the current slice:
						sscur = sliceinfo[workslice].currentslice;
						ssprev = sliceinfo[workslice].previousslice;
						//for (int sx=1;sx<=NUMx;sx++) 
							for (int sy=0;sy<=NUMT+1;sy++)
								for (int sz=0; sz<=NUM+1;sz++)
									for (int i=0;i<NW;i++)  
										sscur->ow[0][sy][sz][i]=ssprev->w[0][sy][sz][i];

						// We also need to copy j from the previous slice to 
						// oj in the current slice:
						//for (int sx=0;sx<=NUMx+1;sx++)
							for (int sy=0;sy<=NUMT+1;sy++)
								for (int sz=0;sz<=NUM+1;sz++)
									for (int i=0;i<3;i++) 
										sscur->oj[0][sy][sz][i] = ssprev->j[0][sy][sz][i];
					};

					// Done with the work.

					if (OPTION_DYNAMICLOADBALANCING)
					{
						// If the current slice was next to the boundary,
						// we should claim victory and demand that the neighbor
						// sends us two of its slices.
						if (workslice == positivemod(firstslice + 1, NUMT))
						{
							// we are at the left border.
							if (MSG_DYNAMIC) cout << myrank << ": I completed the calculation at left border." << endl;

							if (bordermoveleft != 0)
							{
								if (MSG_DYNAMIC) cout << myrank << ": There is a pending left border move:" << bordermoveleft << endl;
								if (MSG_STATEDUMP) dumptimestate();

								// We already have a directive how to proceed:

								// Send finished data to neighbor, and
								// move border right afterwards:

								// Send data to left neighbor
								// Put this into the sendbuffer:
								sinfo.pos = workslice; // what to send
								sinfo.node = leftrank; // node to send it to
								sinfo.time = sliceinfo[workslice].time;
								sinfo.part = PART2;
								
								if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();
								sendbuffer.push(sinfo);
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();

								sliceinfo[workslice].currentblock++; // block the slice during send
								
								// I'm now responsible for freeing old resources:
								sliceinfo[firstslice].reset();

								// Move border
								firstslice = positivemod(firstslice + bordermoveleft, NUMT);
								bordermoveleft = 0;
								if (MSG_STATEDUMP) dumptimestate();

								// After we moved the border, a new element is 
								// the border element, so we may have to send
								// information about that, too to the neighbor:
								
								// First we set a new "workslice" (instead of a temp variable)
								workslice = positivemod(firstslice + 1, NUMT);
								if (sliceinfo[workslice].time == sliceinfo[firstslice].time)
								{
									// The interesting case is only, if the 
									// neighbor has finished some work
									// of the current time. Then we have to send
									// this to the neighbor as well. 
									// (Having a simulation block is equivalent to
									//  having finished the second part)
									// (He could not have finished more, 
									//  since we ourselves just finished this part.
									//  That's why we are at this point in the program.)

									// Send data to left neighbor
									// Put this into the sendbuffer:
									sinfo.pos = workslice; // what to send
									sinfo.node = leftrank; // node to send it to
									sinfo.time = sliceinfo[workslice].time;
									sinfo.part = sliceinfo[workslice].currentpart; // send what there is
									if (sliceinfo[workslice].simulationblock > 0)
									{
										// Simulationblock means, all has been evaluated:
										sinfo.part = ALL_PARTS;
									};
									if ((sliceinfo[workslice].currentpart != PART1)
										&& (sliceinfo[workslice].currentpart != ALL_PARTS))
									{
										cout << myrank << ": WARNING: order of part evaluation messed up!" << endl;
									};
									
									if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
									if (MSG_SENDBUFFER) sendbuffer.dumpstate();
									sendbuffer.push(sinfo);
									if (MSG_SENDBUFFER) sendbuffer.dumpstate();

									sliceinfo[workslice].currentblock++; // block the slice during send
								};
							}
							else if ((sliceinfo[firstslice].time > sliceinfo[workslice].time)
								|| ((sliceinfo[firstslice].time == sliceinfo[workslice].time)
								    && (sliceinfo[firstslice].currentpart == ALL_PARTS)))
							{
								if (MSG_DYNAMIC) cout << myrank << ": Border-Move: Left neighbor was faster." << endl;
								if (MSG_STATEDUMP) dumptimestate();
								// Our neighbor has been faster.
								// He has already sent us his piece.

								// Then we just send our piece to him as usual:

								// Send data to left neighbor
								// Put this into the sendbuffer:
								sinfo.pos = workslice; // what to send
								sinfo.node = leftrank; // node to send it to
								sinfo.time = sliceinfo[workslice].time;
								sinfo.part = PART2;
								
								if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();
								sendbuffer.push(sinfo);
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();

								sliceinfo[workslice].currentblock++; // block the slice during send
							}
							else
							{
								if (MSG_DYNAMIC) cout << myrank << ": I am first, I want to obtain two slices from left neighbor." << endl;
								if (MSG_STATEDUMP) dumptimestate();
								// No border move specified yet:

								// talk to left neighbor:
								MPI_Bsend(
									&sliceinfo[workslice].time, 1, MPI_INT,
									leftrank, // dest
									TAG_BM_PLEASESENDTWO_TORIGHT, // tag
									MPI_COMM_WORLD);

								// Hold the completion of the current part (so that the neighbor won't update too soon)
								sliceinfo[workslice].currentpart = PART1; // omitting PART2
								sliceinfo[workslice].simulationblock++;

								if (dynamicloadstateleft != STATE_BM_NOTWAITING)
								{
									cout << myrank << ": WARNING: Dynamic load left: System is in unexpected state "
										<< dynamicloadstateleft << endl;
								};

								dynamicloadstateleft = STATE_BM_WAITINGFORRESPONSE;
								if (MSG_STATEDUMP) dumptimestate();
							}; // if (bordermoveleft != 0)
						}; // if
							
						if (workslice == positivemod(lastslice - 1, NUMT))
						{
							// we are at the right border.
							if (MSG_DYNAMIC) cout << myrank << ": I completed the calculation at right border." << endl;

							if (bordermoveright != 0)
							{
								if (MSG_DYNAMIC) cout << myrank << ": I have a pending right border move." << endl;
								if (MSG_STATEDUMP) dumptimestate();
								// We already have a directive how to proceed:

								// Send finished data to neighbor, and
								// move border right afterwards:

								// Send data to right neighbor
								// Put this into the sendbuffer:
								sinfo.pos = workslice; // what to send
								sinfo.node = rightrank; // node to send it to
								sinfo.time = sliceinfo[workslice].time;
								sinfo.part = PART2;
								
								if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();
								sendbuffer.push(sinfo);
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();

								sliceinfo[workslice].currentblock++; // block the slice during send

								// I'm now responsible for freeing old resources:
								sliceinfo[lastslice].reset();

								// Move border:
								lastslice = positivemod(lastslice + bordermoveright, NUMT);
								bordermoveright = 0;
								if (MSG_STATEDUMP) dumptimestate();

								// First we set a new "workslice" (instead of a temp variable)
								workslice = positivemod(lastslice - 1, NUMT);
								if (sliceinfo[workslice].time == sliceinfo[lastslice].time)
								{
									// The interesting case is only, if the 
									// neighbor has finished some work
									// of the current time. Then we have to send
									// this to the neighbor as well. 
									// (Having a simulation block is equivalent to
									//  having finished the second part)
									// (He could not have finished more, 
									//  since we ourselves just finished this part.
									//  That's why we are at this point in the program.)

									// Send data to right neighbor
									// Put this into the sendbuffer:
									sinfo.pos = workslice; // what to send
									sinfo.node = rightrank; // node to send it to
									sinfo.time = sliceinfo[workslice].time;
									sinfo.part = sliceinfo[workslice].currentpart; // send what there is
									if (sliceinfo[workslice].simulationblock > 0)
									{
										// Simulationblock means, all has been evaluated:
										sinfo.part = ALL_PARTS;
									};
									if ((sliceinfo[workslice].currentpart != PART1)
										&& (sliceinfo[workslice].currentpart != ALL_PARTS))
									{
										cout << myrank << ": WARNING: order of part evaluation messed up!" << endl;
									};
									
									if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
									if (MSG_SENDBUFFER) sendbuffer.dumpstate();
									sendbuffer.push(sinfo);
									if (MSG_SENDBUFFER) sendbuffer.dumpstate();

									sliceinfo[workslice].currentblock++; // block the slice during send
								};
							}
							else if ((sliceinfo[lastslice].time > sliceinfo[workslice].time)
								|| ((sliceinfo[lastslice].time == sliceinfo[workslice].time)
								    && (sliceinfo[lastslice].currentpart == ALL_PARTS)))
							{
								if (MSG_DYNAMIC) cout << myrank << ": Border-Move: Right neighbor was faster." << endl;
								if (MSG_STATEDUMP) dumptimestate();
								// Our neighbor has been faster.
								// He has already sent us his piece.

								// Then we just send our piece to him as usual:
								// Send data to right neighbor
								// Put this into the sendbuffer:
								sinfo.pos = workslice; // what to send
								sinfo.node = rightrank; // node to send it to
								sinfo.time = sliceinfo[workslice].time;
								sinfo.part = PART2;
								
								if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();
								sendbuffer.push(sinfo);
								if (MSG_SENDBUFFER) sendbuffer.dumpstate();

								sliceinfo[workslice].currentblock++; // block the slice during send
							}
							else
							{
								if (MSG_DYNAMIC) cout << myrank << ": I am first, I want to obtain two slices from right neighbor." << endl;
								if (MSG_STATEDUMP) dumptimestate();
								// No border move specified yet:

								// talk to right neighbor
								MPI_Bsend(
									&sliceinfo[workslice].time, 1, MPI_INT,
									rightrank, // dest
									TAG_BM_PLEASESENDTWO_TOLEFT, // tag
									MPI_COMM_WORLD);

								// Hold the completion of the current part (so that the neighbor won't update too soon)
								sliceinfo[workslice].currentpart = PART1; // omitting PART2
								sliceinfo[workslice].simulationblock++;

								if (dynamicloadstateright != STATE_BM_NOTWAITING)
								{
									cout << myrank << ": WARNING: Dynamic load left: System is in unexpected state "
										<< dynamicloadstateright << endl;
								};

								dynamicloadstateright = STATE_BM_WAITINGFORRESPONSE;
								if (MSG_STATEDUMP) dumptimestate();
							}; // if (bordermoveright != 0)
						}; // if
					}
					else
					{
						// OPTION_DYNAMICLOADBALANCING == false

						// We don't move the border, but just send the
						// updated data to the neighbor:
						if (workslice == positivemod(firstslice + 1, NUMT))
						{
							// talk to left neighbor:
							// Send data to left neighbor
							// Put this into the sendbuffer:
							sinfo.pos = workslice; // what to send
							sinfo.node = leftrank; // node to send it to
							sinfo.time = sliceinfo[workslice].time;
							sinfo.part = PART2;
							
							if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
							if (MSG_SENDBUFFER) sendbuffer.dumpstate();
							sendbuffer.push(sinfo);
							if (MSG_SENDBUFFER) sendbuffer.dumpstate();

							sliceinfo[workslice].currentblock++; // block the slice during send
						}; // if
							
						if (workslice == positivemod(lastslice - 1, NUMT))
						{
							// talk to right neighbor
							// Send data to right neighbor
							// Put this into the sendbuffer:
							sinfo.pos = workslice; // what to send
							sinfo.node = rightrank; // node to send it to
							sinfo.time = sliceinfo[workslice].time;
							sinfo.part = PART2;
							
							if (MSG_SENDBUFFER) cout << myrank << ": sendbuffer.push():" << endl;
							if (MSG_SENDBUFFER) sinfo.dump();
							if (MSG_SENDBUFFER) sendbuffer.dumpstate();
							sendbuffer.push(sinfo);
							if (MSG_SENDBUFFER) sendbuffer.dumpstate();
							sliceinfo[workslice].currentblock++; // block the slice during send
						}; // if

					};

					sinfo.pos = workslice; // what to send
					sinfo.node = 0; // node to send it to (Master)
					sinfo.time = sliceinfo[workslice].time;
					sinfo.part = PART2;

					// Send to Master information about completion
					MPI_Bsend(&sinfo.pos, 4, MPI_INT,
					   sinfo.node, // dest
					   TAG_FINISHEDSLICE, // tag
					   MPI_COMM_WORLD);

					stats.stop(TIMER_PART2);

					// Now we also check if we had just completed a new
					// time-step. (meaning, we are the first completed
					// slice of the new time-step.)
					if (sliceinfo[workslice].time > currentmaximumtime)
					{
						currentmaximumtime = sliceinfo[workslice].time;

						// Let us check, if a new statistics output is due:
						if ((sliceinfo[workslice].time-2) % STATISTICSUPDATE == 0)
						{
							// Yes, send data to master:
							stats.time = sliceinfo[workslice].time;
							stats.stop(TIMER_TOTAL);
							stats.accumulate(); // add to total

							MPI_Send(
								&stats,
								1, 
								myMPI_statisticsVariables,
								0, // dest
								TAG_STATISTICSVARIABLES, // tag
								MPI_COMM_WORLD);

							stats.resetsection();
							stats.start(TIMER_TOTAL);
						};
					};

					// [ ] In case nobody needs the previous slice anymore,
					//     we could get rid of it, to save space (Opt01)

					// ...

					break;
				case PART2: // ---------------------------- PART2 ----
					cout << myrank << ": WARNING: This case should never happen PART2" << endl;
					assert(false); // whichwork() should never return such a slice.
					break;
				default:
					cout << myrank << ": WARNING: worker(): Unknown currentpart " << sliceinfo[workslice].currentpart << endl;
				}; // switch (sliceinfo[workslice].currentpart)
			}; // if (workslice >= 0)

			if ((simulationendreached == false)
				&& (sendstate == STATE_NOTSENDING) && (sendbuffer.empty()))
			{
				// Check if we have reached the end of the simulation:
				if (firstslice != -1)
				{
					simulationendreached = true; // wishful thinking..
					int iend = firstslice + positivemod(lastslice - firstslice, NUMT);
					for (int i = firstslice /*+ 1*/; i <= iend /*- 1*/; i++)
					{
						if (sliceinfo[positivemod(i,NUMT)].time < STEPS+1)
						{
							// cout << myrank << ": no simul end due to pos " << i << ", time " << sliceinfo[i].time << endl;
							simulationendreached = false;
							break;
						};
					};
				}
				else
				{
					// Currently we have no slices... does this mean we 
					// can not end?!? Deadlock?
					// => we must prevent that we later get no slices...
				};

				if (simulationendreached) 
				{
					// Reached end of simulation
					// let the manager quickly know.
					// If we are the first, we get a prize. :-)
					MPI_Send(
					   NULL, // buffer
					   0, // count
					   MPI_INT, // MPI_Datatype
					   0, // dest
					   TAG_REACHEDSIMULATIONEND, // tag
					   MPI_COMM_WORLD);
					// (the prize is to die collectively...)
				}; // if (!simulationendreached)
			} // if (workslice >= 0)
			else
			{
				// (workslice < 0)
				// Idle time:
				if (idlestate)
				{
					stats.check(TIMER_IDLE);
				}
				else
				{
					idlestate = true;
					stats.start(TIMER_IDLE);
				};
				if (MSG_IDLE)
				{
					if (MPI_Wtime() > idletimer)
					{
						cout << myrank << ": Idle" << endl;
						idletimer = MPI_Wtime() + 1; // 1 second before next message.
					};
				};
				// Could start taking the time here, until idle state is lost.

				// Don't waste CPU-time:
				mysleep(DEFAULT_SLEEP_MS);

				stats.check(TIMER_IDLE);
			};

		} // if (startsimulation >= 1)
		else
		{
			// Idle time BEFORE the simulation starts 
			// i.e. during the initial distribution of data.

			// Could take time here by starting a timer
			// that ends as soon as startsimulation changes state.

			// Don't waste CPU-time:
			mysleep(DEFAULT_SLEEP_MS);
		};
		stats.stop(TIMER_BLOCK_WORK);

	}; // while (dowork)


};


int main(int argc, char **argv)
{
	double seconds;

	// Initialize MPI
	MPI_Init(&argc, &argv);

	// Find out my identity in the default communicator
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	MPI_Get_processor_name(processor_name, &namelen);

	if (MSG_MPIINIT) cout << "Process " << myrank << " working on " << processor_name << endl;

	// obtain size of MPI world (= number of nodes)
	MPI_Comm_size(MPI_COMM_WORLD, &MPIsize); 

	// Create structures for communication
	commitoutputVariablesMPIstruct();
	commitstatisticsVariablesMPIstruct();

	// Allocate memory for buffered communication
	// (small packets like outputVariables)
	// (see http://www-unix.mcs.anl.gov/mpi/mpi-standard/mpi-report-1.1/node42.htm#Node42 )
	const int BUFFSIZE = 64*1024; // 64 kB
	MPI_Buffer_attach( malloc(BUFFSIZE), BUFFSIZE); 

	if (NUMT < 5)
	{
		cout << "WARNING! NUMT is too small to work in parallel" << endl;
	};
	if (NUMT < MPIsize - 1)  // exclude manager
	{
		cout << "WARNING! NUMT should be bigger than number of processors" << endl;
	};

	// Start clock
	startMPItime = MPI_Wtime();

	// we need at least 1 manager + 2 workers
	if (MPIsize > 2)
	{
		if (myrank == 0) {
			if (MSG_MPIINIT) cout << "Starting parallel execution "
				<< "on 1 manager and " << MPIsize-1 << " workers." << endl;
			manager();
		} else {
			worker();
		}
	}
	else
	{
		// Not enough workers:
		if (myrank == 0)
		{
			if (MSG_MPIINIT) cout << " ====== NOT ENOUGH WORKERS ====== " << endl;
			if (MSG_MPIINIT) cout << " Starting old routines (single process) " << endl;
			solve();
		}
		else
		{
			// Sorry, no work for the others...
		};
	};

	// Stop clock
	seconds = MPI_Wtime() - startMPItime;
	if ((MSG_MPIINIT) && (myrank == 0))
		cout <<  "Wall clock time used: " << seconds << " seconds." << endl;

	// Free MPI buffer:
	int buffsize;
	char *buff; 
	MPI_Buffer_detach( &buff, &buffsize); 
	free(buff);

	// Shut down MPI
	MPI_Finalize();
	return 0;
}
